/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "Econfig.h"
#include "Ecfg.h"
#include "Etypein.h"
#include "Eview.h"
#include "Emodel.h"
#include "Eactset.h"
#include "Ege.h"
#include "Egecb.h"
#include "Ecolors.h"
#include "Efonts.h"
#include "Eondisk.h"
#include "Eggroup.h"
#include "Esimple.h"

static Widget widget = NULL;    /* handle_command called for this widget */

static void init_command(void);
void uninstall_apply_to_view(EView *v_p, caddrt data);
static void apply_show_grid(EView *v_p, caddrt data, WCRec *);
static void apply_lock_grid(EView *v_p, caddrt data, WCRec *);
static void apply_lock_cplane(EView *v_p, caddrt data, WCRec *);
static void redraw(EView *v_p, caddrt data, WCRec *p);
static long unsigned mask_from_cl(void);
static void load_view_settings(EView *v_p, caddrt f, WCRec *s);
static void save_view_settings(EView *v_p, caddrt f, WCRec *s);

FILE *
ESISelFile(char *prompt, char *ok, char *cancel, char *failed,
           char *init_path, char *mode, char **name_return);


static void
help_me_please(void);
static void send_point(void);
static void send_point_delta(void);
static void show_in_view(void);
static void show_grid(void);
static void lock_in_view(void);
static void lock_to_grid(void);
static void lock_to_cplane(void);
static void save_model(void);
static void load_model(void);
static void fit_into_a_view(void);
static void set_active_style(void);
static void set_active_width(void);
static void set_active_rangle(void);
static void set_active_mtype(void);
static void set_active_msize(void);
static void set_active_color(void);
static void set_vclip(void);
static void set_mclip(void);
static void 
set_active_edgeflag(void);
static void 
set_active_edgecolor(void);
static void set_active_shrink(void);
static void set_active_scale(void);
static void set_active_fill(void);
static void set_active_font(void);
static void set_active_text(void);
static void set_active_pshow(void);
static void set_active_eshow(void);
static void set_fringe_table(void);

static void set_active_notclippable(void);
static void set_active_inviswhenclipped(void);

static void set_active_tesseli(void);
static void set_active_selcrit(void);
static void set_grid_d_or_o(void);
static void set_constrplane(void);
static void layers_setup(void);
static void ggroups(void);
static void view_setup(void);
static void set_keypintervals(void);
static void set_change_mask(void);
static void change_attribs(void);
static void set_render_shade(void);
static void new_frame(void);
static void graphic_info(void);
static void postscript_dump(void);
static void
image_import (void);
static void set_active_u_tesseli(void);
static void set_label(void);
static void cmd (void);
static void handle_pause (void);
static void handle_quit (void);



XEvent command_event;
BOOLEAN on_off_flag;

static BOOLEAN first_time = YES;



static EView *v_p_target = NULL;
static BOOLEAN can_use_target_view = TRUE;

static void
apply_command_to_view (ApplyFuncP func_p,
                       caddrt data_p,
                       ResetFuncP reset_func_p,
                       EModel *m_p
                       )
{
  if (EMCountDepViews (m_p) == 1)
    v_p_target = EMAnyViewP (m_p);
  else
    v_p_target = NULL;

  if (v_p_target != NULL && can_use_target_view) {
    WCRec p = {0,0,0};
    (*func_p) (v_p_target, data_p, &p);
  } else {
    EVSetApplyToViewFunction (func_p, data_p, reset_func_p);
    EMPushHandler (m_p, EVApplyToViewHandler, NULL);
  }
  can_use_target_view = TRUE;
}

static void
unset_target_view (void)
{
  can_use_target_view = FALSE;
}

static void 
set_model (void)
{
  EModel *m_p = NULL; /* NULL means reset to default */

  if (TypeInGetTokensTotal() >= 2) {
    if (TypeInGetTokenType (2) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (2));
    }
  }
  ESISetModel (m_p);
}


static void 
send_point(void)
{
  WCRec p;
  Window win_id;
  
  if (TypeInGetTokenType(2) != NUMBER)
    TypeInBadToken(2, "NUMBER");
  if (TypeInGetTokenType(3) != NUMBER)
    TypeInBadToken(3, "NUMBER");
  if (TypeInGetTokenType(4) != NUMBER)
    TypeInBadToken(4, "NUMBER");

  p.x = (FPNum)TypeInGetTokenAsDouble(2);
  p.y = (FPNum)TypeInGetTokenAsDouble(3);
  p.z = (FPNum)TypeInGetTokenAsDouble(4);

  EVSaveSentPoint(&p);
  win_id = XtWindow(EVViewPToWidget(EMAnyViewP(ESIModel())));
  command_event.xbutton.type       = ButtonPress;
  command_event.xbutton.send_event = True;
  command_event.xbutton.display    = XtDisplay(widget);
  command_event.xbutton.window     = win_id;
  command_event.xbutton.x          = 0;
  command_event.xbutton.y          = 0;
  command_event.xbutton.state      = Button1Mask;
  command_event.xbutton.button     = Button1;

  if (XSendEvent(XtDisplay(widget), win_id, True, (unsigned long)0,
		  &command_event) == 0)
    fprintf(stderr, "XSendEvent failed\n");
}
  


static void 
send_point_delta(void)
{
  WCRec last, p;
  Window win_id;

  if (TypeInGetTokenType(2) != NUMBER)
    TypeInBadToken(2, "NUMBER");
  if (TypeInGetTokenType(3) != NUMBER)
    TypeInBadToken(3, "NUMBER");
  if (TypeInGetTokenType(4) != NUMBER)
    TypeInBadToken(4, "NUMBER");

  p.x = (FPNum)TypeInGetTokenAsDouble(2);
  p.y = (FPNum)TypeInGetTokenAsDouble(3);
  p.z = (FPNum)TypeInGetTokenAsDouble(4);
  
  EVGetLastInputPoint(&last);
  p.x += last.x;
  p.y += last.y;
  p.z += last.z;
  EVSaveSentPoint(&p);
  win_id = XtWindow(EVViewPToWidget(EMAnyViewP(ESIModel())));
  command_event.xbutton.type       = ButtonPress;
  command_event.xbutton.send_event = True;
  command_event.xbutton.display    = XtDisplay(widget);
  command_event.xbutton.window     = win_id;
  command_event.xbutton.x          = 0;
  command_event.xbutton.y          = 0;
  command_event.xbutton.state      = Button1Mask;
  command_event.xbutton.button     = Button1;

  if (XSendEvent(XtDisplay(widget), win_id, True, (unsigned long)0,
		  &command_event) == 0)
    fprintf(stderr, "XSendEvent failed\n");
}
  
 

 




void
handle_command(Widget w, char *s)
{
  if (first_time){
    init_command();
    first_time = NO;
  }

  widget = w; /* called for this widget */
  
  if (TypeInParseLine(s)) {
    TypeInInterpretCurrentCommand();
  }
}



void 
uninstall_apply_to_view(EView *v_p, caddrt data)
{
  EMUninstallHandler(EVGetModel(v_p));
}



static void 
apply_show_grid(EView *v_p, caddrt data, WCRec *p)
{
  EVShowGrid(v_p, *(BOOLEAN*)data);
}



static void 
show_in_view(void)
{
  char *s;
  
  if (TypeInGetTokenType(2) != STRING)
    TypeInBadToken(2, "STRING");
  s = TypeInGetTokenAsString(2);
  if (TypeInCompareTokenWithKeyword(s, "GRID"))
    show_grid();
}



static void 
show_grid(void)
{
  int on_off_flag = YES;
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  
  if (TypeInIsKeywordInLine(2, "ON") > 0)
    on_off_flag = YES;
  else
    on_off_flag = NO;

  apply_command_to_view (apply_show_grid, (caddrt)&on_off_flag,
                         uninstall_apply_to_view, m_p);
}



static void 
lock_in_view(void)
{
  if (TypeInCompareTokenWithKeyword(TypeInGetTokenAsString(2), "GRID"))
    lock_to_grid();
  else if (TypeInCompareTokenWithKeyword(TypeInGetTokenAsString(2), "CPLANE"))
    lock_to_cplane();
}



static void 
apply_lock_grid(EView *v_p, caddrt data, WCRec *p)
{
  EVSetGridLock(v_p, *(BOOLEAN*)data);
  EVShowGrid(v_p, YES);
}



static void 
lock_to_grid(void)
{
  int on_off_flag = YES;
  int   kp;
  EModel *m_p = ESIModel ();
  
  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }

  if (TypeInIsKeywordInLine(2, "ON") > 0)
    on_off_flag = YES;
  else
    on_off_flag = NO;
  apply_command_to_view (apply_lock_grid, (caddrt)&on_off_flag,
                         uninstall_apply_to_view, m_p);
}


static void 
apply_lock_cplane(EView *v_p, caddrt data, WCRec *p)
{
  EVSetConstrPlaneLock(v_p, *(BOOLEAN*)data);
}



static void 
lock_to_cplane(void)
{
  int on_off_flag = YES;
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  
  if (TypeInIsKeywordInLine(2, "ON") > 0)
    on_off_flag = YES;
  else
    on_off_flag = NO;
  apply_command_to_view (apply_lock_cplane, (caddrt)&on_off_flag,
                         uninstall_apply_to_view, m_p);
}



static void 
save_model(void)
{
  char *name = NULL;
  FILE *fp;
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  if ((kp = TypeInIsKeywordInLine (1, "AS")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      name = TypeInGetTokenAsString (kp+1);
    } 
  } else {
    name = OnDiskGetFileName();
  }

  if (name == NULL) {
    fp = ESISelFile("Select file name",
                    "   OK   ", "CANCEL", NULL, NULL, "w", &name);
    if (fp) {
      fclose(fp);
    } else {
      return;                   /* Cancelled */
    }
  }
  OnDiskSetFileName (name);
  if ((kp = TypeInIsKeywordInLine (1, "VISIBLE")) > 0)  {
    EMSaveVisibleGraphics (m_p);  
  } else {
    EMSaveAllGraphics (m_p);
  }
  { char buf[1024];
    sprintf(buf, "\nSaved %d objects\n", EMLastIOCount (m_p));
    ERptStatusMessage(ELIXIR_MESSAGE_CLASS, 1, buf);
  }
}



static void 
load_model(void)
{
  FILE *fp;
  char *name = NULL;
  int kp;
  EModel *m_p = ESIModel ();
  
  /* expecting syntax: LOAD file_name */

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  
  if (TypeInGetTokensTotal() < 2) {
    fp = ESISelFile ("Select ELIXIR file",
                    "   OK   ", "CANCEL", NULL, NULL, "r", &name);
    if (fp) {
      OnDiskSetFileName(name);
      fclose(fp);
    } else {
      return;                   /* Cancelled */
    }
  } else {
    OnDiskSetFileName(TypeInGetTokenAsString(2));
  }
  EMRetrieveAllGraphics (m_p); 
  { char buf[1024];
    sprintf(buf, "\nRetrieved %d objects\n", EMLastIOCount (m_p));
    ERptStatusMessage(ELIXIR_MESSAGE_CLASS, 1, buf);
  }
}



static void 
apply_fit(EView *v_p, caddrt data, WCRec *p)
{
  EVFitAllIntoView(v_p);
}



static void 
fit_into_a_view(void)
{
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
    
  apply_command_to_view (apply_fit, NULL, uninstall_apply_to_view, m_p);
}



static void 
set_active_style(void)
{
  if (TypeInIsKeywordInLine(2, "SOLID") > 0)
    EASValsSetLineStyle(SOLID_STYLE);
  if (TypeInIsKeywordInLine(2, "DASH") > 0)
    EASValsSetLineStyle(DASHED_STYLE);
}



static void 
set_active_width(void)
{
  if (TypeInGetTokenType(2) != NUMBER)
    TypeInBadToken(2, "NUMBER");
  else {
    EASValsSetLineWidth((unsigned)TypeInGetTokenAsInt(2));
  }
}



static void 
set_active_rangle(void)
{
  if (TypeInGetTokenType(2) != NUMBER)
    TypeInBadToken(2, "NUMBER");
  else {
    FPNum angle;

    angle = TypeInGetTokenAsDouble(2);
    if (TypeInIsKeywordInLine(2, "DEG") >0 ||
	TypeInIsKeywordInLine(2, "DEGREE") >0) angle *= PI/180.0;
    
    EASValsSetRotationAngle(angle);
  }
}



static void 
set_active_mtype(void)
{
  if (TypeInGetTokenType(2) != NUMBER)
    TypeInBadToken(2, "NUMBER");
  else {
    EASValsSetMType((EMarkerType) TypeInGetTokenAsInt(2));
  }
}



static void 
set_active_msize(void)
{
  if (TypeInGetTokenType(2) != NUMBER)
    TypeInBadToken(2, "NUMBER");
  else {
    EASValsSetMSize((unsigned)TypeInGetTokenAsInt(2));
  }
}



static void 
set_active_pshow(void)
{
  EASValsSetShowPoly(!(TypeInIsKeywordInLine(2, "OFF") > 0));
}



static void 
set_active_eshow(void)
{
  EASValsSetShowEntity(!(TypeInIsKeywordInLine(2, "OFF") > 0));
}



static void 
set_active_notclippable(void)
{
  EASValsSetNotClippable(!(TypeInIsKeywordInLine(2, "OFF") > 0));
}


static void 
set_active_inviswhenclipped(void)
{
  EASValsSetInvisibleWhenClipped(!(TypeInIsKeywordInLine(2, "OFF") > 0));
}


static void 
set_active_tesseli(void)
{
  if (TypeInGetTokenType(2) != NUMBER)
    TypeInBadToken(2, "NUMBER");
  else {
    EASValsSetTesselIntervals(TypeInGetTokenAsInt(2));
  }
}



static void 
set_active_selcrit(void)
{
  int kp;
  
  if (TypeInIsKeywordInLine(2, "OVERLAP") > 0)
    EASValsSetSelectCriterion(OVERLAP);
  if (TypeInIsKeywordInLine(2, "INTERSECT") > 0)
    EASValsSetSelectCriterion(INTERSECT);
  if (TypeInIsKeywordInLine(2, "INSIDE") > 0)
    EASValsSetSelectCriterion(INSIDE);
  if (TypeInIsKeywordInLine(2, "ATTRIB") > 0) {
    if (TypeInIsKeywordInLine(2, "CLEAR") > 0) {
      EVSetAttribMask((unsigned long)0);
      EVSetTakeAsNOT(NO);
    } else 
      EVSetAttribMask(EVGetAttribMask() | mask_from_cl());
    if (TypeInIsKeywordInLine(2, "NOT") > 0)
      EVSetTakeAsNOT(YES);
  }
  if ((kp = TypeInIsKeywordInLine(2, "PICK_MASK")) > 0) {
    EVSetPickableMask(TypeInGetTokenAsInt(kp+1));
  }
}



static void 
set_keypintervals(void)
{
  if (TypeInGetTokenType(2) != NUMBER)
    TypeInBadToken(2, "NUMBER");
  else {
    EASValsSetKeyPIntervals((unsigned)TypeInGetTokenAsInt(2));
  }
}



static void 
set_active_color(void)
{
  EPixel clr;
  BOOLEAN     success;
  
  if (TypeInGetTokenType(2) != STRING)
    TypeInBadToken(2, "STRING");
  else {
    clr = ColorGetPixelFromString(TypeInGetTokenAsString(2), &success);
    if (success)
      EASValsSetColor(clr);
  }
}


static void 
set_active_edgecolor(void)
{
  EPixel clr;
  BOOLEAN     success;
  
  if (TypeInGetTokenType(2) != STRING)
    TypeInBadToken(2, "STRING");
  else {
    clr = ColorGetPixelFromString(TypeInGetTokenAsString(2), &success);
    if (success)
      EASValsSetEdgeColor(clr);
  }
}




static void 
set_active_font(void)
{
  Font fid;
  BOOLEAN     success;
  
  if (TypeInGetTokenType(2) != STRING)
    TypeInBadToken(2, "STRING");
  else {
    fid = FontGetFontFromString(TypeInGetTokenAsString(2), &success);
    if (success)
      EASValsSetFontId(fid);
  }
}



static void 
set_active_text(void)
{
  if (TypeInGetTokenType(2) != STRING)
    TypeInBadToken(2, "STRING");
  else {
    EASValsSetText(TypeInGetTokenAsString(2));
  }
}

#define DELTA  0
#define ORIGIN 1 

static int gflag = DELTA;
static FPNum val1 = 0., val2 = 0.;



static void apply_gdeltaorigin(EView *v_p, caddrt data, WCRec *p)
{
  if (gflag == DELTA)
    EVSetGridDeltas(v_p, val1, val2);
  else
    EVSetGridOrigin(v_p, val1, val2);
}



static void 
set_grid_d_or_o(void)
{
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
    
  if (TypeInGetTokenType(2) != NUMBER) {
    TypeInBadToken(2, "NUMBER");
    return;
  } else if (TypeInGetTokenType(3) != NUMBER) {
    TypeInBadToken(3, "NUMBER");
    return;
  }
  if (TypeInCompareTokenWithKeyword(TypeInGetTokenAsString(1), "GDELTA"))
    gflag = DELTA;
  else
    gflag = ORIGIN;
  val1 = TypeInGetTokenAsDouble(2);   val2 = TypeInGetTokenAsDouble(3);
  apply_command_to_view (apply_gdeltaorigin, NULL,
                         uninstall_apply_to_view, m_p);
}



static int layer = 0, layer_on = YES;

static void 
toggle_layer_onoff(EView *v_p, caddrt data, WCRec *p)
{
  EVToggleLayerOnOff(v_p, layer);
}

static void 
set_layer_onoff(EView *v_p, caddrt data, WCRec *p)
{
  EVSetLayerOnOff(v_p, layer, layer_on);
}

static void 
layers_setup(void)
{
  int pos;
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  
  if ((pos = TypeInIsKeywordInLine(2, "ACTIVE")) > 0) {
    if ((layer = TypeInGetTokenAsInt(pos+1)) >= 0) 
      EASValsSetLayer(layer);
  } else if (TypeInGetTokenType(2) == NUMBER) {
    if ((layer = TypeInGetTokenAsInt(2)) >= 0) 
      EASValsSetLayer(layer);
  } else if ((pos = TypeInIsKeywordInLine(2, "ON")) > 0) {
    layer_on = YES;
    if ((layer = TypeInGetTokenAsInt(pos+1)) >= 0) {
      apply_command_to_view (set_layer_onoff, NULL,
                             uninstall_apply_to_view, m_p);
    }
  } else if ((pos = TypeInIsKeywordInLine(2, "OFF")) > 0) {
    layer_on = NO;
    if ((layer = TypeInGetTokenAsInt(pos+1)) >= 0) {
      apply_command_to_view (set_layer_onoff, NULL,
                             uninstall_apply_to_view, m_p);
    }
  } else if ((pos = TypeInIsKeywordInLine(2, "TOGGLE")) > 0) {
    if ((layer = TypeInGetTokenAsInt(pos+1)) >= 0) {
      apply_command_to_view (toggle_layer_onoff, NULL,
                             uninstall_apply_to_view, m_p);
    }
  }
}



void 
ggroup_create(EView *v_p, caddrt data_p);
void 
ggroup_break(EView *v_p, caddrt data_p);



static void 
ggroups(void)
{
  int kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  
  if (TypeInIsKeywordInLine (2, "CREATE") > 0) 
    EVSetSelectActionFunction (ggroup_create, NULL, NULL);
  else if (TypeInIsKeywordInLine (2, "BREAK") > 0)
    EVSetSelectActionFunction (ggroup_break, NULL, NULL);
  else
    goto exit;
  EMPushHandler (m_p, EVSelectGraphicsHandler, NULL);
 exit:
  ;
}






static WCRec a_vect = {0, 0, 0};

static void 
view_setup_vo(EView *v_p, caddrt data, WCRec *p)
{
  EVSetOriginVC(v_p, p);
}


static double rotnor_angle = 0.1;
static int rotnor_dir = 2;
static void 
view_setup_rotnor (EView *v_p, caddrt data, WCRec *p)
{
  WCRec normal, e={0,0,0};
  
  switch (rotnor_dir) {
  case 0:
    e.x = rotnor_angle;
    break;
  case 1:
    e.y = rotnor_angle;
    break;
  case 2:
    e.z = rotnor_angle;
    break;
  default:
    return;
  }
  
  EVGetNormal (v_p, &normal);
  RotVectAboutVect (&e, &normal);
  EVSetNormal (v_p, &normal);
}


static double zoom_factor = 0.9;
static void 
view_zoom (EView *v_p, caddrt data, WCRec *p)
{
  FPNum xdim, ydim;
  
  EVGetDimsVC(v_p, &xdim, &ydim);
  EVSetDimsVC(v_p, xdim*zoom_factor, ydim*zoom_factor);
}


static void 
set_view_bkg(EView *v_p, EPixel bkg, caddrt data);

static void 
set_view_bkg(EView *v_p, EPixel bkg, caddrt data)
{
  EVSetBackground(v_p, bkg);
}

static void 
view_setup(void)
{
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  
  if (TypeInGetTokenType(3) == NUMBER &&
      TypeInGetTokenType(4) == NUMBER &&
      TypeInGetTokenType(5) == NUMBER)   {
    a_vect.x = TypeInGetTokenAsDouble(3);
    a_vect.y = TypeInGetTokenAsDouble(4);
    a_vect.z = TypeInGetTokenAsDouble(5);
    if        (TypeInIsKeywordInLine(2, "NOR") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
                             EVSetNormal,
			       (caddrt)&a_vect,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "VUP") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))EVSetVectUp,
			       (caddrt)&a_vect,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "VRP") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))EVSetVRP,
			       (caddrt)&a_vect,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "ORIGIN") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
			       EVSetOriginVC,
			       (caddrt)&a_vect,
			       uninstall_apply_to_view, m_p);
			EVSetApplyToViewPreventRedisplay(YES);
    }
  } else {
    if        (TypeInIsKeywordInLine(2, "TOP") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
			       EVSetViewOrientation,
			       (caddrt)VIEW_ORIENT_TOP,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "BOTTOM") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
			       EVSetViewOrientation,
			       (caddrt)VIEW_ORIENT_BOTTOM,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "LEFT") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
			       EVSetViewOrientation,
			       (caddrt)VIEW_ORIENT_LEFT,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "RIGHT") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
			       EVSetViewOrientation,
			       (caddrt)VIEW_ORIENT_RIGHT,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "FRONT") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
			       EVSetViewOrientation,
			       (caddrt)VIEW_ORIENT_FRONT,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "BACK") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
			       EVSetViewOrientation,
			       (caddrt)VIEW_ORIENT_BACK,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "ISO") > 0) {
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
			       EVSetViewOrientation,
			       (caddrt)VIEW_ORIENT_ISO,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "ORIGIN") > 0) {
      apply_command_to_view (view_setup_vo,
			       (caddrt)&a_vect,
			       uninstall_apply_to_view, m_p);
      unset_target_view ();
			EVSetApplyToViewPreventRedisplay(YES);
    } else if (TypeInIsKeywordInLine(2, "PRESERVE_VUP") > 0) {
      static BOOLEAN yesorno = YES;
      yesorno = YES;
      if        (TypeInIsKeywordInLine(2, "OFF") > 0) yesorno = NO;
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
                             EVPreserveVectUp,
                             (caddrt)yesorno,
                             uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "ZOOM") > 0) {
      zoom_factor = 0.95;
      if      (TypeInGetTokenType (3) == NUMBER)
        zoom_factor = TypeInGetTokenAsDouble (3);
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
                             view_zoom,
                             NULL,
                             uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "ROTNOR") > 0) {
      if      (TypeInGetTokenType (3) == NUMBER)
        rotnor_angle = TypeInGetTokenAsDouble (3);
      else if (TypeInGetTokenType (4) == NUMBER)
        rotnor_angle = TypeInGetTokenAsDouble (4);
      if      (TypeInIsKeywordInLine (1, "X") > 0) 
        rotnor_dir = 0;
      else if (TypeInIsKeywordInLine (1, "Y") > 0) 
        rotnor_dir = 1;
      else if (TypeInIsKeywordInLine (1, "Z") > 0) 
        rotnor_dir = 2;
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
                             view_setup_rotnor,
                             NULL,
                             uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "SAVE") > 0) {
      OnDiskSetFileName(TypeInGetTokenAsString(3));
      apply_command_to_view (save_view_settings,
                             (caddrt)&a_vect,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "LOAD") > 0) {
      OnDiskSetFileName(TypeInGetTokenAsString(3));
      apply_command_to_view (load_view_settings,
			       (caddrt)&a_vect,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "REDRAW") > 0) {
      apply_command_to_view (redraw,
			       (caddrt)&a_vect,
			       uninstall_apply_to_view, m_p);
    } else if (TypeInIsKeywordInLine(2, "BGROUND") > 0) {
      EPixel bkg;
      BOOLEAN suc;

      if (TypeInGetTokenType(3) == STRING) {
        bkg = ColorGetPixelFromString(TypeInGetTokenAsString(3), &suc);
        if (suc) {
          apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
                                   set_view_bkg,
                                   (caddrt)bkg,
                                   uninstall_apply_to_view, m_p);
        }
      }
    } else if (TypeInIsKeywordInLine(2, "AXES") > 0) {
      BOOLEAN on_off;
      on_off = TypeInIsKeywordInLine(2, "ON") > 0;
      apply_command_to_view ((void (*)(EView *, caddrt, WCRec *))
			       EVShowAxes,
			       (caddrt)on_off,
			       uninstall_apply_to_view, m_p);
    }
  }
}




static void change_att(EView *v_p, caddrt data_p);

static long unsigned change_mask = 0x0;

static void 
change_att(EView *v_p, caddrt data_p)
{
  GraphicObj *g_p;
  
  if (EVItemsInSelectionList() <= 0)
    return;

  while ((g_p = (GraphicObj *)EVGetFromSelectionList()) != NULL) 
    EMChangeGraphicAttributes(EVGetModel(v_p), change_mask, g_p);

  if ((change_mask & LAYER_MASK) || (change_mask & ALL_ATTRIB_MASK)) {
    EVForceRedisplay(v_p);
    EVRedrawDisplayList(v_p);
  } 
}


static void
change_reset(EView *v_p, caddrt data_p);
static void
change_reset(EView *v_p, caddrt data_p)
{
  EVSetSelectActionFunction(change_att, NULL, NULL);
  EMPushHandler(EVGetModel (v_p), EVSelectGraphicsHandler, NULL);
}


static void 
change_attribs(void)
{
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  
  if (TypeInGetTokensTotal() > 1) 
    change_mask = mask_from_cl();
  else
    change_mask = EASValsGetChangeMask();
  EVSetSelectActionFunction(change_att, NULL, NULL);
  EMPushHandler (m_p, EVSelectGraphicsHandler, NULL);
}



static long unsigned 
mask_from_cl(void)
{
  unsigned long mask = 0;

  if (TypeInIsKeywordInLine(2, "LAYER") > 0)
    mask |= LAYER_MASK; 
  if (TypeInIsKeywordInLine(2, "STYLE") > 0)
    mask |= STYLE_MASK;
  if (TypeInIsKeywordInLine(2, "COLOR") > 0)
    mask |= COLOR_MASK;
  if (TypeInIsKeywordInLine(2, "EDGECOLOR") > 0)
    mask |= EDGE_COLOR_MASK;
  if (TypeInIsKeywordInLine(2, "EDGEFLAG") > 0)
    mask |= EDGE_FLAG_MASK;
  if (TypeInIsKeywordInLine(2, "FONT") > 0)
    mask |= FONT_MASK;
  if (TypeInIsKeywordInLine(2, "WIDTH") > 0)
    mask |= WIDTH_MASK;
  if (TypeInIsKeywordInLine(2, "MTYPE") > 0)
    mask |= MTYPE_MASK;
  if (TypeInIsKeywordInLine(2, "MSIZE") > 0)
    mask |= MSIZE_MASK;
  if (TypeInIsKeywordInLine(2, "FILL") > 0)
    mask |= FILL_MASK;
  if (TypeInIsKeywordInLine(2, "SHRINK") > 0)
    mask |= SHRINK_MASK;
  if (TypeInIsKeywordInLine(2, "TESSEL") > 0 ||
      TypeInIsKeywordInLine(2, "TESSELATION") > 0)
    mask |= TESSEL_INTERVALS_MASK;
  if (TypeInIsKeywordInLine(2, "PSHOW") > 0)
    mask |= SHOW_POLY_MASK; 
  if (TypeInIsKeywordInLine(2, "ESHOW") > 0)
    mask |= SHOW_ENTITY_MASK;
  if (TypeInIsKeywordInLine(2, "NOTCLIPPABLE") > 0)
    mask |= NOT_CLIPPABLE_MASK;
  if (TypeInIsKeywordInLine(2, "INVISWHENCLIPPED") > 0)
    mask |= INVISIBLE_WHEN_CLIPPED_MASK;
  if (TypeInIsKeywordInLine(2, "UTESSEL") > 0)
    mask |= TESSEL_U_MASK;
  if (TypeInIsKeywordInLine(2, "VTESSEL") > 0)
    mask |= TESSEL_V_MASK;
    
  if (TypeInIsKeywordInLine(2, "CLEAR") > 0)
    mask = 0;
  
  return mask;
}



static void 
set_change_mask(void)
{
  EASValsSetChangeMask(mask_from_cl());
}



static void set_active_fill(void)
{
  int fill = FILL_SOLID;
  
  if      (   TypeInIsKeywordInLine(2, "SOLID") > 0
           || TypeInIsKeywordInLine(2, "ON") > 0)  fill = FILL_SOLID;
  else if (   TypeInIsKeywordInLine(2, "OFF") > 0 
	   || TypeInIsKeywordInLine(2, "HOLLOW") > 0) fill = FILL_HOLLOW;
  else
    return;
  EASValsSetFillStyle(fill);
}



static void 
set_active_edgeflag(void)
{
  int ef = NO;
  
  if      (TypeInIsKeywordInLine(2, "ON") > 0)  ef = YES;

  EASValsSetEdgeFlag(ef);
}



static void 
set_active_shrink(void)
{
  double s;

  if (TypeInGetTokenType(2) != NUMBER) {
    TypeInBadToken(2, "NUMBER");
    return;
  }
  s = TypeInGetTokenAsDouble(2);
  if (s < 0 || s > 1) s = 1;
  EASValsSetShrink(s);
}



static void 
set_active_scale(void)
{
  int pos;

  pos = TypeInIsKeywordInLine(2, "ISO");
  if (pos > 0) {
    if (TypeInGetTokenType(pos+1) != NUMBER){
      TypeInBadToken(pos+1, "NUMBER"); return;
    }
    EASValsSetIsotropicScale(TypeInGetTokenAsDouble(pos+1));
  } else {
    if (TypeInGetTokenType(2) != NUMBER) {TypeInBadToken(2, "NUMBER"); return;}
    if (TypeInGetTokenType(3) != NUMBER) {TypeInBadToken(3, "NUMBER"); return;}
    if (TypeInGetTokenType(4) != NUMBER) {TypeInBadToken(4, "NUMBER"); return;}
    EASValsSetScaleX(TypeInGetTokenAsDouble(2));
    EASValsSetScaleY(TypeInGetTokenAsDouble(3));
    EASValsSetScaleZ(TypeInGetTokenAsDouble(4));
  }
}



static VCRec dir_to_light = {1, 1, 1};

static void set_dir_to_light(EView *v_p, caddrt data, WCRec *p);

static void 
set_dir_to_light(EView *v_p, caddrt data, WCRec *p)
{
  EVSetDirTowardsLight(v_p, &dir_to_light);
}

static void set_render_mode(EView *v_p, caddrt data, WCRec *p);

static ERenderingType the_render_mode = NORMAL_RENDERING;
static EShadingType the_shade_mode = NO_SHADING;

static void 
set_render_mode(EView *v_p, caddrt data, WCRec *p)
{
  EVSetRenderMode(v_p, the_render_mode);
  EVSetShadeMode(v_p, the_shade_mode);
}


static void 
set_render_shade(void)
{
  int pos;
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }

  the_shade_mode = NO_SHADING;
  if (TypeInIsKeywordInLine(2, "NORMAL") > 0) {
    the_render_mode = NORMAL_RENDERING;
    apply_command_to_view (set_render_mode, NULL,
                           uninstall_apply_to_view, m_p);
  } else if ((pos = TypeInIsKeywordInLine(2, "LDIR")) > 0) {
    dir_to_light.u = TypeInGetTokenAsDouble(pos+1);
    dir_to_light.v = TypeInGetTokenAsDouble(pos+2);
    dir_to_light.n = TypeInGetTokenAsDouble(pos+3);
    apply_command_to_view (set_dir_to_light, NULL,
                           uninstall_apply_to_view, m_p);
  } else if ((pos = TypeInIsKeywordInLine(2, "AMBIENT")) > 0) {
    ColorSetAmbientGrey(TypeInGetTokenAsDouble(3));
    return; /* NO HANDLER TO USE */
  } else {
    if (TypeInIsKeywordInLine(2, "SHADE") > 0)  {
      the_render_mode = CONST_SHADING_RENDERING;
      if (TypeInIsKeywordInLine(2, "DITHER") > 0)
	the_shade_mode = DITHER_SHADING;
      else
	the_shade_mode = COLOR_SHADING;
    } else if (TypeInIsKeywordInLine(2, "FILL") > 0)  {
      the_render_mode = FILLED_HIDDEN_RENDERING;
    } else if (TypeInIsKeywordInLine(2, "WIRE") > 0)  {
      the_render_mode = WIRE_RENDERING;
    }
    apply_command_to_view (set_render_mode, NULL,
                           uninstall_apply_to_view, m_p);
  }
}


extern void
do_edit_color_scale (EView *v_p, caddrt data, WCRec *p);

static void 
apply_show_scale (EView *v_p, caddrt data, WCRec *p)
{
  EVShowScale(v_p, data);
}

static void 
set_fringe_table(void)
{
  double low = -1, high = 1;
  int kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }

  if (TypeInGetTokenType(2) == NUMBER && TypeInGetTokenType(3) == NUMBER) {
    low  = TypeInGetTokenAsDouble(2);
    high = TypeInGetTokenAsDouble(3);
    ColorSetupFringeTableByMinMax(EMGetAssocFringeTable(m_p),
                                  low, high);
  } else if (TypeInIsKeywordInLine (2, "SHOW_SCALE") > 0)  {
      BOOLEAN on_off;
      on_off = TypeInIsKeywordInLine(2, "ON") > 0;
      apply_command_to_view (apply_show_scale, (caddrt)on_off,
                             uninstall_apply_to_view, m_p);
  } else if (TypeInIsKeywordInLine (2, "GREY") > 0)  {
    SwitchFringeTableToGrey ();
  } else if (TypeInIsKeywordInLine (2, "COLOR") > 0)  {
    SwitchFringeTableToColor ();
  } else if (TypeInIsKeywordInLine (2, "SMOOTH") > 0)  {
    SetFringeTableSmoothScale ();
  } else if (TypeInIsKeywordInLine (2, "DISCRETE") > 0)  {
    SetFringeTableDiscreteScale ();
  } else {
    char buf[132];

    ColorFringesMinMax(EMGetAssocFringeTable(m_p), &low, &high);
    sprintf(buf, "Fringe table low = %g, high = %g", low, high);
    ERptStatusMessage(ELIXIR_MESSAGE_CLASS, 1, buf);
  }
}

static void 
new_frame(void)
{
  EView *v_p = NULL;
  int width = 500, height = 500, x = 0, y = 0;
  int had_width = TRUE, had_height = TRUE, had_x = FALSE, had_y = FALSE;
  int kp;
  char geom[132];
  char *name = NULL;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  } 
  if ((kp = TypeInIsKeywordInLine (1, "NAME")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      name = TypeInGetTokenAsString (kp+1);
    }
  }
  if ((kp = TypeInIsKeywordInLine (1, "WIDTH")) > 0)  {
    if (TypeInGetTokenType (kp+1) == NUMBER) {
      width = TypeInGetTokenAsInt (kp+1);
      had_width = TRUE;
    }
  }
  if ((kp = TypeInIsKeywordInLine (1, "HEIGHT")) > 0)  {
    if (TypeInGetTokenType (kp+1) == NUMBER) {
      height = TypeInGetTokenAsInt (kp+1);
      had_height = TRUE;
    }
  }
  if ((kp = TypeInIsKeywordInLine (1, "X")) > 0)  {
    if (TypeInGetTokenType (kp+1) == NUMBER) {
      x = TypeInGetTokenAsInt (kp+1);
      had_x = TRUE;
    }
  }
  if ((kp = TypeInIsKeywordInLine (1, "Y")) > 0)  {
    if (TypeInGetTokenType (kp+1) == NUMBER) {
      y = TypeInGetTokenAsInt (kp+1);
      had_y = TRUE;
    }
  }
  if (had_x && had_y)
    sprintf (geom, "%dx%d%+d%+d", width, height, x, y);
  else
    sprintf (geom, "%dx%d", width, height);

  v_p = ElixirNewNamedView (name, (char *) ESIAppName (), (char *) ESIClassName (),
                            "black", "white", geom);
  if (!v_p) {
    ERptErrMessage (1, 1, "Failed to create view (maybe it already exists?)", WARNING_GRADE);
    return;
  }
  
  EMAttachView (m_p, v_p);
  EMRegenerateGraphics (m_p, v_p);
  
  the_render_mode = NORMAL_RENDERING;
  the_shade_mode  = NO_SHADING;
  if (TypeInIsKeywordInLine(2, "SHADE") > 0)  {
    the_render_mode = CONST_SHADING_RENDERING;
    if (TypeInIsKeywordInLine(2, "DITHER") > 0)
      the_shade_mode = DITHER_SHADING;
    else
      the_shade_mode = COLOR_SHADING;
  } else if (TypeInIsKeywordInLine(2, "WIRE") > 0)  {
    the_render_mode = WIRE_RENDERING;
  }
  EVSetRenderMode(v_p, the_render_mode);
  EVSetShadeMode(v_p, the_shade_mode);
}



static void 
save_view_settings(EView *v_p, caddrt f, WCRec *s)
{
  FILE *fp;
  WCRec p;
  float x, y, z;
  
  OnDiskEmptyAGEFile();
  if ((fp = OnDiskOpenAGEFile("w")) == NULL)
    return;

  EVGetOriginVC(v_p, &p);
  x = p.x, y = p.y, z = p.z;
  fprintf(fp, "%f %f %f\n", x, y, z);
  EVGetVectUp(v_p, &p);
  x = p.x, y = p.y, z = p.z;
  fprintf(fp, "%f %f %f\n", x, y, z);
  EVGetNormal(v_p, &p);
  x = p.x, y = p.y, z = p.z;
  fprintf(fp, "%f %f %f\n", x, y, z);
  EVGetVRP(v_p, &p);
  x = p.x, y = p.y, z = p.z;
  fprintf(fp, "%f %f %f\n", x, y, z);
  EVGetDimsVC(v_p, &p.x, &p.y);
  x = p.x, y = p.y, z = p.z;
  fprintf(fp, "%f %f\n", x, y);
  
  OnDiskCloseAGEFile();
  EMUninstallHandler(EVGetModel(v_p)); /* do it for one view only. */
}



static void 
load_view_settings(EView *v_p, caddrt f, WCRec *s)
{
  FILE *fp;
  float x, y, z;
  WCRec p;
  
  if ((fp = OnDiskOpenAGEFile("r")) == NULL)
    return;

  fscanf(fp, "%f %f %f", &x, &y, &z);
  p.x = x, p.y = y, p.z = z;
  EVInhibitViewOrientationSetup(); EVSetOriginVC(v_p, &p);
  fscanf(fp, "%f %f %f", &x, &y, &z);
  p.x = x, p.y = y, p.z = z;
  EVInhibitViewOrientationSetup(); EVSetVectUp(v_p, &p);
  fscanf(fp, "%f %f %f", &x, &y, &z);
  p.x = x, p.y = y, p.z = z;
  EVInhibitViewOrientationSetup(); EVSetNormal(v_p, &p);
  fscanf(fp, "%f %f %f", &x, &y, &z);
  p.x = x, p.y = y, p.z = z;
  EVInhibitViewOrientationSetup(); EVSetVRP(v_p, &p);
  fscanf(fp, "%f %f", &x, &y);
  EVSetDimsVC(v_p, x, y);
  
  OnDiskCloseAGEFile();
  EVSetupViewOrientation(v_p);	/* This call resets the inhibit-flag  */
  EVSetupViewOrientation(v_p);
	EVGetVRP(v_p, &p);
	EVSetVRP(v_p, &p);
}

  
static void 
apply_set_constrpl_center(EView *v_p, caddrt data, WCRec *p);
static void 
apply_set_constrpl_orientation(EView *v_p, caddrt data, WCRec *p);

static WCRec constrpl_orientation_normal;
static WCRec constrpl_orientation_vup;
static WCRec constrpl_orientation_center;

static void 
apply_set_constrpl_center(EView *v_p, caddrt data, WCRec *p)
{
  EVSetConstrPlaneCenter(v_p, &constrpl_orientation_center);
}

static void 
apply_set_constrpl_orientation(EView *v_p, caddrt data, WCRec *p)
{
  EVSetConstrPlaneOrientation(v_p,
                              &constrpl_orientation_normal,
                              &constrpl_orientation_vup);
}

static void 
set_constrplane(void)
{
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }

  if (TypeInIsKeywordInLine(2, "CENTER") > 0) {
    if (TypeInGetTokenType(3) == NUMBER &&
	TypeInGetTokenType(4) == NUMBER &&
	TypeInGetTokenType(5) == NUMBER) {
      constrpl_orientation_center.x = TypeInGetTokenAsDouble(3);
      constrpl_orientation_center.y = TypeInGetTokenAsDouble(4);
      constrpl_orientation_center.z = TypeInGetTokenAsDouble(5);
    }
    apply_command_to_view (apply_set_constrpl_center, NULL,
			     uninstall_apply_to_view, m_p);
  } else if (TypeInIsKeywordInLine(2, "ORIENTATION") > 0) {
    if (TypeInGetTokenType(3) == NUMBER &&
	TypeInGetTokenType(4) == NUMBER &&
	TypeInGetTokenType(5) == NUMBER) {
      constrpl_orientation_normal.x = TypeInGetTokenAsDouble(3);
      constrpl_orientation_normal.y = TypeInGetTokenAsDouble(4);
      constrpl_orientation_normal.z = TypeInGetTokenAsDouble(5);
    }
    if (TypeInGetTokenType(6) == NUMBER &&
	TypeInGetTokenType(7) == NUMBER &&
	TypeInGetTokenType(8) == NUMBER) {
      constrpl_orientation_vup.x = TypeInGetTokenAsDouble(6);
      constrpl_orientation_vup.y = TypeInGetTokenAsDouble(7);
      constrpl_orientation_vup.z = TypeInGetTokenAsDouble(8);
    }
    apply_command_to_view (apply_set_constrpl_orientation,
                             (caddrt)NULL,
                             uninstall_apply_to_view, m_p);
  } else
    return;
}



static void
graphic_info(void)
{
  ginfo_CB(NULL, NULL, NULL);
}



int
ps_dump(EView *v_p, char *fname,
        double width_in_cm, double height_in_cm,
        BOOLEAN portrait, BOOLEAN encapsulated_ps
        );

static char psdump[2560];
static double psdump_x_cm = 0, psdump_y_cm = 0;
static BOOLEAN encapsulated_ps = NO;
static BOOLEAN portrait = NO;

static void 
apply_postscript_dump(EView *v_p, char *file, WCRec *p)
{
  ps_dump(v_p, psdump, psdump_x_cm, psdump_y_cm, portrait, encapsulated_ps);
}


static void
postscript_dump(void)
{
  FILE *fp;
  char *name = NULL;
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  
  strcpy(psdump, "!");
  
  if (TypeInGetTokensTotal() < 2) {
    fp = ESISelFile("Select file name for the Postscript dump",
                    "   OK   ", "CANCEL", NULL, NULL, "w", &name);
    if (fp) {
      strcpy(psdump, name);
      fclose(fp);
    }
    psdump_x_cm = psdump_y_cm = 15;
  } else {
    if (TypeInGetTokenType(2) == STRING) {
      strcpy(psdump, TypeInGetTokenAsString(2));
      if (   TypeInGetTokenType(3) == NUMBER
          && TypeInGetTokenType(4) == NUMBER) {
      psdump_x_cm = TypeInGetTokenAsDouble(3);
      psdump_y_cm = TypeInGetTokenAsDouble(4);
      } else {
        psdump_x_cm = psdump_y_cm = 10;
      }
    }
  }
  
  portrait = (TypeInIsKeywordInLine(2, "PORTRAIT") > 0);
  encapsulated_ps = (   TypeInIsKeywordInLine(2, "EPS") > 0
                     || TypeInIsKeywordInLine(2, "ENCAPSULATED") > 0);
  
  if (!(psdump[0] == '!')) {
    apply_command_to_view (apply_postscript_dump, (caddrt)psdump,
                             uninstall_apply_to_view, m_p);
  }
}



static WCRec  mclip_plane_center = {0, 0, 0};
static WCRec  mclip_plane_normal = {0, 0, 1};
static BOOLEAN mclip_plane_on = NO;

static void set_mclip_params(EView *v_p, caddrt data, WCRec *p);

static void 
set_mclip_params(EView *v_p, caddrt data, WCRec *p)
{
  EVSetModelClipPlane(v_p, 0, &mclip_plane_center, &mclip_plane_normal,
                      mclip_plane_on);
  EVSetModelPlaneClipOnOff(v_p, mclip_plane_on);
}


static void
set_mclip(void)
{
  int pos;
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  
  if (TypeInIsKeywordInLine(2, "ON") > 0)
    mclip_plane_on = YES;
  if (TypeInIsKeywordInLine(2, "OFF") > 0)
    mclip_plane_on = NO;
  if ((pos = TypeInIsKeywordInLine(2, "CENTER")) > 0) {
    mclip_plane_center.x = TypeInGetTokenAsDouble(pos+1);
    mclip_plane_center.y = TypeInGetTokenAsDouble(pos+2);
    mclip_plane_center.z = TypeInGetTokenAsDouble(pos+3);
  }
  if ((pos = TypeInIsKeywordInLine(2, "NORMAL")) > 0) {
    mclip_plane_normal.x = TypeInGetTokenAsDouble(pos+1);
    mclip_plane_normal.y = TypeInGetTokenAsDouble(pos+2);
    mclip_plane_normal.z = TypeInGetTokenAsDouble(pos+3);
  }
  apply_command_to_view (set_mclip_params, NULL,
                         uninstall_apply_to_view, m_p);
}



static void
set_vclip(void)
{
}


static void 
set_active_u_tesseli(void)
{
  if (TypeInGetTokenType(2) != NUMBER)
    TypeInBadToken(2, "NUMBER");
  else {
    EASValsSetUTesselIntervals(TypeInGetTokenAsInt(2));
  }
}

static void 
set_active_v_tesseli(void)
{
  if (TypeInGetTokenType(2) != NUMBER)
    TypeInBadToken(2, "NUMBER");
  else {
    EASValsSetVTesselIntervals(TypeInGetTokenAsInt(2));
  }
}



static void 
redraw (EView *v_p, caddrt data, WCRec *p)
{
  EVFastRedraw (v_p);
}



static void 
apply_image_import (EView *v_p, char *file, WCRec *p)
{
  char command[512];
  char buf[1024];
  
  sprintf (command, "import -silent -window 0x%x %s",
           (unsigned int)XtWindow (XtParent (EVViewPToWidget (v_p))),
           file);
  sprintf (buf, "\n  Saving view dump: %s", command);
  ERptStatusMessage(ELIXIR_MESSAGE_CLASS, 1, buf);
  system (command);
  ERptStatusMessage(ELIXIR_MESSAGE_CLASS, 1, "\n  Done");
}



static char imagefile[512];


static void
image_import (void)
{
  FILE *fp;
  char *name = NULL;
  int   kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
  
  strcpy (imagefile, "!");
  
  if (TypeInGetTokensTotal() < 2) {
    fp = ESISelFile("Select file name for the Postscript dump",
                    "   OK   ", "CANCEL", NULL, NULL, "w", &name);
    if (fp) {
      strcpy(imagefile, name);
      fclose(fp);
    } else {
      return;
    }
  } else {
    if (TypeInGetTokenType(2) == STRING) {
      strcpy(imagefile, TypeInGetTokenAsString(2));
    }
  }
  
  if (!(imagefile[0] == '!')) {
    apply_command_to_view (apply_image_import, (caddrt)imagefile,
                             uninstall_apply_to_view, m_p);
  }
}

static char *label = NULL;


static void 
apply_set_label (EView *v_p, char *l, WCRec *p)
{
  EVSetLabel(v_p, l);
  EVShowLabel(v_p, (l != NULL));
}

static void
set_label (void)
{
  int kp;
  EModel *m_p = ESIModel ();

  if ((kp = TypeInIsKeywordInLine (1, "MODEL")) > 0)  {
    if (TypeInGetTokenType (kp+1) == STRING) {
      m_p = ESINamedModel (TypeInGetTokenAsString (kp+1));
    }
  }
 
  label = NULL;
  if (TypeInGetTokensTotal() >= 2) {
    if (TypeInGetTokenType(2) == STRING) {
      label = malloc(strlen(TypeInGetTokenAsString(2))+1);
      sprintf(label,"%s",TypeInGetTokenAsString(2));
    }
  }
  
  apply_command_to_view (apply_set_label, (caddrt)label,
                         uninstall_apply_to_view, m_p);
  if (label) {
    free(label);
    label = NULL;
  }
}



static top_command   commands[] = {
 /* command string      its length    command handler */ 
  { "HELP",               4,           help_me_please     ,
      "HELP keyword" },
  { "XY",                 2,           send_point         ,
      "# # # (set point absolute coordinates)" },
  { "DX",                 2,           send_point_delta   ,
      "# # # (set point by delta from the last input point)" },
  { "SHOW",               4,           show_in_view       ,
      "{GRID} (show ... in view)" },
  { "LOCK",               4,           lock_in_view       ,
      "{GRID|CPLANE} (lock in view to ...)" },
  { "SAVE",               4,           save_model         ,
      "AS file_name (save graphics to a file)" },
  { "LOAD",               4,           load_model         ,
      "file_name (load graphics from a file)" },
  { "FIT",                3,           fit_into_a_view    ,
      "(fit all graphics into a view)" },
  { "STYLE",              5,           set_active_style   ,
      "{SOLID|DASH} (set line style)" },
  { "WIDTH",              5,           set_active_width   ,
      "# (set line width)" },
  { "RANGLE",             6,           set_active_rangle  ,
      "# [DEG] (set rotation angle)" },
  { "MTYPE",              5,           set_active_mtype   ,
      "# (set marker type - 0,1,2,3)" },
  { "MSIZE",              5,           set_active_msize   ,
      "# (set marker size in pixels)" },
  { "COLOR",              5,           set_active_color   ,
      "color_name (set color)" },
  { "FRINGE",              5,           set_fringe_table   ,
      "low high | ADOPT (set fringe table by minmax values)" },
  { "FILL",               4,           set_active_fill    ,
      "{OFF|{ON|SOLID}} (set fill style)" },
  { "EDGECOLOR",          8,           set_active_edgecolor,
      "color_name (set edge color)" },
  { "EDGEFLAG",          8,            set_active_edgeflag,
      "{ON|OFF} (set edge flag)" },
  { "SHRINK",             6,           set_active_shrink  ,
      "# (set shrink 0 < s && s <= 1)" },
  { "SCALE",              5,           set_active_scale  ,
      "# (set scale > 0)" },
  { "FONT",               4,           set_active_font    ,
      "font_name (set font)" },
  { "TEXT",               4,           set_active_text    ,
      "the_text (set text)" },
  { "PSHOW",              5,           set_active_pshow   ,
      "{ON|OFF} (set show-control-polygon flag)" },
  { "NOTCLIP",            7,           set_active_notclippable,
      "{ON|OFF} (set \"not clippable\" flag)" },
  { "INVISWH",            7,           set_active_inviswhenclipped,
      "{ON|OFF} (set \"invisible when clipped\" flag)" },
  { "ESHOW",              5,           set_active_eshow   ,
      "{ON|OFF} (set show-entity flag)" },
  { "TESSEL",             6,           set_active_tesseli ,
      "# (set tesselation intervals >0)" },
  { "SELCRIT",            7,           set_active_selcrit ,
      "{\tOVERLAP|INTERSECT|INSIDE|\n\t\tATTRIB {CLEAR|NOT|attrib_mask}|\n\t\tPICK_MASK #} (set selection criteria)" },
  { "LAYER",              5,           layers_setup       ,
      "{ACTIVE #|ON #|OFF #|TOGGLE} (set layer)" },
  { "GGROUP",             6,           ggroups            ,
      "{CREATE|BREAK} (create or break a graphic group)" },
  { "GDELTA",             6,           set_grid_d_or_o    ,
      "# # (set grid deltas in VC or in construction plane coordinates)" },
  { "GORIGIN",            7,           set_grid_d_or_o    ,
      "# # # (set grid origin in VC or in construction plane coordinates)" },
  { "CPLANE",             6,           set_constrplane    ,
      "{CENTER|ORIENTATION} # # # # # # (set construction plane \n"
        "                 orientation -- normal and the vector \"up\")" },
  { "VIEW",               4,           view_setup         ,
      "{{\tNOR|VUP|VRP|ORIGIN|ROTNOR|ZOOM} # # #|\n\tTOP|BOTTOM|LEFT|RIGHT|FRONT|BACK|ISO|\n\tSAVE name|LOAD name\n\tBGROUND color\n\tAXES {ON|OFF}} (set viewing parameters)" },
  { "KEYP",               4,           set_keypintervals  ,
      "# (set number of keypoints per graphic object)" },
  { "CHMASK",             6,           set_change_mask    ,
      "{\tLAYER|STYLE|COLOR|FONT|WIDTH|MTYPE|MSIZE|FILL|\n\t\tSHRINK|TESSEL|PSHOW|ESHOW} (set change mask)" },
  { "CHANGE",             6,           change_attribs     ,
      "change_mask (see CHMASK command) (change named attribute)" },
  { "RENDER",             6,           set_render_shade   ,
      "{NORMAL|LDIR # # #|AMBIENT #|SHADE [DITHER]|FILL|WIRE}"
      "(set rendering parameters in a view)" },
  { "NEWFRAME",           4,           new_frame          ,
      "[{SHADE|WIRE|HIDDEN}] (opens new view window)" },
  { "GINFO",              5,           graphic_info,
      "(prints info about graphic object)" },
  { "PSDUMP",             6,           postscript_dump,
      "ps_file [# #] [PORTRAIT][EPS] (creates postscript dump of a view window)" },
  { "IMAGE",              5,           image_import,
      "image_file (creates dump of a view window in formats recognized by import from ImageMagick)" },
  { "VCLIP",              5,           set_vclip,
      "(sets view clip params)" },
  { "MCLIP",              5,           set_mclip,
      "{CENTER # # #|NORMAL # # #|ON|OFF} (sets modelling clip params)" },
  { "UTESSEL",            7,           set_active_u_tesseli ,
      "# (set u tesselation intervals >0)" },
  { "VTESSEL",            7,           set_active_v_tesseli ,
      "# (set v tesselation intervals >0)" },
  { "MODEL",              5,           set_model ,
      "[#] (set target model; reset to default if no name is specified)" },
  { "LABEL",              5,           set_label ,
    "[#] (set view label; reset to no-label if no argument is specified)" },
  { "CMD",                3,           cmd,
    "[#] (execute command file)" },
  { "PAUSE",              5,           handle_pause,
    "[#] (pause for specified number of seconds)" },
  { "QUIT",              5,           handle_quit,
    " (quit)" },
  
  
};


static void 
init_command(void)
{
  TypeInAugmentCommandTable(commands, XtNumber(commands));
}


static void
help_me_please(void)
{
  int i, n = XtNumber(commands);
  char *s;
  top_command *cmd;

  if (TypeInGetTokensTotal() > 1) {
    s = TypeInGetTokenAsString(2);
    cmd = TypeInGetCommandData(s);
    if (cmd != NULL) {
      fprintf(stderr, "\n       Syntax of command %s\n", cmd->keyword);
      fprintf(stderr, "Notation: \n"
              "\t[] ...optional item,\n "
              "\t{s1|s2|..|sn} ...one of s1, s2,...,sn,\n"
              "\t# ... number\n");
      fprintf(stderr, "%s %s\n", cmd->keyword, cmd->syntax);
    } else {
      fprintf(stderr, "\n    No such command %s\n", cmd->keyword);
    }
    fprintf(stderr, "\n ");
  } else {
    fprintf(stderr,
            "=============>    Available commands:   <=============\n");
    for (i = 0; i < n; i++) {
      fprintf(stderr, "\t%s", commands[i].keyword);
      if (i/6*6 == i) fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n  Use command \"HELP keyword\" to get syntax. \n");
  }
}

static void
cmd (void)
{
  FILE *fp;
  char *name = NULL;
  int   kp;
  char cmdf[2048];

  cmdf[0] = ' ';
  if (TypeInGetTokensTotal() < 2) {
    fp = ESISelFile("Select file name of command file",
                    "   OK   ", "CANCEL", NULL, NULL, "w", &name);
    if (fp) {
      sprintf(cmdf, "%s", name);
      fclose(fp);
    } else {
      return;
    }
  } else {
    if (TypeInGetTokenType(2) == STRING) {
      sprintf(cmdf, "%s", TypeInGetTokenAsString(2));
    }
  }
  
  if (cmdf[0] != ' ') {
    tokens_parser p = tokens_new_parser ();
    if (p) {
      if (tokens_open_file (p, cmdf)) {
        while (tokens_next_line (p)) {
          const char *s = tokens_curr_line_contents (p);
          if (s && TypeInParseLine(s)) {
            TypeInInterpretCurrentCommand();
          }
        }
      }
      tokens_delete_parser (p);
    }
  }
}

static void
handle_pause (void)
{
  if (TypeInGetTokensTotal() >= 2) {
    double rs, us, ss;
    double r, u, s;
    double i;
    
    i = TypeInGetTokenAsDouble(2);
    
    ckit_get_times (&rs, &us, &ss);
    r = rs;
    while (r - rs < i) {
      ESIEventLoop (FALSE, "pausing...");
      ckit_get_times (&r, &u, &s);
    }
  }
}

static void
handle_quit (void)
{
  exit (1);
}
