#include "hexvue.h"
#include "hex.h"
#include "cut.h"
#include "crack.h"
#include "hvupriv.h"
#include "tri.h"
#include "quad.h"
#include "invmap.h"

/* 3D model of data on a hexahedral mesh  */

#define CMD_IS(cmd, keyw) (tokens_token_and_keyword_equiv (cmd, keyw))

#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

double tolerance = FLT_MAX; 

#undef NEED_HEX_DUMP
#ifdef NEED_HEX_DUMP

static int ddump = 0;
static double dshrink = 1.0;
static char dwhite[] = "white";
static char dyellow[] = "yellow";
static char dred[] = "red";
static char dblue[] = "blue";
static char *dcolor = dred;
static char *dedgecolor = dwhite;

static void
dump_hex_elixir_format(hexvue_hex_volume_t *h);
#  define DH(H) { if (ddump) dump_hex_elixir_format(H); } 
#else
#  define DH(H)
#endif

#define ALL_HEX_COORDS(hex)                                           \
  hex->points[0].loc.x, hex->points[0].loc.y, hex->points[0].loc.z,   \
  hex->points[1].loc.x, hex->points[1].loc.y, hex->points[1].loc.z,   \
  hex->points[2].loc.x, hex->points[2].loc.y, hex->points[2].loc.z,   \
  hex->points[3].loc.x, hex->points[3].loc.y, hex->points[3].loc.z,   \
  hex->points[4].loc.x, hex->points[4].loc.y, hex->points[4].loc.z,   \
  hex->points[5].loc.x, hex->points[5].loc.y, hex->points[5].loc.z,   \
  hex->points[6].loc.x, hex->points[6].loc.y, hex->points[6].loc.z,   \
  hex->points[7].loc.x, hex->points[7].loc.y, hex->points[7].loc.z

#define HEX_HAS_OFFSPRING(hex) (   hex->offspring[0] != NULL    \
                                || hex->offspring[1] != NULL    \
                                || hex->offspring[2] != NULL    \
                                || hex->offspring[3] != NULL    \
                                || hex->offspring[4] != NULL    \
                                || hex->offspring[5] != NULL    \
                                || hex->offspring[6] != NULL    \
                                || hex->offspring[7] != NULL)
#define Q4_HAS_OFFSPRING(q4) (   q4->offspring[0] != NULL       \
                              || q4->offspring[1] != NULL       \
                              || q4->offspring[2] != NULL       \
                              || q4->offspring[3] != NULL)



#define N_IPOL(hex, Nmat, dir)                  \
  (  Nmat[0] * hex->points[0].loc.dir           \
   + Nmat[1] * hex->points[1].loc.dir           \
   + Nmat[2] * hex->points[2].loc.dir           \
   + Nmat[3] * hex->points[3].loc.dir           \
   + Nmat[4] * hex->points[4].loc.dir           \
   + Nmat[5] * hex->points[5].loc.dir           \
   + Nmat[6] * hex->points[6].loc.dir           \
   + Nmat[7] * hex->points[7].loc.dir)
#define N_IPOL_LOCS(dest, vert, src, Nmat)              \
  {                                                     \
    dest->points[vert].loc.x = N_IPOL(src, Nmat, x);    \
    dest->points[vert].loc.y = N_IPOL(src, Nmat, y);    \
    dest->points[vert].loc.z = N_IPOL(src, Nmat, z);    \
  }



static int 
read_header(hexvue_model_t *v);
static int
read_data(hexvue_model_t *v, hexvue_bool_t split_by_cracks);

static int
read_crack_surface(hexvue_model_t *v, 
                   hexvue_bool_t split_by_cracks);
static int
read_hex(hexvue_model_t *v);
static int
read_point(hexvue_model_t *v);
static int
read_quad4_patch(hexvue_model_t *v);
static int
draw_cut(hexvue_cut_t *c, hexvue_model_t *v);
static int
draw_crack_surface(hexvue_crack_surface_t *c, hexvue_model_t *v);
static int
draw_contour(hexvue_contour_t *ct, hexvue_model_t *v);
static int
draw_poly(hexvue_cut_poly_t *p, hexvue_cut_t *c, hexvue_model_t *v);
static hexvue_contour_t *
new_contour(hexvue_model_t *v, double contour_value);
static void
pierce_hex(hexvue_hex_volume_t *h,
           hexvue_front_segment_t *cfs,
           hexvue_bool_t allow_lambda_outside,
           hexvue_crack_surface_t *cs);

hexvue_model_t *
hexvue_new_model (char *input_file,
                  hexvue_bool_t allow_presplit,
                  double prescribed_presplit_size,
                  hexvue_bool_t crack_split)
{
  hexvue_model_t *v;

  v = hexvue_new(hexvue_model_t);
  CHECK_MALLOC(v);
  DUPLICATE_STRING (v->in_file, input_file);
  v->hex_volumes    = make_list();
  v->crack_surfaces = make_list();

  v->quad4_patches  = make_list();

  v->points         = make_list();

  v->allow_presplit = allow_presplit;

  v->prescribed_presplit_size = prescribed_presplit_size;

  if (!tokens_open_file (input_file))
    REPORT_ERROR ("Failed open on input file");
  
  if (!read_header (v))
    goto error_exit;
  
  if (!read_data (v, crack_split))
    goto error_exit;

  tokens_close_curr_file ();

  v->active_value = NO_ACTIVE_VALUE;
  v->on_x = v->on_y = v->on_z = NO_ACTIVE_VALUE;
  v->scale_x = v->scale_y = v->scale_z = 1;

  v->min_value = -1;
  v->max_value =  1;

  v->cuts = NULL;
  v->contours = NULL;

  v->model = NULL;
  
  return v;
  
 error_exit:
  if (v) {
    if (v->data.values)
      hexvue_free(v->data.values);
    if (v->hex_volumes)
      free_list(v->hex_volumes, NOT_NODES);
    hexvue_free(v);
  }
  return NULL;
}


static int
merge_read_header (hexvue_model_t *v)
{
  char *s;
  int n, i;
  char *comment;
  int num_of_values;
  
  GET_LINE ({return 0;});
  DUPLICATE_STRING (comment, tokens_curr_line_contents ());

  GET_LINE ({return 0;});
  TOKEN_INT (1, num_of_values, {return 0;});

  if (num_of_values != v->data.num_of_values) {
    REPORT_ERROR ("num_of_values do not match for merged models");
    return 0;
  }

  n = (strlen (v->comment)
       + strlen ("; merged with   ")
       + strlen (comment)
       + 1);
  s = hexvue_new_array (n, char);
  sprintf  (s, "%s; merged with %s", v->comment, comment);
  free (v->comment);
  v->comment = s;

  FOR (i, v->data.num_of_values) {
    GET_LINE ({return 0;});
  }

  return 1;
}


hexvue_model_t *
hexvue_merge_model (hexvue_model_t *existing_model,
                    char *input_file,
                    hexvue_bool_t allow_presplit,
                    hexvue_bool_t split_when_merging)
{
  if (!tokens_open_file (input_file))
    REPORT_ERROR ("Failed open on input file");
  
  if (!merge_read_header (existing_model))
    goto error_exit;
  
  if (!read_data (existing_model, split_when_merging))
    goto error_exit;

  tokens_close_curr_file ();

  return existing_model;
  
 error_exit:
  return existing_model;
}


     
static int
read_header(hexvue_model_t *v)
{
  char *s;
  int i;
  
  GET_LINE ({return 0;});
  DUPLICATE_STRING (v->comment, tokens_curr_line_contents ());
  GET_LINE ({return 0;});
  TOKEN_INT (1, v->data.num_of_values, {return 0;});
  v->data.values
    = hexvue_new_array(v->data.num_of_values, hexvue_data_entry_t);
  CHECK_MALLOC (v->data.values);
  FOR (i, v->data.num_of_values) {
    GET_LINE ({return 0;});
    TOKEN_STR (1, s, {return 0;})
    DUPLICATE_STRING (v->data.values[i].description, s);
  }
  return 1;
}

  


static int
read_hex_binary (hexvue_model_t *v);
static int
read_hex_compressed (hexvue_model_t *v);
static void
split_by_cracks(hexvue_model_t *v);


static int
read_data (hexvue_model_t *v, hexvue_bool_t split)
{
  char *cmd;
  
  while (tokens_next_line ()) {
    TOKEN_STR(1, cmd, {goto error_exit;});
    if        (CMD_IS(cmd, "HV")) {
      if (!read_hex (v))
        return 0;
    } else if (CMD_IS(cmd, "BHV")) {
      if (!read_hex_binary (v))
        return 0;
    } else if (CMD_IS(cmd, "CHV")) {
      if (!read_hex_compressed (v))
        return 0;
    } else if (CMD_IS(cmd, "CS")) {
      if (!read_crack_surface (v, split))
        return 0;
    } else if (CMD_IS(cmd, "Q4")) {
      if (!read_quad4_patch (v))
        return 0;  
    } else if (CMD_IS(cmd, "PT")) {
      if (!read_point (v))
        return 0;
    } else if (CMD_IS(cmd, "END")) {
      if (split)
        split_by_cracks (v);
      return 1;
    }
  }
 error_exit:
  fprintf (stderr, "%s\n", tokens_curr_line_contents ());
  return 0;
}


static void
cut_by_crack(hexvue_crack_surface_t *cs, hexvue_model_t *v);


static void
split_by_cracks (hexvue_model_t *v)
{
  hexvue_crack_surface_t *cs;
  
  if (v->crack_surfaces) {
    loop_through_list (v->crack_surfaces, cs, hexvue_crack_surface_t *) {
      cut_by_crack (cs, v);
    }
  }
}

#define CBND(hbbox, hex, which, func, N)                        \
  {                                                             \
    hbbox.which.x = func(hbbox.which.x, hex->points[N].loc.x);  \
    hbbox.which.y = func(hbbox.which.y, hex->points[N].loc.y);  \
    hbbox.which.z = func(hbbox.which.z, hex->points[N].loc.z);  \
  }
#define DEFINE_HBBOX(hex, hbbox)                                     \
  {                                                                  \
    hbbox.lower.x = hbbox.upper.x = hex->points[0].loc.x;            \
    hbbox.lower.y = hbbox.upper.y = hex->points[0].loc.y;            \
    hbbox.lower.z = hbbox.upper.z = hex->points[0].loc.z;            \
    CBND(hbbox, hex, lower, min, 1);                                 \
    CBND(hbbox, hex, lower, min, 2);                                 \
    CBND(hbbox, hex, lower, min, 3);                                 \
    CBND(hbbox, hex, lower, min, 4);                                 \
    CBND(hbbox, hex, lower, min, 5);                                 \
    CBND(hbbox, hex, lower, min, 6);                                 \
    CBND(hbbox, hex, lower, min, 7);                                 \
    CBND(hbbox, hex, upper, max, 1);                                 \
    CBND(hbbox, hex, upper, max, 2);                                 \
    CBND(hbbox, hex, upper, max, 3);                                 \
    CBND(hbbox, hex, upper, max, 4);                                 \
    CBND(hbbox, hex, upper, max, 5);                                 \
    CBND(hbbox, hex, upper, max, 6);                                 \
    CBND(hbbox, hex, upper, max, 7);                                 \
  }
#define CBNDCP(cpbbox, cp, which, func, N)                       \
  {                                                              \
    cpbbox.which.x = func(cpbbox.which.x, cp->points[N].loc.x);  \
    cpbbox.which.y = func(cpbbox.which.y, cp->points[N].loc.y);  \
    cpbbox.which.z = func(cpbbox.which.z, cp->points[N].loc.z);  \
  }
#define DEFINE_CPBBOX(cp, cpbbox)                               \
  {                                                             \
    cpbbox.lower.x = cpbbox.upper.x = cp->points[0].loc.x;      \
    cpbbox.lower.y = cpbbox.upper.y = cp->points[0].loc.y;      \
    cpbbox.lower.z = cpbbox.upper.z = cp->points[0].loc.z;      \
    CBNDCP(cpbbox, cp, lower, min, 1);                          \
    CBNDCP(cpbbox, cp, lower, min, 2);                          \
    if (cp->npts == 4) {                                        \
      CBNDCP(cpbbox, cp, lower, min, 3);                        \
    }                                                           \
    CBNDCP(cpbbox, cp, upper, max, 1);                          \
    CBNDCP(cpbbox, cp, upper, max, 2);                          \
    if (cp->npts == 4) {                                        \
      CBNDCP(cpbbox, cp, upper, max, 3);                        \
    }                                                           \
  }
#define BOXES_OVERLAP(r1, r2)                                         \
         (   (r1.lower.x <= r2.upper.x) && (r1.upper.x >= r2.lower.x) \
          && (r1.lower.y <= r2.upper.y) && (r1.upper.y >= r2.lower.y) \
          && (r1.lower.z <= r2.upper.z) && (r1.upper.z >= r2.lower.z) \
          ) 
#define COMPUTE_DIAG_LENGTH(cpbbox)                                           \
 sqrt(  (cpbbox.upper.x - cpbbox.lower.x) * (cpbbox.upper.x - cpbbox.lower.x) \
      + (cpbbox.upper.y - cpbbox.lower.y) * (cpbbox.upper.y - cpbbox.lower.y) \
      + (cpbbox.upper.z - cpbbox.lower.z) * (cpbbox.upper.z - cpbbox.lower.z))
  


static int
how_to_presplit (hexvue_hex_volume_t *h, hexvue_crack_poly_t *cp,
                 hexvue_crack_surface_t *cs, double presplit_size)
{
  hexvue_box_t hbb, cspbb;
  double hcd, desired = cs->poly_char_dim;
  int nsubdiv = 0;

  DEFINE_CPBBOX(cp, cspbb);
  DEFINE_HBBOX(h, hbb);

  if (BOXES_OVERLAP (hbb, cspbb)) {
    hcd = COMPUTE_DIAG_LENGTH(hbb);
    desired = cs->poly_char_dim;
    if (presplit_size > 0)
      desired = presplit_size;
    nsubdiv = hcd / desired + 1;
  }
  return nsubdiv;
}


static int
do_presplit_hex(hexvue_model_t *v, hexvue_hex_volume_t *hp,
                int nsubdiv, LIST add_to_this);


static int
presplit_hex (hexvue_hex_volume_t *h, hexvue_crack_surface_t *cs,
              hexvue_model_t *v, LIST add_to_this)
{
  hexvue_crack_poly_t *csp;
  int nsubdiv;

  if (cs->crack_polygons) {
    loop_through_list(cs->crack_polygons, csp, hexvue_crack_poly_t *) {
      nsubdiv = how_to_presplit (h, csp, cs, v->prescribed_presplit_size);
      if (nsubdiv > 1) {
        do_presplit_hex (v, h, nsubdiv, add_to_this);
        return 1;
      }
    }
  }
  return 0;
}


#define P_NEWVA() hexvue_new_array(n, hexvue_float_t); 
#define P_NEWHEX(nh)                                            \
                   {  nh = hexvue_new(hexvue_hex_volume_t);     \
                      CHECK_MALLOC(nh);                         \
                      nh->points[0].values = P_NEWVA();         \
                      CHECK_MALLOC(nh->points[0].values);       \
                      nh->points[1].values = P_NEWVA();         \
                      CHECK_MALLOC(nh->points[1].values);       \
                      nh->points[2].values = P_NEWVA();         \
                      CHECK_MALLOC(nh->points[2].values);       \
                      nh->points[3].values = P_NEWVA();         \
                      CHECK_MALLOC(nh->points[3].values);       \
                      nh->points[4].values = P_NEWVA();         \
                      CHECK_MALLOC(nh->points[4].values);       \
                      nh->points[5].values = P_NEWVA();         \
                      CHECK_MALLOC(nh->points[5].values);       \
                      nh->points[6].values = P_NEWVA();         \
                      CHECK_MALLOC(nh->points[6].values);       \
                      nh->points[7].values = P_NEWVA();         \
                      CHECK_MALLOC(nh->points[7].values);       \
                      nh->level  = 0;                           \
                      nh->parent = NULL;                        \
                      nh->offspring[0]                          \
                        = nh->offspring[1]                      \
                          = nh->offspring[2]                    \
                            = nh->offspring[3]                  \
                              = nh->offspring[4]                \
                                = nh->offspring[5]              \
                                  = nh->offspring[6]            \
                                    = nh->offspring[7]          \
                                      = NULL;                   \
                      nh->grep = nh->plain_grep = NULL;         \
                      {                                         \
                        int _k_;                                \
                        FOR(_k_, 12) {                          \
                          nh->edge_greps[_k_] = NULL;           \
                        }                                       \
                      }                                         \
                    }
#define P_N_LOC(h, v, DIR, n)                  \
   v.DIR = (  n[0] * h->points[0].loc.DIR      \
            + n[1] * h->points[1].loc.DIR      \
            + n[2] * h->points[2].loc.DIR      \
            + n[3] * h->points[3].loc.DIR      \
            + n[4] * h->points[4].loc.DIR      \
            + n[5] * h->points[5].loc.DIR      \
            + n[6] * h->points[6].loc.DIR      \
            + n[7] * h->points[7].loc.DIR)

#define MAKE_VERTEX_AT(H, N, ksi, eta, tht, loc)                            \
   {                                                                        \
     hex_compute_N_matrix(N, ksi, eta, tht);                                \
     P_N_LOC(H, loc, x, N);                                                 \
     P_N_LOC(H, loc, y, N);                                                 \
     P_N_LOC(H, loc, z, N);                                                 \
   }
#define CPY_VALS(H, HP, V, _N_)                                             \
  {                                                                         \
    int _Mm_;                                                               \
    FOR(_Mm_, _N_) {                                                        \
      H->points[V].values[_Mm_] = HP->points[V].values[_Mm_];               \
    }                                                                       \
  }
#define IPOL_VALS(H, HP, V, _N_, Nmat)                                    \
  {                                                                       \
    int _M_;                                                              \
    FOR(_M_, _N_) {                                                       \
      H->points[V].values[_M_]  = (  HP->points[0].values[_M_] * Nmat[0]  \
                                   + HP->points[1].values[_M_] * Nmat[1]  \
                                   + HP->points[2].values[_M_] * Nmat[2]  \
                                   + HP->points[3].values[_M_] * Nmat[3]  \
                                   + HP->points[4].values[_M_] * Nmat[4]  \
                                   + HP->points[5].values[_M_] * Nmat[5]  \
                                   + HP->points[6].values[_M_] * Nmat[6]  \
                                   + HP->points[7].values[_M_] * Nmat[7]);\
    }                                                                     \
  }




static int
do_presplit_hex(hexvue_model_t *v, hexvue_hex_volume_t *hp,
                int nsubdiv, LIST add_to_this)
{
  hexvue_hex_volume_t *ho;
  double ksi, eta, tht, dksi, deta, dtht;
  int i, j, k;
  int n = v->data.num_of_values;
#define PARENT_MULT 1000000
  hexvue_float_t N[8];

  dksi = deta = dtht = (double)2 / (double)nsubdiv;

  ksi = -1;
  FOR(i, nsubdiv) {
    eta = -1;
    FOR(j, nsubdiv) {
      tht = -1;
      FOR(k, nsubdiv) {
        P_NEWHEX(ho);

        MAKE_VERTEX_AT(hp, N, ksi     , eta     , tht     , ho->points[0].loc);
        IPOL_VALS(ho, hp, 0, n, N);
        MAKE_VERTEX_AT(hp, N, ksi+dksi, eta     , tht     , ho->points[1].loc);
        IPOL_VALS(ho, hp, 1, n, N);
        MAKE_VERTEX_AT(hp, N, ksi+dksi, eta+deta, tht     , ho->points[2].loc);
        IPOL_VALS(ho, hp, 2, n, N);
        MAKE_VERTEX_AT(hp, N, ksi     , eta+deta, tht     , ho->points[3].loc);
        IPOL_VALS(ho, hp, 3, n, N);
        MAKE_VERTEX_AT(hp, N, ksi     , eta     , tht+dtht, ho->points[4].loc);
        IPOL_VALS(ho, hp, 4, n, N);
        MAKE_VERTEX_AT(hp, N, ksi+dksi, eta     , tht+dtht, ho->points[5].loc);
        IPOL_VALS(ho, hp, 5, n, N);
        MAKE_VERTEX_AT(hp, N, ksi+dksi, eta+deta, tht+dtht, ho->points[6].loc);
        IPOL_VALS(ho, hp, 6, n, N);
        MAKE_VERTEX_AT(hp, N, ksi     , eta+deta, tht+dtht, ho->points[7].loc);
        IPOL_VALS(ho, hp, 7, n, N);

        add_to_tail(add_to_this, ho);
        tht += dtht;
      } /* FOR(k, nsubdiv) */
      eta += deta;
    } /* FOR(j, nsubdiv)  */
    ksi += dksi;
  } /* FOR(i, nsubdiv) { */
  return 1;
}




static void
check_for_vertices_on_crack_surface(
                                    hexvue_hex_volume_t *h,
                                    hexvue_crack_surface_t *cs,
                                    hexvue_model_t *v
                                    );


static LIST hv = NULL;
static LIST h_list = NULL;

static void
cut_by_crack (hexvue_crack_surface_t *cs, hexvue_model_t *v)
{
  hexvue_hex_volume_t *h;
  hexvue_crack_front_t *cf;
  LIST tmp;
#define SWAP_LISTS() {tmp = v->hex_volumes; v->hex_volumes = hv; hv = tmp;} 
#define APPEND_LIST(dest, src)                                          \
  {                                                                     \
    hexvue_hex_volume_t *h_;                                            \
                                                                        \
    while ((h_ = (hexvue_hex_volume_t *)get_from_head(src)) != NULL)    \
      add_to_tail(dest, h_);                                            \
  }
  
  if (hv == NULL)       hv       = make_list();
  if (h_list == NULL) h_list = make_list();

  flush_list(hv, NOT_NODES);
  flush_list(h_list, NOT_NODES); 
  while ((h = (hexvue_hex_volume_t *)get_from_head(v->hex_volumes)) != NULL) {
    if (v->allow_presplit) {
      if (presplit_hex (h, cs, v, h_list)) {
	APPEND_LIST(hv, h_list);
	flush_list(h_list, NOT_NODES);
      } else {
        add_to_tail(hv, h);
      }
    } else {
      add_to_tail(hv, h);
    }
  }
  SWAP_LISTS(); /* v->hex_volumes is empty, hv holds those unsplit */
                /* and those generated inside presplit_hex() */

  loop_through_list(cs->crack_fronts, cf, hexvue_crack_front_t *) {
    flush_list(hv, NOT_NODES);
    flush_list(h_list, NOT_NODES); 
    while ((h = (hexvue_hex_volume_t *)get_from_head(v->hex_volumes)) != NULL){
      if (hvupriv_cut_hex_by_crack_front(h, cf,
					 v->data.num_of_values, h_list)) {
        APPEND_LIST(hv, h_list);
        flush_list(h_list, NOT_NODES);
      } else {
        add_to_tail(hv, h);
      }
    }
    SWAP_LISTS();
  }

  flush_list(hv, NOT_NODES);
  flush_list(h_list, NOT_NODES);
  loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
    if (hvupriv_cut_hex_by_crack_face(h, cs, v->data.num_of_values, h_list)) {
      APPEND_LIST(hv, h_list);
      flush_list(h_list, NOT_NODES);
    } else {
      check_for_vertices_on_crack_surface(h, cs, v);
      add_to_tail(hv, h);
    }
  }
  SWAP_LISTS(); 
}


static int
normal_intersects_poly(hexvue_loc_t *ip,
                       double *lambda,
                       hexvue_loc_t *a,
                       hexvue_vect_t *ab,
                       hexvue_crack_poly_t *cp,
                       double tol
                       );
static void
project_crack_polys(hexvue_vect_t *normal, hexvue_vect_t *center,
                    ListOf(hexvue_crack_poly_t *) polys
                    );
static void
ipol_on_polys(hexvue_hex_volume_t *h, int vert,
              hexvue_loc_t *center, int nvals);


static void
check_for_vertices_on_crack_surface(
                                    hexvue_hex_volume_t *h,
                                    hexvue_crack_surface_t *cs,
                                    hexvue_model_t *v
                                    )
{
  int i, n = v->data.num_of_values;
  hexvue_crack_poly_t *cp;
  hexvue_plane_t plane;
  double tol = tolerance / 1000;
  hexvue_loc_t a, ab, ip;
  double lambda;
  static ListOf(hexvue_crack_poly_t *) poly_list = NULL;
  hexvue_float_t centera[3];
  hexvue_loc_t center;
#define COMPUTE_CENTER(HEX)                                                  \
  {                                                                          \
    hex_center(ALL_HEX_COORDS(HEX), centera);                                \
    center.x = centera[0]; center.y = centera[1]; center.z = centera[2];     \
  }
#define COMP_NORMAL(cp)                                                  \
  {                                                                      \
    if        (cp->npts == 3) {                                          \
      plane_consts_3(&plane, &cp->points[0].loc,                         \
                     &cp->points[1].loc, &cp->points[2].loc);            \
    } else if (cp->npts == 4) {                                          \
      plane_consts_4(&plane,                                             \
                     &cp->points[0].loc, &cp->points[1].loc,             \
                     &cp->points[2].loc, &cp->points[3].loc);            \
    }                                                                    \
  }

  if (poly_list == NULL) poly_list = make_list();

  COMPUTE_CENTER(h);
  
  loop_through_list(cs->crack_polygons, cp, hexvue_crack_poly_t *) {
    COMP_NORMAL(cp);
    FOR(i, 8) {
      a = h->points[i].loc;
      v3_add(&a, &plane.n, &ab);
      if (normal_intersects_poly(&ip, &lambda, &a, &ab, cp, tol)) {
        if (fabs(lambda) < tol) {
          flush_list(poly_list, NOT_NODES);
          add_to_tail(poly_list, cp);
          project_crack_polys(&plane.n, &center, poly_list);
          ipol_on_polys(h, i, &center, n);
        }
      }
    }
  }
}



static int
normal_intersects_poly(hexvue_loc_t *ip,
                       double *lambda,
                       hexvue_loc_t *a,
                       hexvue_vect_t *ab,
                       hexvue_crack_poly_t *cp,
                       double tol
                       )
{
  hexvue_quad_t q;
  hexvue_tri_t t;
#define CPPT(n) cp->points[n].loc
#define TPT(n) t.loc[n]
#define QPT(n) q.loc[n]
  
  if (cp->npts == 3) {
    TPT(0) = CPPT(0); TPT(1) = CPPT(1); TPT(2) = CPPT(2);
    return hexvue_intersect_tri_and_line(ip, lambda, a, ab, &t, tol);
  } else if (cp->npts == 4) {
    QPT(0) = CPPT(0); QPT(1) = CPPT(1); QPT(2) = CPPT(2); QPT(3) = CPPT(3);
    return hexvue_intersect_quad_and_line(ip, lambda, a, ab, &q, tol);
  } else {
    REPORT_ERROR("poly not implemented");
  }
  return 0;
#undef CPPT
}



#define MAX_XP_PTS 12
typedef struct isect_points_t {
  hexvue_loc_t            xp;
  hexvue_loc_t            parspxp;
  hexvue_front_segment_t *xcfs;
  int                     on_edge;
  int                     on_vertex;
}              isect_points_t;

static isect_points_t isectp[MAX_XP_PTS];
static int nisectp;
#define MAX_HEXES_DIM 20
static int hexes_dim = MAX_HEXES_DIM;
static int nhexes;
static hexvue_float_t hexes[MAX_HEXES_DIM][8][3];
static hexvue_hex_vol_split_data_t hvs_hexes[MAX_HEXES_DIM];
static hexvue_plane_t plane;
static hexvue_vect_t AB, A;

static int
section_by_front(hexvue_hex_volume_t *h, LIST add_to_this,
                 hexvue_crack_front_t *cf, int num_of_values);



int
hvupriv_cut_hex_by_crack_front (hexvue_hex_volume_t *h,
                                hexvue_crack_front_t *cf,
                                int num_of_values,
                                LIST add_to_this)
{
  hexvue_front_segment_t *cfs;
  lobs_3d_front_end_t  *lobs = hvu_front_lobs ();
  G3D_box_t hbbox;
  
  DEFINE_HBBOX (h, hbbox);
  
  nisectp = 0;
  lobs_3d_search (lobs,
                  hbbox.lower.x, hbbox.lower.y, hbbox.lower.z,
                  hbbox.upper.x, hbbox.upper.y, hbbox.upper.z);
  cfs = lobs_3d_first_overlapping_object (lobs);
  while (cfs != NULL) {
    pierce_hex (h, cfs, FALSE, cf->crack_surface);
    cfs = lobs_3d_next_overlapping_object (lobs);
  }
  if (nisectp == 1 || nisectp == 2)
    return section_by_front (h, add_to_this, cf, num_of_values);
  else if (nisectp > 2) {
    /* RAW need to presplit the hex to achieve good cuts */
    return 0;
  } else
    return 0;
}



#define DEFINE_PBBOX(p, pbbox)                  \
  {                                             \
    pbbox.lower.x = pbbox.upper.x = p.x;        \
    pbbox.lower.y = pbbox.upper.y = p.y;        \
    pbbox.lower.z = pbbox.upper.z = p.z;        \
    pbbox.lower.x -= cs->poly_char_dim/2;       \
    pbbox.lower.y -= cs->poly_char_dim/2;       \
    pbbox.lower.z -= cs->poly_char_dim/2;       \
    pbbox.upper.x += cs->poly_char_dim/2;       \
    pbbox.upper.y += cs->poly_char_dim/2;       \
    pbbox.upper.z += cs->poly_char_dim/2;       \
  }

 
static void
approx_plane (hexvue_hex_volume_t *h, hexvue_crack_front_t *cf);


static void
approx_plane (hexvue_hex_volume_t *h, hexvue_crack_front_t *cf)
{
  hexvue_crack_surface_t *cs = cf->crack_surface;
  hexvue_crack_poly_t *cp;
  hexvue_box_t pbb0, pbb1, cspbb;
  hexvue_vect_t n = {0,0,0}, s, pss0, pss1;
  hexvue_float_t center[3];
  lobs_3d_front_end_t  *lobs = hvu_crack_lobs ();
  G3D_box_t hbbox;
#define CPPT(n) &cp->points[n].loc
#define INCR_N(n)                               \
  {                                             \
    if (v3_dot(&n, &plane.n) < 0) {             \
      n.x -= plane.n.x;                         \
      n.y -= plane.n.y;                         \
      n.z -= plane.n.z;                         \
    } else {                                    \
      n.x += plane.n.x;                         \
      n.y += plane.n.y;                         \
      n.z += plane.n.z;                         \
    }                                           \
  }

  DEFINE_HBBOX(h, hbbox);

  DEFINE_PBBOX(isectp[0].xp, pbb0);
  DEFINE_PBBOX(isectp[1].xp, pbb1);
  lobs_3d_search (lobs,
                  hbbox.lower.x, hbbox.lower.y, hbbox.lower.z,
                  hbbox.upper.x, hbbox.upper.y, hbbox.upper.z);
  cp = lobs_3d_first_overlapping_object (lobs);
  while (cp != NULL) {
    DEFINE_CPBBOX(cp, cspbb);
    if (BOXES_OVERLAP(cspbb, pbb0) || BOXES_OVERLAP(cspbb, pbb1)) {
      if (cp->npts == 3) {
        plane_consts_3(&plane, CPPT(0), CPPT(1), CPPT(2));
      } else if (cp->npts == 4) {
        plane_consts_4(&plane, CPPT(0), CPPT(1), CPPT(2), CPPT(3));
      } else {
        REPORT_ERROR("bad polygon");
      }
      INCR_N(n);
    }
    cp = lobs_3d_next_overlapping_object (lobs);
  }
  if (v3_length(&n) == 0) {
    REPORT_WARNING("no normal");
    n.x = 0; n.y = 0; n.z = 1;
  }
  hex_center(ALL_HEX_COORDS(h), center);
  s.x = center[0]; s.y = center[1]; s.z = center[2];
  hexvue_hex_inv_map(h, &s, 0.001, &pss0);
  s.x += n.x * cs->poly_char_dim / 100;
  s.y += n.y * cs->poly_char_dim / 100;
  s.z += n.z * cs->poly_char_dim / 100;
  hexvue_hex_inv_map(h, &s, 0.001, &pss1);
  plane.n.x = pss1.x - pss0.x;
  plane.n.y = pss1.y - pss0.y;
  plane.n.z = pss1.z - pss0.z;
  v3_normalize(&plane.n);
}



static int
section_by_front (hexvue_hex_volume_t *h, LIST add_to_this,
                  hexvue_crack_front_t *cf,
                  int num_of_values)
{
  int j, i, n = num_of_values;
  hexvue_float_t Nmat[8];
  hexvue_hex_volume_t *ho;
  hexvue_float_t xp[2][3];
#define ASSXP(i) { xp[i][0] = isectp[i].parspxp.x;      \
                   xp[i][1] = isectp[i].parspxp.y;      \
                   xp[i][2] = isectp[i].parspxp.z; }

  if (nisectp == 1) {
    nisectp = 0;
    pierce_hex(h, isectp[0].xcfs, TRUE, cf->crack_surface);
    if (nisectp != 2) return 0;
  }

  approx_plane (h, cf);

  FOR(i, 2) { ASSXP(i); }
  split_hex_by_front (&plane.n, xp, hexes_dim, &nhexes, hexes);

  if (nhexes == 0 )
    return 0;
  
  FOR(i, nhexes) {
    P_NEWHEX (ho);
    FOR(j, 8) {
      hex_compute_N_matrix (Nmat,
                            hexes[i][j][0], hexes[i][j][1], hexes[i][j][2]);
      N_IPOL_LOCS (ho, j, h, Nmat);
      IPOL_VALS (ho, h, j, n, Nmat);
    }
    add_to_tail (add_to_this, ho);
  }
  return 1;
}




static int
have_this_cut_point(hexvue_loc_t *p, double tol);

#define FV(face, vertex) hexvue_face_vertex_indices[face][vertex]

static void
pierce_hex(hexvue_hex_volume_t *h,
           hexvue_front_segment_t *cfs,
           hexvue_bool_t allow_lambda_outside,
           hexvue_crack_surface_t *cs)
{
  int face, vertex;
  double lambda;
  hexvue_vect_t s, pss;
  hexvue_quad_t q;
#define TOPO_EPS 0.001
#define IS_ON_SURFACE(pss) 1 /* (   fabs(fabs(pss.x) - 1) < TOPO_EPS      \
                            || fabs(fabs(pss.y) - 1) < TOPO_EPS      \
                            || fabs(fabs(pss.z) - 1) < TOPO_EPS) */
#define QPTS(face, vertex)                                      \
  {                                                             \
    q.loc[vertex].x = h->points[FV(face, vertex)].loc.x;        \
    q.loc[vertex].y = h->points[FV(face, vertex)].loc.y;        \
    q.loc[vertex].z = h->points[FV(face, vertex)].loc.z;        \
  }
  
  A = cfs->start;
  v3_sub(&cfs->end, &cfs->start, &AB);

  FOR(face, 6) {
    FOR(vertex, 4) { QPTS(face, vertex); }
    if (hexvue_intersect_quad_and_line(&s, &lambda, &A, &AB, &q,
                                       cs->poly_char_dim/100)) {
      if ((lambda >= 0 && lambda <= 1)
          || allow_lambda_outside) {
        hexvue_hex_inv_map(h, &s, TOPO_EPS, &pss);
        if (!IS_ON_SURFACE(pss))
          REPORT_WARNING("not on surface");
        if (nisectp < MAX_XP_PTS) {
          if (!have_this_cut_point(&s, cs->poly_char_dim/100)) {
            isectp[nisectp].parspxp.x  = pss.x;
            isectp[nisectp].parspxp.y  = pss.y;
            isectp[nisectp].parspxp.z  = pss.z;
            isectp[nisectp].xp.x       = s.x;
            isectp[nisectp].xp.y       = s.y;
            isectp[nisectp].xp.z       = s.z;
            isectp[nisectp].xcfs       = cfs;
            nisectp++;
          }
        } /* if (nisectp < MAX_XP_PTS)  */
      } /* if (   lambda >= 0 && lambda <= 1 */
    } /* if (hexvue_intersect_quad_and_line */
  } /* FOR(face, 6) */
}



static int
have_this_cut_point(hexvue_loc_t *p, double tol)
{
  int i;

  FOR(i, nisectp) {
    if (   fabs(p->x - isectp[i].xp.x) < tol
        && fabs(p->y - isectp[i].xp.y) < tol
        && fabs(p->z - isectp[i].xp.z) < tol)
      return 1;
  }
  return 0;
}


static void
update_tolerance (hexvue_hex_volume_t *h);




static int
read_crack_surface (hexvue_model_t *v,
                    hexvue_bool_t cut_when_merging)
{
  hexvue_crack_surface_t *cs;

  cs = hvupriv_read_crack_surface (v->data.num_of_values);
  if (cs == NULL)
    return 0;
  
  if (v->crack_surfaces == NULL) v->crack_surfaces = make_list();
  add_to_tail(v->crack_surfaces, cs);

  return 1;
}



static lobs_3d_front_end_t  *crack_lobs = NULL;
static G3D_box_t csp_bbox;
static lobs_3d_front_end_t  *front_lobs = NULL;
static G3D_box_t cfs_bbox;


lobs_3d_front_end_t  *
hvu_crack_lobs (void)
{
  return crack_lobs;
}


lobs_3d_front_end_t  *
hvu_front_lobs (void)
{
  return front_lobs;
}



hexvue_crack_surface_t *
hvupriv_read_crack_surface (int num_of_values)
{
  hexvue_crack_surface_t *cs;
  char *cmd;
  hexvue_crack_front_t *cf;
  hexvue_crack_poly_t *csp;

  if (crack_lobs == NULL) crack_lobs = lobs_3d_new_instance ();
  if (front_lobs == NULL) front_lobs = lobs_3d_new_instance ();
  
  cs = hexvue_new (hexvue_crack_surface_t);
  CHECK_MALLOC (cs);

  cs->crack_polygons = make_list();
  cs->crack_fronts   = make_list();
  cs->poly_char_dim  = 0;

  while (tokens_next_line ()) {
    TOKEN_STR(1, cmd, {return NULL;});
    if        (CMD_IS(cmd, "CP")) {
      csp = hvupriv_read_crack_polygon (num_of_values);
      if (csp == NULL)
	return NULL;
      add_to_tail (cs->crack_polygons, csp);
      G3D_INIT_BOX (csp_bbox, csp->points[0].loc);
      G3D_UPDT_BOX (csp_bbox, csp->points[1].loc);
      G3D_UPDT_BOX (csp_bbox, csp->points[2].loc);
      if (csp->npts == 4) 
        G3D_UPDT_BOX (csp_bbox, csp->points[3].loc);
      lobs_3d_add_object (crack_lobs, csp,
                          csp_bbox.lower.x,csp_bbox.lower.y,csp_bbox.lower.z,
                          csp_bbox.upper.x,csp_bbox.upper.y,csp_bbox.upper.z);
    } else if (CMD_IS(cmd, "CF")) {
      cf = hvupriv_read_front (num_of_values);
      if (cf == NULL)
	return NULL;
      cf->crack_surface = cs;
      if (cs->crack_fronts == NULL) cs->crack_fronts = make_list();
      add_to_tail (cs->crack_fronts, cf);
    } else if (CMD_IS(cmd, "CHAR_DIM")) {
      TOKEN_DBL(2, cs->poly_char_dim, {goto error_exit;});
    } else if (CMD_IS(cmd, "CSEND")) {  
      hvupriv_setup_crack_surface (num_of_values, cs);
      return cs;
    }
  }
 error_exit:
  return NULL;
}


hexvue_crack_front_t *
hvupriv_read_front (int num_of_values)
{
  hexvue_front_segment_t *cfs;
  hexvue_crack_front_t *cf;
  char *cmd;

  cf = hexvue_new (hexvue_crack_front_t);
  CHECK_MALLOC(cf);
  cf->crack_surface = NULL;
  cf->segments = make_list();

  while (tokens_next_line ()) {
    TOKEN_STR (1, cmd, {return NULL;});
    if        (CMD_IS (cmd, "SEG")) {
      cfs = hexvue_new (hexvue_front_segment_t);
      CHECK_MALLOC (cfs);
      TOKEN_DBL (2, cfs->start.x, {goto error_exit;});
      TOKEN_DBL (3, cfs->start.y, {goto error_exit;});
      TOKEN_DBL (4, cfs->start.z, {goto error_exit;});
      TOKEN_DBL (5, cfs->end.x,   {goto error_exit;});
      TOKEN_DBL (6, cfs->end.y,   {goto error_exit;});
      TOKEN_DBL (7, cfs->end.z,   {goto error_exit;});
      add_to_tail (cf->segments, cfs);
      G3D_INIT_BOX (cfs_bbox, cfs->start);
      G3D_UPDT_BOX (cfs_bbox, cfs->end);
      lobs_3d_add_object (front_lobs, cfs,
                          cfs_bbox.lower.x,cfs_bbox.lower.y,cfs_bbox.lower.z,
                          cfs_bbox.upper.x,cfs_bbox.upper.y,cfs_bbox.upper.z);
    } else if (CMD_IS (cmd, "CFEND")) {
      return cf;
    }
  }
 error_exit:
  return NULL;  
}



static hexvue_crack_poly_t *
hvupriv_read_crack_polygon(int num_of_values)
{
  hexvue_crack_poly_t *csp;
  int j, i, n = num_of_values;

  csp = hexvue_new (hexvue_crack_poly_t);
  CHECK_MALLOC(csp);
  GET_LINE({goto error_exit;});
  TOKEN_INT(1, csp->npts, {goto error_exit;});
  if (!(csp->npts == 3 || csp->npts == 4)) {
    REPORT_ERROR("Bad number of crack surface polygon vertices");
  }
    
  FOR(i, csp->npts) {
    GET_LINE({goto error_exit;});
    TOKEN_DBL(1, csp->points[i].loc.x, {goto error_exit;});
    TOKEN_DBL(2, csp->points[i].loc.y, {goto error_exit;});
    TOKEN_DBL(3, csp->points[i].loc.z, {goto error_exit;});
    csp->points[i].values_on_plus_side
      = hexvue_new_array(n, hexvue_float_t);
    csp->points[i].values_on_minus_side
      = hexvue_new_array(n, hexvue_float_t);
    FOR(j, n) {
      GET_LINE({goto error_exit;});
      TOKEN_DBL(1, csp->points[i].values_on_plus_side[j], {goto error_exit;});
      TOKEN_DBL(2, csp->points[i].values_on_minus_side[j], {goto error_exit;});
    }
  }
  csp->plus_side_grep        = NULL;
  csp->minus_side_grep       = NULL;
  csp->plus_side_plain_grep  = NULL;
  csp->minus_side_plain_grep = NULL;


  return csp;
  
 error_exit:
  return NULL;
}


static void
construct_hex_buf(int num_of_values);

static hexvue_hex_volume_t *hex_buf= NULL;

static void
construct_hex_buf(int num_of_values)
{
  int j, n = num_of_values;
  
  hex_buf = hexvue_new (hexvue_hex_volume_t);
  CHECK_MALLOC(hex_buf);
  FOR(j, 8) {
    hex_buf->points[j].values = hexvue_new_array(n, hexvue_float_t);
  }
  hex_buf->parent = NULL; /* top-level hex */
  hex_buf->offspring[0] = 
  hex_buf->offspring[1] = 
  hex_buf->offspring[2] = 
  hex_buf->offspring[3] = 
  hex_buf->offspring[4] = 
  hex_buf->offspring[5] = 
  hex_buf->offspring[6] = 
  hex_buf->offspring[7] = NULL;
  hex_buf->level = 0;
  hex_buf->grep = NULL;
  hex_buf->plain_grep = NULL;
  FOR(j, 12) {
    hex_buf->edge_greps[j] = NULL;
  }
}


static hexvue_hex_volume_t *
copy_hex_buf(int num_of_values);


static hexvue_hex_volume_t *
copy_hex_buf(int num_of_values)
{
  hexvue_hex_volume_t *h;
  int j, i, n = num_of_values;

  h = hexvue_new (hexvue_hex_volume_t);
  CHECK_MALLOC(h);
  FOR(j, 8) {
    h->points[j].loc    = hex_buf->points[j].loc;
    h->points[j].values = hexvue_new_array(n, hexvue_float_t);
    CHECK_MALLOC(h->points[j].values);
    FOR(i, n) {
      h->points[j].values[i] = hex_buf->points[j].values[i];
    }
  }
  h->parent       = NULL; /* top-level hex */
  h->offspring[0] = 
  h->offspring[1] = 
  h->offspring[2] = 
  h->offspring[3] = 
  h->offspring[4] = 
  h->offspring[5] = 
  h->offspring[6] = 
  h->offspring[7] = NULL;
  h->level        = 0;
  h->grep         = NULL;
  h->plain_grep   = NULL;
  FOR(j, 12) {
    h->edge_greps[j] = NULL;
  }
  return h;
}


static int
read_point(hexvue_model_t *v)
{
  hexvue_isolated_point_t *pt;

  pt = hexvue_new (hexvue_isolated_point_t);
  CHECK_MALLOC(pt);

  GET_LINE({goto error_exit;});
  TOKEN_INT(1, pt->id   , {goto error_exit;});
  TOKEN_DBL(2, pt->loc.x, {goto error_exit;});
  TOKEN_DBL(3, pt->loc.y, {goto error_exit;});
  TOKEN_DBL(4, pt->loc.z, {goto error_exit;});
  pt->grep   = NULL;
  pt->values = NULL;
  
  add_to_tail(v->points, pt);
  return 1;
  
 error_exit:
  return 0;
}



static int
read_hex (hexvue_model_t *v)
{
  hexvue_hex_volume_t *h;
  
  h = hvupriv_read_hex (v->data.num_of_values);
  if (h == NULL) goto error_exit;
  add_to_tail (v->hex_volumes, h);
  return 1;
  
 error_exit:
  return 0;
}



hexvue_hex_volume_t *
hvupriv_read_hex_binary (int num_of_values);

static int
read_hex_binary (hexvue_model_t *v)
{
  hexvue_hex_volume_t *h;
  
  h = hvupriv_read_hex_binary (v->data.num_of_values);
  if (h == NULL) goto error_exit;
  add_to_tail(v->hex_volumes, h);
  return 1;
  
 error_exit:
  return 0;
}



hexvue_hex_volume_t *
hvupriv_read_hex_compressed (int num_of_values);

static int
read_hex_compressed (hexvue_model_t *v)
{
  hexvue_hex_volume_t *h;
  
  h = hvupriv_read_hex_compressed (v->data.num_of_values);
  if (h == NULL) goto error_exit;
  add_to_tail(v->hex_volumes, h);
  return 1;
  
 error_exit:
  return 0;
}


hexvue_hex_volume_t *
hvupriv_read_hex (int num_of_values)
{
  int j, i, n = num_of_values;

  if (hex_buf == NULL) 
    construct_hex_buf (num_of_values);
  
  FOR(j, 8) {
    GET_LINE({goto error_exit;});
    TOKEN_DBL(1, hex_buf->points[j].loc.x, {goto error_exit;});
    TOKEN_DBL(2, hex_buf->points[j].loc.y, {goto error_exit;});
    TOKEN_DBL(3, hex_buf->points[j].loc.z, {goto error_exit;});
    FOR(i, n) {
      GET_LINE({goto error_exit;});
      TOKEN_DBL(1, hex_buf->points[j].values[i], {goto error_exit;});
    }
  }

  update_tolerance(hex_buf);
  return copy_hex_buf (num_of_values);
  
 error_exit:
  return NULL;
}


#define RUN_IOP(action)                                         \
   {                                                            \
     if (!action) {                                             \
       REPORT_ERROR ((char *)ckit_io_error_explanation ());     \
     }                                                          \
   }

hexvue_hex_volume_t *
hvupriv_read_hex_binary (int num_of_values)
{
  int j, i, n = num_of_values;
  int reqlen = n;
  ckit_iobuf_t IOB;
  float geom[3];
  static int buflen        = 0;
  static float *minv       = NULL;
  static float *maxv       = NULL;
  static unsigned char *lo = NULL;
  static unsigned char *hi = NULL;

  if (buflen < reqlen) {
    minv   = hexvue_realloc_array (minv, reqlen, float);
    maxv   = hexvue_realloc_array (maxv, reqlen, float);
    lo     = hexvue_realloc_array (lo, reqlen, unsigned char);
    hi     = hexvue_realloc_array (hi, reqlen, unsigned char);
    buflen = reqlen;
  }

  if (hex_buf == NULL) 
    construct_hex_buf (num_of_values);

  IOB = tokens_get_iobuf ();
  RUN_IOP (ckit_io_scan_data (IOB, FALSE, minv, n * sizeof (float)));
  RUN_IOP (ckit_io_scan_data (IOB, FALSE, maxv, n * sizeof (float)));
  FOR(j, 8) {
    RUN_IOP (ckit_io_scan_data (IOB, FALSE, geom, 3 * sizeof (float)));
    hex_buf->points[j].loc.x = geom[0];
    hex_buf->points[j].loc.y = geom[1];
    hex_buf->points[j].loc.z = geom[2];
    RUN_IOP (ckit_io_scan_data (IOB, FALSE, lo, n * sizeof (unsigned char)));
    RUN_IOP (ckit_io_scan_data (IOB, FALSE, hi, n * sizeof (unsigned char)));
    FOR(i, n) {
      hex_buf->points[j].values[i] =
        ckit_io_convert_bytes_to_double (lo[i], hi[i], minv[i], maxv[i]);
    }
  }

  update_tolerance (hex_buf);
  return copy_hex_buf (num_of_values);
  
 error_exit:
  return NULL;
}



hexvue_hex_volume_t *
hvupriv_read_hex_compressed (int num_of_values)
{
  int j, i, n = num_of_values, k;
  int reqlen = n;
  ckit_iobuf_t IOB;
  float flbox[6];
  unsigned char geombytes[24];
  static int buflen        = 0;
  static float *minv       = NULL;
  static float *maxv       = NULL;
  static unsigned char *byte = NULL;

  if (buflen < reqlen) {
    minv   = hexvue_realloc_array (minv, reqlen, float);
    maxv   = hexvue_realloc_array (maxv, reqlen, float);
    byte   = hexvue_realloc_array (byte, reqlen, unsigned char);
    buflen = reqlen;
  }

  if (hex_buf == NULL) 
    construct_hex_buf (num_of_values);

  IOB = tokens_get_iobuf ();
  RUN_IOP (ckit_io_scan_data (IOB, FALSE, minv, n * sizeof (float)));
  RUN_IOP (ckit_io_scan_data (IOB, FALSE, maxv, n * sizeof (float)));
  RUN_IOP (ckit_io_scan_data (IOB, FALSE, flbox, 6 * sizeof (float)));
  RUN_IOP (ckit_io_scan_data (IOB, FALSE, geombytes, 24));
  k = 0;
  FOR (j, 8) {
    hex_buf->points[j].loc.x
      = ckit_io_convert_byte_to_double (geombytes[k], flbox[0], flbox[3]); k++;
    hex_buf->points[j].loc.y
      = ckit_io_convert_byte_to_double (geombytes[k], flbox[1], flbox[4]); k++;
    hex_buf->points[j].loc.z
      = ckit_io_convert_byte_to_double (geombytes[k], flbox[2], flbox[5]); k++;
    RUN_IOP (ckit_io_scan_data (IOB, FALSE, byte, n * sizeof (unsigned char)));
    FOR(i, n) {
      hex_buf->points[j].values[i] =
        ckit_io_convert_byte_to_double (byte[i], minv[i], maxv[i]);
    }
  }

  update_tolerance (hex_buf);
  return copy_hex_buf (num_of_values);
  
 error_exit:
  return NULL;
}


static void
construct_quad4_buf(hexvue_model_t *v);

static hexvue_quad4_patch_t *quad4_buf= NULL;

static void
construct_quad4_buf(hexvue_model_t *v)
{
  int j, n = v->data.num_of_values;
  
  quad4_buf = hexvue_new (hexvue_quad4_patch_t);
  CHECK_MALLOC(quad4_buf);
  FOR(j, 8) {
    quad4_buf->points[j].values = hexvue_new_array(n, hexvue_float_t);
  }
  quad4_buf->parent       = NULL; /* top-level */
  quad4_buf->offspring[0] = 
  quad4_buf->offspring[1] = 
  quad4_buf->offspring[2] = 
  quad4_buf->offspring[3] = NULL;
  quad4_buf->level        = 0;
  quad4_buf->contour_grep = NULL;
  quad4_buf->grep         = NULL;
}


static hexvue_quad4_patch_t *
copy_quad4_buf(hexvue_model_t *v);


static hexvue_quad4_patch_t *
copy_quad4_buf(hexvue_model_t *v)
{
  hexvue_quad4_patch_t *q;
  int j, i, n = v->data.num_of_values;

  q = hexvue_new (hexvue_quad4_patch_t);
  CHECK_MALLOC(q);
  FOR(j, 4) {
    q->points[j].loc    = quad4_buf->points[j].loc;
    q->points[j].values = hexvue_new_array(n, hexvue_float_t);
    CHECK_MALLOC(q->points[j].values);
    FOR(i, n) {
      q->points[j].values[i] = quad4_buf->points[j].values[i];
    }
  }
  q->parent       = NULL; /* top-level hex */
  q->offspring[0] = 
  q->offspring[1] = 
  q->offspring[2] = 
  q->offspring[3] = NULL;
  q->level        = 0;
  q->grep         = NULL;
  q->contour_grep = NULL;
  
  return q;
}


static int
read_quad4_patch(hexvue_model_t *v)
{
  int j, i, n = v->data.num_of_values;

  if (quad4_buf == NULL) 
    construct_quad4_buf(v);
  
  FOR(j, 4) {
    GET_LINE({goto error_exit;});
    TOKEN_DBL(1, quad4_buf->points[j].loc.x, {goto error_exit;});
    TOKEN_DBL(2, quad4_buf->points[j].loc.y, {goto error_exit;});
    TOKEN_DBL(3, quad4_buf->points[j].loc.z, {goto error_exit;});
    FOR(i, n) {
      GET_LINE({goto error_exit;});
      TOKEN_DBL(1, quad4_buf->points[j].values[i], {goto error_exit;});
    }
  }

  add_to_tail(v->quad4_patches, copy_quad4_buf(v));
  return 1;
  
 error_exit:
  return 0;
}





static void
update_tolerance(hexvue_hex_volume_t *h)
{
  int i;
  double dist;

  for(i = 1; i < 8; i++) {
    dist = 0.01 * v3_distance(&h->points[0].loc, &h->points[i].loc);
    if (dist > 0) {
      tolerance = min(tolerance, dist);
    }
  }
}




static int
draw_hex(hexvue_hex_volume_t *h, hexvue_model_t *v);
static int
draw_quad4(hexvue_quad4_patch_t *h, hexvue_model_t *v);
static int
draw_point(hexvue_isolated_point_t *pt, hexvue_model_t *v);


int
hexvue_generate_graphics(hexvue_model_t *v, EModel *mod)
{
  hexvue_hex_volume_t *h;
  hexvue_quad4_patch_t *q4;
  hexvue_cut_t *c;
  hexvue_contour_t *ct;
  hexvue_crack_surface_t *cs;
  hexvue_isolated_point_t *pt;

  v->model = mod;
  if (v->hex_volumes) {
    loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
      draw_hex(h, v);
    }
  }
  if (v->points) {
    loop_through_list(v->points, pt, hexvue_isolated_point_t *) {
      draw_point(pt, v);
    }
  }
  if (v->crack_surfaces) {
    loop_through_list(v->crack_surfaces, cs, hexvue_crack_surface_t *) {
      draw_crack_surface(cs, v);
    }
  }
  if (v->cuts) {
    loop_through_list(v->cuts, c, hexvue_cut_t *) {
      draw_cut(c, v);
    }
  }
  if (v->contours) {
    loop_through_list(v->contours, ct, hexvue_contour_t *) {
      draw_contour(ct, v);
    }
  }
  if (v->quad4_patches) {
    loop_through_list(v->quad4_patches, q4, hexvue_quad4_patch_t *) {
      draw_quad4(q4, v);
    }
  }
  return 1;
}


static int
draw_plain_hex(hexvue_hex_volume_t *h, hexvue_model_t *v);
static int
draw_outline_hex(hexvue_hex_volume_t *h, hexvue_model_t *v);
static int
update_outline(hexvue_hex_volume_t *h, hexvue_model_t *v);


int
hexvue_generate_outline_hex_rep(hexvue_model_t *v, EModel *mod)
{
  hexvue_hex_volume_t *h;

  v->model = mod;
  if (v->hex_volumes) {
    loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
      update_outline(h, v);
    }
    loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
      draw_outline_hex(h, v);
    }
  }

  return 1;
}


int
hexvue_generate_plain_hex_rep(hexvue_model_t *v, EModel *mod)
{
  hexvue_hex_volume_t *h;

  v->model = mod;
  if (v->hex_volumes) {
    loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
      draw_plain_hex(h, v);
    }
  }

  return 1;
}

int
hexvue_regenerate_graphics(hexvue_model_t *v)
{
  if (v->model == NULL)
    return 0;
  hexvue_generate_graphics(v, v->model);
  return 1;
}


static int
draw_hex(hexvue_hex_volume_t *h, hexvue_model_t *v)
{
  WCRec pnts[8];
  FPNum vals[8];
  int i, layer;

  if (h->level == 0) {
    FOR(i, 8) {
      pnts[i].x = h->points[i].loc.x;
      if (v->on_x != NO_ACTIVE_VALUE)
        pnts[i].x += v->scale_x * h->points[i].values[v->on_x];
      
      pnts[i].y = h->points[i].loc.y;
      if (v->on_y != NO_ACTIVE_VALUE)
        pnts[i].y += v->scale_y * h->points[i].values[v->on_y];
      
      pnts[i].z = h->points[i].loc.z;
      if (v->on_z != NO_ACTIVE_VALUE)
        pnts[i].z += v->scale_z * h->points[i].values[v->on_z];
      
      vals[i] = 0;
      if (v->active_value != NO_ACTIVE_VALUE)
        vals[i] = h->points[i].values[v->active_value];
    }
    
    if (h->grep == NULL) {
      h->grep = CreateHexahedronWD(pnts, vals);
      layer = EASValsGetLayer();
      EASValsSetLayer(HEX_VOL_LAYER);
      EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, h->grep);
      EASValsSetLayer(layer);
      EMNoDrawAddGraphicsToModel(v->model, h->grep);
    } else {
      EGModifyGraphicsGeometry(h->grep, pnts);
      EGModifyAssocData(h->grep, vals);
    }
  }
  
  return 1;
}

static int
draw_plain_hex(hexvue_hex_volume_t *h, hexvue_model_t *v)
{
  WCRec pnts[8];
  int i, layer;

  if (h->grep == NULL) draw_hex(h, v);

  EGGetGraphicsGeometry(h->grep, pnts, &i);
  h->plain_grep = CreateHexahedron(pnts);
  layer = EASValsGetLayer();
  EASValsSetLayer(HEX_PLAIN_VOL_LAYER);
  EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, h->plain_grep);
  EASValsSetLayer(layer);
  EMNoDrawAddGraphicsToModel(v->model, h->plain_grep);
  return TRUE;
}


int
is_outline_edge(hexvue_point_t *p0, hexvue_point_t *p1);
int
add_outline_edge(hexvue_point_t *p0, hexvue_point_t *p1, double tol);


static int
update_outline(hexvue_hex_volume_t *h, hexvue_model_t *v)
{
  int i;

  FOR(i, 12) {
    int k, l;
    k = hexvue_edge_vertex_indices[i][0];
    l = hexvue_edge_vertex_indices[i][1];
    add_outline_edge(&h->points[k], &h->points[l], tolerance);
  }
  return TRUE;
}



static int
draw_outline_hex(hexvue_hex_volume_t *h, hexvue_model_t *v)
{
  WCRec pnts[8];
  int i, layer;

  if (h->grep == NULL) draw_hex(h, v);

  EGGetGraphicsGeometry(h->grep, pnts, &i);
  FOR(i, 12) {
    WCRec pts[2];
    int k, l;
    k = hexvue_edge_vertex_indices[i][0];
    l = hexvue_edge_vertex_indices[i][1];
    if (is_outline_edge(&h->points[k], &h->points[l])) {
      pts[0] = pnts[k];
      pts[1] = pnts[l];
      h->edge_greps[i] = CreateLine3D(pts);
      layer = EASValsGetLayer();
      EASValsSetLayer(HEX_OUTLINE_VOL_LAYER);
      EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, h->edge_greps[i]);
      EASValsSetLayer(layer);
      EMNoDrawAddGraphicsToModel(v->model, h->edge_greps[i]);
    }
  }
  return TRUE;
}


static void
destroy_hex_plain_grep(hexvue_hex_volume_t *h, hexvue_model_t *v);
static void
destroy_hex_outline_grep(hexvue_hex_volume_t *h, hexvue_model_t *v);


static void
destroy_hex_plain_grep(hexvue_hex_volume_t *h, hexvue_model_t *v)
{
  if (h->plain_grep)
    EMDeleteGraphics(v->model, h->plain_grep);
  h->plain_grep = NULL;
}

static void
destroy_hex_outline_grep(hexvue_hex_volume_t *h, hexvue_model_t *v)
{
  int i;

  FOR(i, 12) {
    if (h->edge_greps[i]) {
      EMDeleteGraphics(v->model, h->edge_greps[i]);
    }
    h->edge_greps[i] = NULL;
  }
}


static int
draw_point(hexvue_isolated_point_t *pt, hexvue_model_t *v)
{
  WCRec pnts[1];
  int layer;
  char buf[64];
  Font fnt;

  pnts[0].x = pt->loc.x;
  pnts[0].y = pt->loc.y; 
  pnts[0].z = pt->loc.z;
  if (pt->grep == NULL) {
    sprintf(buf, "%d", pt->id);
    pt->grep = CreateAnnText3D(pnts, buf);
    layer = EASValsGetLayer();
    fnt   = EASValsGetFontId();
    EASValsSetLayer(POINT_LAYER);
    EASValsSetFontId(FontDefaultFont());
    EGWithMaskChangeAttributes(COLOR_MASK | LAYER_MASK | FONT_MASK, pt->grep);
    EASValsSetLayer(layer);
    EASValsSetFontId(fnt);
    EMNoDrawAddGraphicsToModel(v->model, pt->grep);
  } else {
    EGModifyGraphicsGeometry(pt->grep, pnts);
  }
   
  return 1;
}



static int
draw_quad4(hexvue_quad4_patch_t *q4, hexvue_model_t *v)
{
  WCRec pnts[4];
  FPNum vals[4];
  int i, layer;

  if (q4->level == 0) {
    FOR(i, 4) {
      pnts[i].x = q4->points[i].loc.x;
      if (v->on_x != NO_ACTIVE_VALUE)
        pnts[i].x += v->scale_x * q4->points[i].values[v->on_x];
      
      pnts[i].y = q4->points[i].loc.y;
      if (v->on_y != NO_ACTIVE_VALUE)
        pnts[i].y += v->scale_y * q4->points[i].values[v->on_y];
      
      pnts[i].z = q4->points[i].loc.z;
      if (v->on_z != NO_ACTIVE_VALUE)
        pnts[i].z += v->scale_z * q4->points[i].values[v->on_z];
      
      vals[i] = 0;
      if (v->active_value != NO_ACTIVE_VALUE)
        vals[i] = q4->points[i].values[v->active_value];
    }
    
    if (q4->contour_grep == NULL) {
      q4->contour_grep = CreateQuadWD3D(pnts,
                                        vals[0], vals[1], vals[2], vals[3]);
      layer = EASValsGetLayer();
      EASValsSetLayer(QUAD4_CONTOUR_LAYER);
      EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, q4->contour_grep);
      EASValsSetLayer(layer);
      EMNoDrawAddGraphicsToModel(v->model, q4->contour_grep);
    } else {
      EGModifyGraphicsGeometry(q4->contour_grep, pnts);
      EGModifyAssocData(q4->contour_grep, vals);
    }
    if (q4->grep == NULL) {
      q4->grep = CreateQuad3D(pnts);
      layer = EASValsGetLayer();
      EASValsSetLayer(QUAD4_PLAIN_LAYER);
      EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, q4->grep);
      EASValsSetLayer(layer);
      EMNoDrawAddGraphicsToModel(v->model, q4->grep);
    } else {
      EGModifyGraphicsGeometry(q4->grep, pnts);
    }
    
  }
  
  return 1;
}



static void
destroy_point(hexvue_isolated_point_t *pt, hexvue_model_t *v);
static void
destroy_cut(hexvue_cut_t *c, hexvue_model_t *v);
static void
destroy_contour(hexvue_contour_t *ct, hexvue_model_t *v);
static void
destroy_quad4_patch(hexvue_quad4_patch_t *q4, hexvue_model_t *v);


void
hexvue_free_model (hexvue_model_t *v)
{
  int i;
  hexvue_hex_volume_t *h;
  hexvue_quad4_patch_t *q4;
  hexvue_cut_t *c;
  hexvue_contour_t *ct;
  hexvue_isolated_point_t *pt;

  if (front_lobs != NULL) {
    lobs_3d_free_instance (front_lobs);
    front_lobs = NULL;
  }
  if (crack_lobs != NULL) {
    lobs_3d_free_instance (crack_lobs);
    crack_lobs = NULL;
  }
  
  if (v->model != NULL) {
    hexvue_destroy_graphics(v);
  }
  FOR(i, v->data.num_of_values) {
    hexvue_free(v->data.values[i].description);
  }
  hexvue_free(v->data.values);
  if (v->hex_volumes) {
    loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
      hvupriv_destroy_hex_volume(h, v);
    }
    free_list(v->hex_volumes, NOT_NODES);
  }
  if (v->points) {
    loop_through_list(v->points, pt, hexvue_isolated_point_t *) {
      destroy_point(pt, v);
    }
    free_list(v->points, NOT_NODES);
  }
  if (v->quad4_patches) {
    loop_through_list(v->quad4_patches, q4, hexvue_quad4_patch_t *) {
      destroy_quad4_patch(q4, v);
    }
    free_list(v->quad4_patches, NOT_NODES);
  }
  if (v->cuts) {
    loop_through_list(v->cuts, c, hexvue_cut_t *) {
      destroy_cut(c, v);
    }
    free_list(v->cuts, NOT_NODES);
  }
  if (v->contours) {
    loop_through_list(v->contours, ct, hexvue_contour_t *) {
      destroy_contour(ct, v);
    }
    free_list(v->contours, NOT_NODES);
  }
  hexvue_free(v->comment);
  hexvue_free(v->in_file);
  hexvue_free(v);
}


static void
destroy_hex_grep(hexvue_hex_volume_t *h, hexvue_model_t *v);


void
hvupriv_destroy_hex_volume (hexvue_hex_volume_t *h, hexvue_model_t *v)
{
  int i;

  if (HEX_HAS_OFFSPRING (h)) {
    FOR (i, 8) {
      hvupriv_destroy_hex_volume (h->offspring[i], v);
    }
  } 
  FOR (i, 8) {
    hexvue_free (h->points[i].values);
  }
  if (h->grep != NULL) destroy_hex_grep (h, v);
  if (h->plain_grep != NULL) destroy_hex_plain_grep (h, v);
  destroy_hex_outline_grep (h, v);
  hexvue_free(h);
}

static void
destroy_point_grep(hexvue_isolated_point_t *pt, hexvue_model_t *v);

static void
destroy_point(hexvue_isolated_point_t *pt, hexvue_model_t *v)
{
  if (pt->grep != NULL) destroy_point_grep (pt, v);
  hexvue_free (pt);
}



static void
destroy_quad4_grep(hexvue_quad4_patch_t *q4, hexvue_model_t *v);


static void
destroy_quad4_patch(hexvue_quad4_patch_t *q4, hexvue_model_t *v)
{
  int i;

  if (Q4_HAS_OFFSPRING(q4)) {
    FOR(i, 4) {
      destroy_quad4_patch(q4->offspring[i], v);
    }
  } 
  FOR(i, 4) {
    hexvue_free(q4->points[i].values);
  }
  if (q4->grep != NULL) destroy_quad4_grep(q4, v);
  hexvue_free(q4);
}

static void
destroy_poly(hexvue_cut_poly_t *p, hexvue_model_t *v);


static void
destroy_cut(hexvue_cut_t *c, hexvue_model_t *v)
{
  hexvue_cut_poly_t *p;
  
  if (c->polys) {
    loop_through_list(c->polys, p, hexvue_cut_poly_t *) {
      destroy_poly(p, v);
    }
  }
  hexvue_free(c);
}



static void
destroy_contour(hexvue_contour_t *ct, hexvue_model_t *v)
{
  hexvue_cut_poly_t *p;
  
  if (ct->polys) {
    loop_through_list(ct->polys, p, hexvue_cut_poly_t *) {
      destroy_poly(p, v);
    }
  }
  hexvue_free(ct);
}


static void
destroy_poly_grep(hexvue_cut_poly_t *p, hexvue_model_t *v);



static void
destroy_poly(hexvue_cut_poly_t *p, hexvue_model_t *v)
{
  if (p->grep) destroy_poly_grep(p, v);
  hexvue_free(p);
}





static void
destroy_cut_grep(hexvue_cut_t *c, hexvue_model_t *v);
static void
destroy_contour_grep(hexvue_contour_t *ct, hexvue_model_t *v);


void
hexvue_destroy_graphics(hexvue_model_t *v)
{
  hexvue_hex_volume_t *h;
  hexvue_cut_t *c;
  hexvue_contour_t *ct;
  hexvue_quad4_patch_t *q4;
  hexvue_isolated_point_t *pt;
  
  if (v->hex_volumes) {
    loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
      destroy_hex_grep(h, v);
      destroy_hex_plain_grep(h, v);
      destroy_hex_outline_grep(h, v);
    }
  }
  if (v->points) {
    loop_through_list(v->points, pt, hexvue_isolated_point_t *) {
      destroy_point_grep(pt, v);
    }
  }
  if (v->cuts) {
    loop_through_list(v->cuts, c, hexvue_cut_t *) {
      destroy_cut_grep(c, v);
    }
  }
  if (v->contours) {
    loop_through_list(v->contours, ct, hexvue_contour_t *) {
      destroy_contour_grep(ct, v);
    }
  }
  if (v->quad4_patches) {
    loop_through_list(v->quad4_patches, q4, hexvue_quad4_patch_t *) {
      destroy_quad4_grep(q4, v);
    }
  }
  v->model = NULL;
}



void
hexvue_destroy_plain_hex_rep(hexvue_model_t *v)
{
  hexvue_hex_volume_t *h;
  
  if (v->hex_volumes) {
    loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
      destroy_hex_plain_grep(h, v);
    }
  }
}


void
hexvue_destroy_outline_hex_rep(hexvue_model_t *v)
{
  hexvue_hex_volume_t *h;
  
  if (v->hex_volumes) {
    loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
      destroy_hex_outline_grep(h, v);
    }
  }
}


static void
destroy_hex_grep(hexvue_hex_volume_t *h, hexvue_model_t *v)
{
  EMDeleteGraphics(v->model, h->grep);
  h->grep = NULL;
}


static void
destroy_quad4_grep(hexvue_quad4_patch_t *q4, hexvue_model_t *v)
{
  EMDeleteGraphics(v->model, q4->grep);
  EMDeleteGraphics(v->model, q4->contour_grep);
  q4->grep         = NULL;
  q4->contour_grep = NULL;
}

static void
destroy_point_grep(hexvue_isolated_point_t *pt, hexvue_model_t *v)
{
  EMDeleteGraphics(v->model, pt->grep);
  pt->grep         = NULL;
}

static void
destroy_cut_grep(hexvue_cut_t *c, hexvue_model_t *v)
{
  hexvue_cut_poly_t *p;
  
  if (c->polys) {
    loop_through_list(c->polys, p, hexvue_cut_poly_t *) {
      destroy_poly_grep(p, v);
    }
  }
}



static void
destroy_contour_grep(hexvue_contour_t *ct, hexvue_model_t *v)
{
  hexvue_cut_poly_t *p;
  
  if (ct->polys) {
    loop_through_list(ct->polys, p, hexvue_cut_poly_t *) {
      destroy_poly_grep(p, v);
    }
  }
}


static void
destroy_poly_grep(hexvue_cut_poly_t *p, hexvue_model_t *v)
{
  if (p->grep) {
    EMDeleteGraphics(v->model, p->grep);
  }
  p->grep = NULL;
}



static int
hex_update_minmax(hexvue_hex_volume_t *h, int set, double *minv, double *maxv);
static int
quad4_update_minmax(hexvue_quad4_patch_t *q4, int set,
                    double *minv, double *maxv);
static int
crack_surf_update_minmax(hexvue_crack_surface_t *cs,
                         int set, double *minv, double *maxv);


static int
crack_surf_update_minmax(hexvue_crack_surface_t *cs,
                         int set, double *minv, double *maxv)
{
  int i;
  double val;
  hexvue_crack_poly_t *csp;

  if (cs->crack_polygons) {
    loop_through_list(cs->crack_polygons, csp, hexvue_crack_poly_t *) {
      FOR(i, csp->npts) {
        val = csp->points[i].values_on_plus_side[set];
        if (val < *minv) *minv = val;
        if (val > *maxv) *maxv = val;
        val = csp->points[i].values_on_minus_side[set];
        if (val < *minv) *minv = val;
        if (val > *maxv) *maxv = val;
      }
    }
  }

  return 1;
}


int
hexvue_find_minmax_value (hexvue_model_t *v, int set,
                          double *minv, double *maxv)
{
  if (set < 0 || set >= v->data.num_of_values)
    return 0;
  else {
    hexvue_hex_volume_t *h;
    hexvue_quad4_patch_t *q4;
    hexvue_crack_surface_t *cs;

    *minv = FLT_MAX;
    *maxv = -FLT_MAX;
    loop_through_list (v->hex_volumes, h, hexvue_hex_volume_t *) {
      hex_update_minmax (h, set, minv, maxv);
    }

    loop_through_list (v->quad4_patches, q4, hexvue_quad4_patch_t *) {
      quad4_update_minmax (q4, set, minv, maxv);
    }

    loop_through_list (v->crack_surfaces, cs, hexvue_crack_surface_t *) {
      crack_surf_update_minmax (cs, set, minv, maxv);
    }
    
    return 1;
  }
}


static int
hex_update_minmax (hexvue_hex_volume_t *h, int set, double *minv, double *maxv)
{
  int i;
  double val;

  FOR(i, 8) {
    val = h->points[i].values[set];
    if (val < *minv) *minv = val;
    if (val > *maxv) *maxv = val;
  }

  return 1;
}

static int
quad4_update_minmax(hexvue_quad4_patch_t *q4,
                    int set, double *minv, double *maxv)
{
  int i;
  double val;

  FOR(i, 4) {
    val = q4->points[i].values[set];
    if (val < *minv) *minv = val;
    if (val > *maxv) *maxv = val;
  }

  return 1;
}


void
hexvue_print_header(hexvue_model_t *v, FILE *fp)
{
  int i;
  double minv, maxv;

  fprintf(fp,
          "===============================================================\n");
  fprintf(fp, "HEXVUE model from %s\n", v->in_file);
  fprintf(fp, "   (%s)\n", v->comment);
  fprintf(fp,
          "===============================================================\n");
  fprintf(fp, "Data:\n");
  FOR(i, v->data.num_of_values) {
    if (!hexvue_find_minmax_value(v, i, &minv, &maxv))
      fprintf(fp, "\t%d = %s (no min/max values!?)\n",
              i, v->data.values[i].description);
    else
      fprintf(fp, "\t%d = %s (min %g, max %g)\n",
              i, v->data.values[i].description,
              minv, maxv);
  }
  fprintf(fp,
          "===============================================================\n");
}




int
hexvue_pick_set(hexvue_model_t *v, int set)
{
  if (set < 0 || set >= v->data.num_of_values)
    return 0;
  else {
    double minv, maxv;
    
    v->active_value = set;
    hexvue_find_minmax_value(v, set, &minv, &maxv);
    ColorSetupFringeTableByMinMax(EMGetAssocFringeTable(v->model),
                                  minv, maxv);
  }
  return 1;
}


int
hexvue_pick_displ(hexvue_model_t *v, int axis, int set, double scale)
{
  if (set < NO_ACTIVE_VALUE || set >= v->data.num_of_values)
    return 0;
  else {
    if        (axis == 'x' || axis == 'X') {
      if ((v->scale_x = scale) != 0) v->on_x    = set;
      else                 v->on_x    = -1;
    } else if (axis == 'y' || axis == 'Y') {
      if ((v->scale_y = scale) != 0) v->on_y    = set;
      else                 v->on_y    = -1;
    } else if (axis == 'z' || axis == 'Z') {
      if ((v->scale_z = scale) != 0) v->on_z    = set;
      else                 v->on_z    = -1;
    } else {
      return 0;
    }
  }
  return 1;
}


static hexvue_cut_t *
new_cut(hexvue_model_t *v,
                    hexvue_vect_t *normal, hexvue_vect_t *center);



hexvue_cut_t *
hexvue_generate_cut(hexvue_model_t *v,
                    hexvue_vect_t *normal, hexvue_vect_t *center)
{
  hexvue_cut_t *c;
  
  if (v->cuts == NULL) v->cuts = make_list();
  c = new_cut (v, normal, center);
  add_to_tail(v->cuts, c);

  return c;
}





static void
add_to_cut(hexvue_cut_t *c, hexvue_model_t *v, hexvue_hex_volume_t *h);



static hexvue_cut_t *
new_cut(hexvue_model_t *v,
                    hexvue_vect_t *normal, hexvue_vect_t *center)
{
  hexvue_cut_t *c;

  c = hexvue_new(hexvue_cut_t);
  CHECK_MALLOC(c);
  c->normal = *normal;
  c->center = *center;
  plane_consts_normal(&c->plane, &c->normal, &c->center);
  c->polys  = NULL;
  c->layer  = EASValsGetLayer();
  {
    hexvue_hex_volume_t *h;
    
    loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
      add_to_cut(c, v, h);
    }
  }

  return c;
}



int
cut_hex(hexvue_hex_volume_t *h, /* IN */
        hexvue_plane_t *plane, /* IN */
        hexvue_cut_point_t *cut_points,
        int *ncutpts   /* OUT # generated points */
        );


static void
add_to_cut(hexvue_cut_t *c, hexvue_model_t *v, hexvue_hex_volume_t *h)
{
  hexvue_cut_point_t  cut_points[12];
  int ncutpts, i;
  hexvue_cut_poly_t *p;
  
#define NEW_POLY()                              \
  {                                             \
    p = hexvue_new(hexvue_cut_poly_t);          \
    CHECK_MALLOC(p);                            \
    p->hex  = h;                                \
    p->grep = NULL;                             \
  }

  if (c->polys == NULL) c->polys = make_list();
  
  cut_hex(h, &c->plane, cut_points, &ncutpts);
  if (ncutpts > 0) {
    switch (ncutpts) {
    case 3:
      NEW_POLY();
      p->npts = 3;
      FOR(i, 3) {
        p->pts[i] = cut_points[i];
      }
      add_to_tail(c->polys, p);
      break;
    case 4:
      NEW_POLY();
      p->npts = 4;
      FOR(i, 4) {
        p->pts[i] = cut_points[i];
      }
      add_to_tail(c->polys, p);
      break;
    case 5:
      NEW_POLY();
      p->npts = 3;
      FOR(i, 3) {
        p->pts[i] = cut_points[i];
      }
      add_to_tail(c->polys, p);
      NEW_POLY();
      p->npts = 4;
      p->pts[0] = cut_points[2];
      p->pts[1] = cut_points[3];
      p->pts[2] = cut_points[4];
      p->pts[3] = cut_points[0];
      add_to_tail(c->polys, p);
      break;
    case 6:
      NEW_POLY();
      p->npts = 4;
      p->pts[0] = cut_points[0];
      p->pts[1] = cut_points[1];
      p->pts[2] = cut_points[2];
      p->pts[3] = cut_points[5];
      add_to_tail(c->polys, p);
      NEW_POLY();
      p->npts = 4;
      p->pts[0] = cut_points[2];
      p->pts[1] = cut_points[3];
      p->pts[2] = cut_points[4];
      p->pts[3] = cut_points[5];
      add_to_tail(c->polys, p);
      break;
    default:
      /* RAW need to try to reduce the number of cut points for degenerate */
      /* cases (cuts through vertices) */
      REPORT_ERROR("bad number of cut points");
      break;
    }
  }
}



static int
draw_cut(hexvue_cut_t *c, hexvue_model_t *v)
{
  hexvue_cut_poly_t *p;
  
  if (c->polys) {
    loop_through_list(c->polys, p, hexvue_cut_poly_t *) {
      draw_poly(p, c, v);
    }
  }
  return 1;
}


static int
draw_crack_poly(hexvue_crack_poly_t *csp, hexvue_crack_surface_t *c,
                hexvue_model_t *v);


static int
draw_crack_surface(hexvue_crack_surface_t *c, hexvue_model_t *v)
{
  hexvue_crack_poly_t *csp;

  if (c->crack_polygons) {
    loop_through_list(c->crack_polygons, csp, hexvue_crack_poly_t *) {
      draw_crack_poly(csp, c, v);
    }
  }
  return 1;
}


static int
draw_crack_poly(hexvue_crack_poly_t *csp, hexvue_crack_surface_t *c,
                hexvue_model_t *v)
{
  WCRec pnts[4];
  int i;
  int prev_layer;
  FPNum vals[4];

  prev_layer = EASValsGetLayer();
  FOR(i, csp->npts) {
    pnts[i].x = csp->points[i].loc.x;
    if (v->on_x != NO_ACTIVE_VALUE)
      pnts[i].x += v->scale_x * csp->points[i].values_on_plus_side[v->on_x];

    pnts[i].y = csp->points[i].loc.y;
    if (v->on_y != NO_ACTIVE_VALUE)
      pnts[i].y += v->scale_y * csp->points[i].values_on_plus_side[v->on_y];

    pnts[i].z = csp->points[i].loc.z;
    if (v->on_z != NO_ACTIVE_VALUE)
      pnts[i].z += v->scale_z * csp->points[i].values_on_plus_side[v->on_z];
    vals[i] = 0;
    if (v->active_value != NO_ACTIVE_VALUE)
      vals[i] = csp->points[i].values_on_plus_side[v->active_value];
  }

  if (csp->plus_side_grep == NULL) {
    switch (csp->npts) {
    case 3:
      csp->plus_side_grep
        = CreateTriangleWD3D(pnts, vals[0], vals[1], vals[2]);
      csp->plus_side_plain_grep
        = CreateTriangle3D(pnts);

      break;
    case 4:
      csp->plus_side_grep
        = CreateQuadWD3D(pnts, vals[0], vals[1], vals[2], vals[3]);
      csp->plus_side_plain_grep
        = CreateQuad3D(pnts);
      break;
    default:
      REPORT_ERROR("bad poly");
      break;
    }
    EASValsSetLayer(CRACK_SURFACE_LAYER);
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, csp->plus_side_grep);
    EMNoDrawAddGraphicsToModel(v->model, csp->plus_side_grep);
    EASValsSetLayer(CRACK_PLAIN_SURFACE_LAYER);
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, csp->plus_side_plain_grep);
    EMNoDrawAddGraphicsToModel(v->model, csp->plus_side_plain_grep);
  } else {
    EGModifyGraphicsGeometry(csp->plus_side_grep, pnts);
    EGModifyAssocData(csp->plus_side_grep, vals);
    EGModifyGraphicsGeometry(csp->plus_side_plain_grep, pnts);
  }

  FOR(i, csp->npts) {
    pnts[i].x = csp->points[i].loc.x;
    if (v->on_x != NO_ACTIVE_VALUE)
      pnts[i].x += v->scale_x * csp->points[i].values_on_minus_side[v->on_x];

    pnts[i].y = csp->points[i].loc.y;
    if (v->on_y != NO_ACTIVE_VALUE)
      pnts[i].y += v->scale_y * csp->points[i].values_on_minus_side[v->on_y];

    pnts[i].z = csp->points[i].loc.z;
    if (v->on_z != NO_ACTIVE_VALUE)
      pnts[i].z += v->scale_z * csp->points[i].values_on_minus_side[v->on_z];
    vals[i] = 0;
    if (v->active_value != NO_ACTIVE_VALUE)
      vals[i] = csp->points[i].values_on_minus_side[v->active_value];
  }

  if (csp->minus_side_grep == NULL) {
    switch (csp->npts) {
    case 3:
      csp->minus_side_grep
        = CreateTriangleWD3D(pnts, vals[0], vals[1], vals[2]);
      csp->minus_side_plain_grep
        = CreateTriangle3D(pnts);
      break;
    case 4:
      csp->minus_side_grep
        = CreateQuadWD3D(pnts, vals[0], vals[1], vals[2], vals[3]);
      csp->minus_side_plain_grep
        = CreateQuad3D(pnts);
      break;
    default:
      REPORT_ERROR("bad poly");
      break;
    }
    EASValsSetLayer(CRACK_SURFACE_LAYER);
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, csp->minus_side_grep);
    EMNoDrawAddGraphicsToModel(v->model, csp->minus_side_grep);
    EASValsSetLayer(CRACK_PLAIN_SURFACE_LAYER);
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, csp->minus_side_plain_grep);
    EMNoDrawAddGraphicsToModel(v->model, csp->minus_side_plain_grep);
  } else {
    EGModifyGraphicsGeometry(csp->minus_side_grep, pnts);
    EGModifyAssocData(csp->minus_side_grep, vals);
    EGModifyGraphicsGeometry(csp->minus_side_plain_grep, pnts);
  }
  EASValsSetLayer(prev_layer);
  
  return 1;
}



static int
draw_poly(hexvue_cut_poly_t *p, hexvue_cut_t *c, hexvue_model_t *v)
{
  WCRec hex_pnts[8];
  FPNum hex_vals[8];
  WCRec pnts[8];
  FPNum vals[8];
  int i;
  hexvue_hex_volume_t *h = p->hex;
  int edge, v1, v2;
  double lambda;
  int prev_layer;

  prev_layer = EASValsGetLayer();
  FOR(i, 8) {
    hex_pnts[i].x = h->points[i].loc.x;
    if (v->on_x != NO_ACTIVE_VALUE)
      hex_pnts[i].x += v->scale_x * h->points[i].values[v->on_x];

    hex_pnts[i].y = h->points[i].loc.y;
    if (v->on_y != NO_ACTIVE_VALUE)
      hex_pnts[i].y += v->scale_y * h->points[i].values[v->on_y];

    hex_pnts[i].z = h->points[i].loc.z;
    if (v->on_z != NO_ACTIVE_VALUE)
      hex_pnts[i].z += v->scale_z * h->points[i].values[v->on_z];

    hex_vals[i] = 0;
    if (v->active_value != NO_ACTIVE_VALUE)
      hex_vals[i] = h->points[i].values[v->active_value];
  }

  FOR(i, p->npts) {
    lambda = p->pts[i].lambda;
    edge   = p->pts[i].on_edge;
    v1     = hexvue_edge_vertex_indices[edge][0];
    v2     = hexvue_edge_vertex_indices[edge][1];
    pnts[i].x = hex_pnts[v1].x + (hex_pnts[v2].x - hex_pnts[v1].x) * lambda;
    pnts[i].y = hex_pnts[v1].y + (hex_pnts[v2].y - hex_pnts[v1].y) * lambda;
    pnts[i].z = hex_pnts[v1].z + (hex_pnts[v2].z - hex_pnts[v1].z) * lambda;
    vals[i]   = hex_vals[v1]   + (hex_vals[v2]   -   hex_vals[v1]) * lambda;
  }
  
  if (p->grep == NULL) {
    switch (p->npts) {
    case 3:
      p->grep = CreateTriangleWD3D(pnts, vals[0], vals[1], vals[2]);
      break;
    case 4:
      p->grep = CreateQuadWD3D(pnts, vals[0], vals[1], vals[2], vals[3]);
      break;
    default:
      REPORT_ERROR("bad poly");
      break;
    }
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, p->grep);
    EMNoDrawAddGraphicsToModel(v->model, p->grep);
  } else {
    EGModifyGraphicsGeometry(p->grep, pnts);
    EGModifyAssocData(p->grep, vals);
  }
  EASValsSetLayer(prev_layer);
  
  return 1;
}


void
hexvue_remove_cuts(hexvue_model_t *v)
{
  hexvue_cut_t *c;
  
  if (v->cuts) {
    loop_through_list(v->cuts, c, hexvue_cut_t *) {
      destroy_cut(c, v);
    }
    free_list(v->cuts, NOT_NODES);
    v->cuts = NULL;
  }
}



void
hexvue_remove_cut(hexvue_model_t *v, hexvue_cut_t *c_to_remove)
{
  hexvue_cut_t *c;
  
  if (v->cuts) {
    c = (hexvue_cut_t *)delete_list_node(v->cuts, c_to_remove);
    if (c == c_to_remove) {
      destroy_cut(c, v);
    }
  }
}




void
hexvue_remove_contours(hexvue_model_t *v)
{
  hexvue_contour_t *ct;
  
  if (v->contours) {
    loop_through_list(v->contours, ct, hexvue_contour_t *) {
      destroy_contour(ct, v);
    }
    free_list(v->contours, NOT_NODES);
    v->contours = NULL;
  }
}




void
hexvue_remove_contour(hexvue_model_t *v, hexvue_contour_t *ct_to_remove)
{
  hexvue_contour_t *ct;
  
  if (v->contours) {
    ct = (hexvue_contour_t *)delete_list_node(v->contours, ct_to_remove);
    if (ct == ct_to_remove) {
      destroy_contour(ct, v);
    }
  }
}


int
hexvue_write_cuts(hexvue_model_t *v, char *name)
{
  FILE *fp;
  hexvue_cut_t *c;

  fp = fopen(name, "w");
  if (fp == NULL)
    return 0;
  
  fprintf(fp, "! HEXVUE cuts file (C) 1995 Petr Krysl\n");
  fprintf(fp, "! %s (file %s). \n",
          v->comment, v->in_file);
  fprintf(fp, "!   There are %d cuts here.\n",
          v->cuts? list_length(v->cuts): 0);
  fprintf(fp, "!   normal x,y,z          center x,y,z          layer \n");

  if (v->cuts) {
    loop_through_list(v->cuts, c, hexvue_cut_t *) {
      fprintf(fp, "%g %g %g   %g %g %g   %d\n",
              c->normal.x, c->normal.y, c->normal.z, 
              c->center.x, c->center.y, c->center.z,
              c->layer);
    }
  }
  fprintf(fp, "! END HEXVUE cut file.\n"); 
  
  fclose(fp);

  return 1;
}


int
hexvue_read_cuts(hexvue_model_t *v, char *name)
{
  hexvue_vect_t cut_normal;
  hexvue_vect_t cut_center;
  int cut_layer;
  
  if (!tokens_open_file(name))
    return 0;

  while(tokens_next_line()) {
    if (tokens_total_of_tokens() != 7)
      return 0;
    cut_normal.x = tokens_token_as_double(1);
    cut_normal.y = tokens_token_as_double(2);
    cut_normal.z = tokens_token_as_double(3);
    cut_center.x = tokens_token_as_double(4);
    cut_center.y = tokens_token_as_double(5);
    cut_center.z = tokens_token_as_double(6);
    cut_layer    = tokens_token_as_int(7);
    EASValsSetLayer(cut_layer);
    hexvue_generate_cut(v, &cut_normal, &cut_center);
  }

  tokens_close_curr_file();
  return 1;
}


static hexvue_contour_t *
new_contour(hexvue_model_t *v, double contour_value);


hexvue_contour_t *
hexvue_generate_contour(hexvue_model_t *v, double contour_value)
{
  hexvue_contour_t *ct;
  
  if (v->contours == NULL) v->contours = make_list();
  ct = new_contour(v, contour_value);
  add_to_tail(v->contours, ct);

  return ct;
}

static int
add_to_contour(hexvue_contour_t *ct,
               hexvue_model_t *v, hexvue_hex_volume_t *h);



static hexvue_contour_t *
new_contour(hexvue_model_t *v, double contour_value)
{
  hexvue_contour_t *ct;
  hexvue_hex_volume_t *h;

  ct = hexvue_new(hexvue_contour_t);
  CHECK_MALLOC(ct);
  ct->contour_value = contour_value;
  ct->polys  = NULL;
  ct->layer  = EASValsGetLayer();

  loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
    while (!add_to_contour(ct, v, h));
  }

  return ct;
}


static int
draw_poly_contour(hexvue_cut_poly_t *p,
                  hexvue_contour_t *ct, hexvue_model_t *v);


static int
draw_contour(hexvue_contour_t *ct, hexvue_model_t *v)
{
  hexvue_cut_poly_t *p;
  
  if (ct->polys) {
    loop_through_list(ct->polys, p, hexvue_cut_poly_t *) {
      draw_poly_contour(p, ct, v);
    }
  }
  return 1;
}



static int
draw_poly_contour(hexvue_cut_poly_t *p,
                  hexvue_contour_t *ct, hexvue_model_t *v)
{
  WCRec hex_pnts[8];
  FPNum hex_vals[8];
  WCRec pnts[8];
  int i;
  hexvue_hex_volume_t *h = p->hex;
  int edge, v1, v2;
  double lambda;

  FOR(i, 8) {
    hex_pnts[i].x = h->points[i].loc.x;
    if (v->on_x != NO_ACTIVE_VALUE)
      hex_pnts[i].x += v->scale_x * h->points[i].values[v->on_x];

    hex_pnts[i].y = h->points[i].loc.y;
    if (v->on_y != NO_ACTIVE_VALUE)
      hex_pnts[i].y += v->scale_y * h->points[i].values[v->on_y];

    hex_pnts[i].z = h->points[i].loc.z;
    if (v->on_z != NO_ACTIVE_VALUE)
      hex_pnts[i].z += v->scale_z * h->points[i].values[v->on_z];

    hex_vals[i] = 0;
    if (v->active_value != NO_ACTIVE_VALUE)
      hex_vals[i] = h->points[i].values[v->active_value];
  }

  FOR(i, p->npts) {
    lambda = p->pts[i].lambda;
    edge   = p->pts[i].on_edge;
    v1     = hexvue_edge_vertex_indices[edge][0];
    v2     = hexvue_edge_vertex_indices[edge][1];
    pnts[i].x = hex_pnts[v1].x + (hex_pnts[v2].x - hex_pnts[v1].x) * lambda;
    pnts[i].y = hex_pnts[v1].y + (hex_pnts[v2].y - hex_pnts[v1].y) * lambda;
    pnts[i].z = hex_pnts[v1].z + (hex_pnts[v2].z - hex_pnts[v1].z) * lambda;
  }
  
  if (p->grep == NULL) {
    switch (p->npts) {
    case 3:
      p->grep = CreateTriangle3D(pnts);
      break;
    case 4:
      p->grep = CreateQuad3D(pnts);
      break;
    default:
      REPORT_ERROR("bad poly");
      break;
    }
    EASValsSetColor(
        ColorFringeValueToColorHSV(EMGetAssocFringeTable(ESIModel()),
                                   ct->contour_value));
    EGWithMaskChangeAttributes(ALL_ATTRIB_MASK, p->grep);
    EMNoDrawAddGraphicsToModel(v->model, p->grep);
  } else {
    EGModifyGraphicsGeometry(p->grep, pnts);
  }
  return 1;
}


int
cut_hex_contour(hexvue_hex_volume_t *h, /* IN */
                int active_value, /* IN */
                double contour_value, /* IN */
                hexvue_cut_point_t *cutpts, /* OUT */
                int *ncutpts   /* OUT # generated points */
                );

static int
split_hex(hexvue_hex_volume_t *h, hexvue_model_t *v);


static int
add_to_contour(hexvue_contour_t *ct,
               hexvue_model_t *v,
               hexvue_hex_volume_t *h)
{
#define MAX_ISECTP 24
  hexvue_cut_point_t  cut_points[MAX_ISECTP];
  int ncutpts, i;
  hexvue_cut_poly_t *p;
  
#define NEW_POLY()                              \
  {                                             \
    p = hexvue_new(hexvue_cut_poly_t);          \
    CHECK_MALLOC(p);                            \
    p->hex  = h;                                \
    p->grep = NULL;                             \
  }

  if (ct->polys == NULL) ct->polys = make_list();

  if (HEX_HAS_OFFSPRING(h)) {
    int result = 1;
    FOR(i, 8) {
      if (h->offspring[i] != NULL) {
        if (!add_to_contour(ct, v, h->offspring[i]))
          result = 0;
      }
    }
    return result;
  } else { /* no offspring */
    if (!cut_hex_contour(h, v->active_value, ct->contour_value,
                         cut_points, &ncutpts)) {
      return split_hex(h, v);
    } else {
      if (ncutpts > 0) {
        switch (ncutpts) {
        case 3:
          NEW_POLY();
          p->npts = 3;
          FOR(i, 3) {
            p->pts[i] = cut_points[i];
          }
          add_to_tail(ct->polys, p);
          return 1;
        case 4:
          NEW_POLY();
          p->npts = 4;
          FOR(i, 4) {
            p->pts[i] = cut_points[i];
          }
          add_to_tail(ct->polys, p);
          return 1;
        case 5:
          NEW_POLY();
          p->npts = 3;
          FOR(i, 3) {
            p->pts[i] = cut_points[i];
          }
          add_to_tail(ct->polys, p);
          NEW_POLY();
          p->npts = 4;
          p->pts[0] = cut_points[2];
          p->pts[1] = cut_points[3];
          p->pts[2] = cut_points[4];
          p->pts[3] = cut_points[0];
          add_to_tail(ct->polys, p);
          break;
        case 6:
          NEW_POLY();
          p->npts = 4;
          p->pts[0] = cut_points[0];
          p->pts[1] = cut_points[1];
          p->pts[2] = cut_points[2];
          p->pts[3] = cut_points[5];
          add_to_tail(ct->polys, p);
          NEW_POLY();
          p->npts = 4;
          p->pts[0] = cut_points[2];
          p->pts[1] = cut_points[3];
          p->pts[2] = cut_points[4];
          p->pts[3] = cut_points[5];
          add_to_tail(ct->polys, p);
          break;
        default:
          /* Split the hex and add its offspring to the tail of the list */
          REPORT_WARNING("suspicious cut point count");
          return 0;
        }
      }
    }
  } /* no offspring  */
  return 1;
}


#define subt_vec(ab, a, b)    \
   (ab).x = (a).x - (b).x, \
   (ab).y = (a).y - (b).y, \
   (ab).z = (a).z - (b).z
#define avg_vec(ab, a, b)             \
   (ab).x = 0.5 * ((a).x + (b).x), \
   (ab).y = 0.5 * ((a).y + (b).y), \
   (ab).z = 0.5 * ((a).z + (b).z)
#define cross_p(ab, a, b)                      \
  (ab).x = ((a).y*(b).z) - ((a).z*(b).y), \
  (ab).y = ((a).z*(b).x) - ((a).x*(b).z), \
  (ab).z = ((a).x*(b).y) - ((a).y*(b).x)
#define dot_p(a, b) ((a).x*(b).x + (a).y*(b).y + (a).z*(b).z)
#define lin_comb(r, p, pq, l)                   \
  (r).x = (p).x + (l) * (pq).x,              \
  (r).y = (p).y + (l) * (pq).y,              \
  (r).z = (p).z + (l) * (pq).z
#define NEWVA() hexvue_new_array(n, hexvue_float_t); 
#define NEWHEX(nh) {  nh = hexvue_new(hexvue_hex_volume_t);        \
                      CHECK_MALLOC(nh);                            \
                      nh->points[0].values = NEWVA();              \
                      CHECK_MALLOC(nh->points[0].values);          \
                      nh->points[1].values = NEWVA();              \
                      CHECK_MALLOC(nh->points[1].values);          \
                      nh->points[2].values = NEWVA();              \
                      CHECK_MALLOC(nh->points[2].values);          \
                      nh->points[3].values = NEWVA();              \
                      CHECK_MALLOC(nh->points[3].values);          \
                      nh->points[4].values = NEWVA();              \
                      CHECK_MALLOC(nh->points[4].values);          \
                      nh->points[5].values = NEWVA();              \
                      CHECK_MALLOC(nh->points[5].values);          \
                      nh->points[6].values = NEWVA();              \
                      CHECK_MALLOC(nh->points[6].values);          \
                      nh->points[7].values = NEWVA();              \
                      CHECK_MALLOC(nh->points[7].values);          \
                      nh->level  = h->level + 1;                   \
                      nh->parent = h;                              \
                      nh->offspring[0]                             \
                        = nh->offspring[1]                         \
                          = nh->offspring[2]                       \
                            = nh->offspring[3]                     \
                              = nh->offspring[4]                   \
                                = nh->offspring[5]                 \
                                  = nh->offspring[6]               \
                                    = nh->offspring[7]             \
                                      = NULL;                      \
                      nh->grep   = NULL;                           \
                      nh->plain_grep   = NULL;                  \
                      {                                         \
                        int _k_;                                \
                        FOR(_k_, 12) {                          \
                          nh->edge_greps[_k_] = NULL;           \
                        }                                       \
                      }                                         \
                    }
#define avg_vals(valdest, val1, val2)                                        \
{                                                                            \
  int avg_values_i;                                                          \
  FOR(avg_values_i, n) {                                                     \
    valdest[avg_values_i] = 0.5 * (val1[avg_values_i] + val2[avg_values_i]); \
  }                                                                          \
}
#define avg_vals_4(valdest, val1, val2, val3, val4)                          \
{                                                                            \
  register int avg_values_i;                                                 \
  FOR(avg_values_i, n) {                                                     \
    valdest[avg_values_i] = 0.25 * (  val1[avg_values_i]                     \
                                    + val2[avg_values_i]                     \
                                    + val3[avg_values_i]                     \
                                    + val4[avg_values_i]                     \
                                    );                                       \
  }                                                                          \
}
#define cpy_vals(valdest, valsrc)                                            \
{                                                                            \
  int avg_values_i;                                                          \
  FOR(avg_values_i, n) {                                                     \
    valdest[avg_values_i] = valsrc[avg_values_i];                            \
  }                                                                          \
}
#define HV(hex, pnt) hex->points[pnt].values
#define compute_center_vals(hex)                                             \
{                                                                            \
  int avg_values_i;                                                          \
  FOR(avg_values_i, n) {                                                     \
    center_vals[avg_values_i] = 0.125 * (  HV(hex, 0)[avg_values_i]          \
                                         + HV(hex, 1)[avg_values_i]          \
                                         + HV(hex, 2)[avg_values_i]          \
                                         + HV(hex, 3)[avg_values_i]          \
                                         + HV(hex, 4)[avg_values_i]          \
                                         + HV(hex, 5)[avg_values_i]          \
                                         + HV(hex, 6)[avg_values_i]          \
                                         + HV(hex, 7)[avg_values_i]);        \
  }                                                                          \
}




static int
split_hex(hexvue_hex_volume_t *h, hexvue_model_t *v)
{
  hexvue_hex_volume_t *nh;
  int n = v->data.num_of_values;
  hexvue_loc_t p0, p1, p2, p3, p4, p5, p6, p7;
  hexvue_loc_t p01, p12, p23, p30, p45, p56, p67, p74, p04, p15, p26, p37;
  hexvue_loc_t p0145, p1256, p2367, p3074, p4567, p0123;
  hexvue_loc_t center;
  hexvue_float_t center_vals[100];
#define ADD_TO_HEX_VOLUMES(hex, num, newhex)  hex->offspring[num] = newhex 
#define CUT_OFF_LEVEL 3

  if (h->level == CUT_OFF_LEVEL)
    return 1;

  p0 = h->points[0].loc; 
  p1 = h->points[1].loc;
  p2 = h->points[2].loc;
  p3 = h->points[3].loc;
  p4 = h->points[4].loc;
  p5 = h->points[5].loc;
  p6 = h->points[6].loc;
  p7 = h->points[7].loc;
  avg_vec(p01, p0, p1);
  avg_vec(p12, p1, p2);
  avg_vec(p23, p2, p3);
  avg_vec(p30, p3, p0);
  avg_vec(p45, p4, p5);
  avg_vec(p56, p5, p6);
  avg_vec(p67, p6, p7);
  avg_vec(p74, p7, p4);
  avg_vec(p04, p0, p4);
  avg_vec(p15, p1, p5);
  avg_vec(p26, p2, p6);
  avg_vec(p37, p3, p7);
  avg_vec(p0145, p01, p45);
  avg_vec(p1256, p12, p56);
  avg_vec(p2367, p23, p67);
  avg_vec(p3074, p30, p74);
  avg_vec(p4567, p45, p67);
  avg_vec(p0123, p01, p23);
  avg_vec(center, p0145, p2367);
  compute_center_vals(h);
  
#define PT(num, setto) nh->points[num].loc = setto
#define CV(to, from) cpy_vals(HV(nh, to), HV(h, from))
#define AV(to, v1, v2) avg_vals(HV(nh, to), HV(h, v1), HV(h, v2));
#define AV4(to, v1, v2, v3, v4) \
  avg_vals_4(HV(nh, to), HV(h, v1), HV(h, v2), HV(h, v3), HV(h, v4))
#define CCV(num) cpy_vals(HV(nh, num), center_vals)
#define PC(num) { PT(num, p##num); CV(num, num); }
#define PA(at, num1, num2) { PT(at, p##num1##num2); AV(at, num1, num2); }
#define PA4(at, num1, num2, num3, num4) \
    { PT(at, p##num1##num2##num3##num4); AV4(at, num1, num2, num3, num4); }
#define PCC(at) { PT(at, center); CCV(at); }


  /* Vertex 0 */
  NEWHEX(nh);
  PC(0);
  PA(1, 0, 1);
  PA4(2, 0, 1, 2, 3);
  PA(3, 3, 0);
  PA(4, 0, 4);
  PA4(5, 0, 1, 4, 5);
  PCC(6);
  PA4(7, 3, 0, 7, 4);
  ADD_TO_HEX_VOLUMES(h, 0, nh);


  /* Vertex 1 */
  NEWHEX(nh);
  PA(0, 0, 1);
  PC(1);
  PA(2, 1, 2);
  PA4(3, 0, 1, 2, 3);
  PA4(4, 0, 1, 4, 5);
  PA(5, 1, 5);
  PA4(6, 1, 2, 5, 6);
  PCC(7);
  ADD_TO_HEX_VOLUMES(h, 1, nh); 


  /* Vertex 2 */
  NEWHEX(nh);
  PA4(0, 0, 1, 2, 3);
  PA(1, 1, 2);
  PC(2);
  PA(3, 2, 3);
  PCC(4);
  PA4(5, 1, 2, 5, 6);
  PA(6, 2, 6);
  PA4(7, 2, 3, 6, 7);
  ADD_TO_HEX_VOLUMES(h, 2, nh);


  /* Vertex 3 */
  NEWHEX(nh); 
  PA(0, 3, 0);
  PA4(1, 0, 1, 2, 3);
  PA(2, 2, 3);
  PC(3);     
  PA4(4, 3, 0, 7, 4);
  PCC(5);
  PA4(6, 2, 3, 6, 7);
  PA(7, 3, 7);
  ADD_TO_HEX_VOLUMES(h, 3, nh); 


  /* Vertex 4 */
  NEWHEX(nh);  
  PA(0, 0, 4);
  PA4(1, 0, 1, 4, 5);
  PCC(2);
  PA4(3, 3, 0, 7, 4);
  PC(4);  
  PA(5, 4, 5);
  PA4(6, 4, 5, 6, 7);
  PA(7, 7, 4);
  ADD_TO_HEX_VOLUMES(h, 4, nh); 


  /* Vertex 5 */
  NEWHEX(nh);  
  PA4(0, 0, 1, 4, 5);
  PA(1, 1, 5);
  PA4(2, 1, 2, 5, 6);
  PCC(3);
  PA(4, 4, 5);
  PC(5);
  PA(6, 5, 6);
  PA4(7, 4, 5, 6, 7);
  ADD_TO_HEX_VOLUMES(h, 5, nh);  


  /* Vertex 6 */
  NEWHEX(nh); 
  PCC(0);
  PA4(1, 1, 2, 5, 6);
  PA(2, 2, 6);
  PA4(3, 2, 3, 6, 7);
  PA4(4, 4, 5, 6, 7);
  PA(5, 5, 6);
  PC(6);
  PA(7, 6, 7);
  ADD_TO_HEX_VOLUMES(h, 6, nh); 


  /* Vertex 7 */
  NEWHEX(nh); 
  PA4(0, 3, 0, 7, 4);
  PCC(1);
  PA4(2, 2, 3, 6, 7);
  PA(3, 3, 7);
  PA(4, 7, 4);
  PA4(5, 4, 5, 6, 7);
  PA(6, 6, 7);
  PC(7); 
  ADD_TO_HEX_VOLUMES(h, 7, nh);

  return 0;
}



int
hexvue_write_contours(hexvue_model_t *v, char *name)
{
  FILE *fp;
  hexvue_contour_t *ct;

  fp = fopen(name, "w");
  if (fp == NULL)
    return 0;
  
  fprintf(fp, "! HEXVUE contours file (C) 1995 Petr Krysl\n");
  fprintf(fp, "! %s (file %s). \n",
          v->comment, v->in_file);
  fprintf(fp, "!   There are %d contours here.\n",
          v->contours? list_length(v->contours): 0);
  fprintf(fp, "!   The active set.\n");
  fprintf(fp, "     %d\n", v->active_value);
  fprintf(fp, "!   value   layer \n");

  if (v->contours) { 
    loop_through_list(v->contours, ct, hexvue_contour_t *) {
      fprintf(fp, "%g   %d\n", ct->contour_value, ct->layer);
    }
  }
  fprintf(fp, "! END HEXVUE contour file.\n"); 
  
  fclose(fp);

  return 1;
}


int
hexvue_read_contours(hexvue_model_t *v, char *name)
{
  double contour_value;
  int contour_layer, active_set;
  
  if (!tokens_open_file(name))
    return 0;

  if (!tokens_next_line())
    return 0;
  active_set = tokens_token_as_int(1);
  
  hexvue_pick_set(v, active_set); 
  
  while(tokens_next_line()) {
    if (tokens_total_of_tokens() != 2)
      return 0;
    contour_value   = tokens_token_as_double(1);
    contour_layer   = tokens_token_as_int(2);
    EASValsSetLayer(contour_layer);
    hexvue_generate_contour(v, contour_value);
  }

  tokens_close_curr_file();
  return 1;
}




#ifdef NEED_HEX_DUMP

static void
dump_hex_elixir_format(hexvue_hex_volume_t *h)
{
  printf("811700647 1 17 0 0 0\n");
  printf(
         "%f %f %f %f %f %f %f %f %f %f %f %f\n"
         "%f %f %f %f %f %f %f %f %f %f %f %f\n%d %d %d %f\n%s\n%s\n",
         h->points[0].loc.x, h->points[0].loc.y, h->points[0].loc.z, 
         h->points[1].loc.x, h->points[1].loc.y, h->points[1].loc.z, 
         h->points[2].loc.x, h->points[2].loc.y, h->points[2].loc.z, 
         h->points[3].loc.x, h->points[3].loc.y, h->points[3].loc.z, 
         h->points[4].loc.x, h->points[4].loc.y, h->points[4].loc.z, 
         h->points[5].loc.x, h->points[5].loc.y, h->points[5].loc.z, 
         h->points[6].loc.x, h->points[6].loc.y, h->points[6].loc.z, 
         h->points[7].loc.x, h->points[7].loc.y, h->points[7].loc.z,
         0, 1, 1, dshrink, dcolor, dedgecolor);
}

#endif /* #ifdef NEED_HEX_DUMP */




int
hvupriv_cut_hex_by_crack_face (hexvue_hex_volume_t *h,
                               hexvue_crack_surface_t *cs,
                               int num_of_values,
                               LIST add_to_this
                               )
{
  int j, i, n = num_of_values;
  hexvue_float_t Nmat[8];
  hexvue_hex_volume_t *ho;
  hexvue_hex_volume_t *th;
  static ListOf(hexvue_crack_poly_t *) polys_secting_hex = NULL;
  static hexvue_vect_t approx_face_normal;
#define MAX_CUTPTS 20
  static int ncutpts;
  static hexvue_cut_point_t cutpts[MAX_CUTPTS];
  static hexvue_hex_volume_t *top_hexes[2] = { NULL, NULL };
  hexvue_float_t centera[3];
  hexvue_loc_t center;
#define COMPUTE_CENTER(HEX)                                                  \
  {                                                                          \
    hex_center(ALL_HEX_COORDS(HEX), centera);                                \
    center.x = centera[0]; center.y = centera[1]; center.z = centera[2];     \
  }
#define HVSKSI(I, J)                            \
       hvs_hexes[I].vertex[J].xyz[0],           \
       hvs_hexes[I].vertex[J].xyz[1],           \
       hvs_hexes[I].vertex[J].xyz[2]
#define COMP_N(arg) hex_compute_N_matrix(Nmat, arg)
#define SAVE_TOP_HEXES(hex)                                          \
  {                                                                  \
    int _v_;                                                         \
    if (top_hexes[0] == NULL) {                                      \
      top_hexes[0] = copy_hex_buf(num_of_values);                    \
      top_hexes[1] = copy_hex_buf(num_of_values);                    \
    }                                                                \
    FOR(_v_, 8) {                                                    \
      COMP_N(HVSKSI(0, _v_));                                        \
      N_IPOL_LOCS(top_hexes[0], _v_, hex, Nmat);                     \
      IPOL_VALS(top_hexes[0], h, _v_, n, Nmat);                      \
      COMP_N(HVSKSI(1, _v_));                                        \
      N_IPOL_LOCS(top_hexes[1], _v_, hex, Nmat);                     \
      IPOL_VALS(top_hexes[1], h, _v_, n, Nmat);                      \
    }                                                                \
  }
#define E0 REPORT_WARNING("cut_hex_by_crack_surface failed")
#define E1 REPORT_WARNING("split recurs on the 2nd level")
#define CUT_BY_CS(HEX)                                          \
   flush_list(polys_secting_hex, NOT_NODES);                    \
   if (!cut_hex_by_crack_surface(HEX, cs, cutpts, &ncutpts,     \
                                 &approx_face_normal,           \
                                 polys_secting_hex)) { E0; }
#define SPLIT_BY_SD(HEX)                                                \
   split_hex_by_section_data(cutpts, ncutpts, HEX, MAX_HEXES_DIM,       \
                             &nhexes, hvs_hexes)
#define HOSP(I, J) hvs_hexes[I].vertex[J].is_on_split_plane
#define PROJECT_POLYS()  \
     project_crack_polys(&approx_face_normal, &center, polys_secting_hex)
#define IPOL_ON_POLYS(HEX, VERT, NVALS) \
   ipol_on_polys(HEX, VERT, &center, NVALS)

  if (!polys_secting_hex) polys_secting_hex = make_list();

  CUT_BY_CS(h);

  if (ncutpts <= 2) 
    return 0;/* Done -- partial cut is not expected here */

  if (!SPLIT_BY_SD (h)) { 

    SAVE_TOP_HEXES (h); DH (h); DH (top_hexes[0]); DH (top_hexes[1]);
    
    th = top_hexes[0];
    CUT_BY_CS (th);
    if (!SPLIT_BY_SD (th)) { E1; }  
    FOR(i, nhexes) {
      P_NEWHEX (ho);
      FOR(j, 8) { COMP_N(HVSKSI(i, j)); N_IPOL_LOCS(ho, j, th, Nmat); }
      COMPUTE_CENTER (ho); PROJECT_POLYS ();
      FOR(j, 8) {
        if (HOSP (i, j)) {IPOL_ON_POLYS(ho, j, n);}     /* RAW */
        else             {COMP_N(HVSKSI(i, j)); IPOL_VALS(ho, th, j, n, Nmat);}
      }
      add_to_tail (add_to_this, ho); DH (ho);
    } /* FOR(i, nhexes) */
    
    th = top_hexes[1];
    CUT_BY_CS(th);
    if (!SPLIT_BY_SD (th)) { E1; }
    FOR(i, nhexes) {
      P_NEWHEX (ho);
      FOR(j, 8) { COMP_N(HVSKSI(i, j)); N_IPOL_LOCS(ho, j, th, Nmat); }
      COMPUTE_CENTER (ho); PROJECT_POLYS ();
      FOR(j, 8) {
        if (HOSP (i, j)) {IPOL_ON_POLYS(ho, j, n); }      /* RAW */
        else             {COMP_N(HVSKSI(i, j)); IPOL_VALS(ho, th, j, n, Nmat);}
      }
      add_to_tail (add_to_this, ho); DH (ho);
    } /* FOR(i, nhexes) */
    
    return 1;
  } else {
    if (nhexes == 0) return 0;
    DH (h);
    FOR(i, nhexes) {
      P_NEWHEX(ho);
      FOR(j, 8) { COMP_N(HVSKSI(i, j)); N_IPOL_LOCS(ho, j, h, Nmat); }
      COMPUTE_CENTER(ho); PROJECT_POLYS();
      FOR(j, 8) {
        if (HOSP(i, j)) { IPOL_ON_POLYS(ho, j, n); }  
        else            { COMP_N(HVSKSI(i, j)); IPOL_VALS(ho, h, j, n, Nmat); }
      }
      add_to_tail(add_to_this, ho); DH(ho);
    } /* FOR(i, nhexes) */

    return 1;
  } /* else */
}




typedef struct crack_poly_projection_t {
  hexvue_crack_poly_t *cp;
  int                  npts;
  hexvue_loc2d_t       locs[4];
  hexvue_loc_t         param_coords;
}              crack_poly_projection_t;


static int num_poly_projections = 0;
static int poly_projections_dim = 0;
static crack_poly_projection_t *poly_projections = NULL;
static hexvue_plane_t poly_proj_plane;
static hexvue_loc_t poly_proj_plane_center;

static void
alloc_proj_buf(int n);


static void
project_crack_polys(hexvue_vect_t *normal, hexvue_vect_t *center,
                    ListOf(hexvue_crack_poly_t *) polys
                    )
{
  hexvue_crack_poly_t *cp;
  int np, n, i, p;
  hexvue_vect_t a;

  np = list_length(polys);
  if (np < 1) return;

  alloc_proj_buf(np);

  plane_consts_normal(&poly_proj_plane, normal, center);
  poly_proj_plane_center = *center;

  p = 0;
  loop_through_list(polys, cp, hexvue_crack_poly_t *) {
    poly_projections[p].cp   = cp;
    poly_projections[p].npts = n = cp->npts;
    FOR(i, n) {
      v3_sub(&cp->points[i].loc, center, &a);
      poly_projections[p].locs[i].x = v3_dot(&poly_proj_plane.u, &a);
      poly_projections[p].locs[i].y = v3_dot(&poly_proj_plane.v, &a);
    }
    p++;
  }
}


static void
alloc_proj_buf(int n)
{
  if (poly_projections == NULL) {
    poly_projections = hexvue_new_array(n, crack_poly_projection_t);
    CHECK_MALLOC(poly_projections);
    poly_projections_dim = n;
  } else {
    if (poly_projections_dim < n) {
      poly_projections
        = hexvue_realloc_array(poly_projections, n, crack_poly_projection_t);
      CHECK_MALLOC(poly_projections);
      poly_projections_dim = n;
    }
  }
  num_poly_projections = n;
}



static int
which_side_of_poly(int poly, hexvue_loc_t *hex_center);


static void
ipol_on_polys(hexvue_hex_volume_t *h,
              int vert,
              hexvue_loc_t *center, int nvals)
{
  hexvue_loc2d_t hvproj;
  hexvue_loc_t a;
  int i;
  double p0, p1, p2;
  hexvue_loc2d_t qpc;
  double tol;
  static hexvue_float_t *dest_vals;
  int side = 1;                 /* RAW */
#define COMPUTE_PARAM_COORDS_TRIANGLE()
#define COMPUTE_PARAM_COORDS_QUAD()

  v3_sub(&h->points[vert].loc, &poly_proj_plane_center, &a);
  hvproj.x = v3_dot(&poly_proj_plane.u, &a);
  hvproj.y = v3_dot(&poly_proj_plane.v, &a);

  FOR(i, num_poly_projections) {
    if        (poly_projections[i].cp->npts == 3) {
      hexvue_inv_map_triangle(&p0, &p1, &p2,
                              hvproj.x, hvproj.y,
                              poly_projections[i].locs[0].x,
                              poly_projections[i].locs[0].y,
                              poly_projections[i].locs[1].x,
                              poly_projections[i].locs[1].y,
                              poly_projections[i].locs[2].x,
                              poly_projections[i].locs[2].y);
      poly_projections[i].param_coords.x = p0;
      poly_projections[i].param_coords.y = p1;
      poly_projections[i].param_coords.z = p2;
    } else if (poly_projections[i].cp->npts == 4) {
      hexvue_quad2d_t q2d;

      q2d.loc[0].x = poly_projections[i].locs[0].x;
      q2d.loc[0].y = poly_projections[i].locs[0].y;
      q2d.loc[1].x = poly_projections[i].locs[1].x;
      q2d.loc[1].y = poly_projections[i].locs[1].y;
      q2d.loc[2].x = poly_projections[i].locs[2].x;
      q2d.loc[2].y = poly_projections[i].locs[2].y;
      q2d.loc[3].x = poly_projections[i].locs[3].x;
      q2d.loc[3].y = poly_projections[i].locs[3].y;
      hexvue_quad_inv_map(&q2d, &hvproj, 1.e-12, &qpc);
      poly_projections[i].param_coords.x = qpc.x;
      poly_projections[i].param_coords.y = qpc.y;
      poly_projections[i].param_coords.z = 0;
    } else {
      REPORT_ERROR("bad polygon");
    }
  }
  dest_vals = h->points[vert].values;
#define IN_TOL(v, u, l) (v < u && v > l)
#define CHECK_IN_AND_IPOL(I)                                                 \
  side = which_side_of_poly(I, center);                                      \
  if (poly_projections[I].npts == 3) {                                       \
    double x, y, z, u = 1+tol, l = -tol;                                     \
    x = poly_projections[I].param_coords.x;                                  \
    y = poly_projections[I].param_coords.y;                                  \
    z = poly_projections[I].param_coords.z;                                  \
    if (IN_TOL(x, u, l) && IN_TOL(y, u, l) && IN_TOL(z, u, l)) {             \
      int j_;                                                                \
      hexvue_float_t *_vals_[3];                                             \
      if (side > 0) {                                                        \
        _vals_[0] = poly_projections[I].cp->points[0].values_on_plus_side;   \
        _vals_[1] = poly_projections[I].cp->points[1].values_on_plus_side;   \
        _vals_[2] = poly_projections[I].cp->points[2].values_on_plus_side;   \
      } else {                                                               \
        _vals_[0] =poly_projections[I].cp->points[0].values_on_minus_side;   \
        _vals_[1] =poly_projections[I].cp->points[1].values_on_minus_side;   \
        _vals_[2] =poly_projections[I].cp->points[2].values_on_minus_side;   \
      }                                                                      \
      FOR(j_, nvals) {                                                       \
        dest_vals[j_]                                                        \
          = ( x * _vals_[0][j_] + y * _vals_[1][j_] + z * _vals_[2][j_]);    \
      }                                                                      \
      return;                                                                \
    }                                                                        \
  } else {                                                                   \
    double x, y, z, u = 1+tol, l = -1-tol;                                   \
    x = poly_projections[I].param_coords.x;                                  \
    y = poly_projections[I].param_coords.y;                                  \
    z = poly_projections[I].param_coords.z;                                  \
    if (IN_TOL(x, u, l) && IN_TOL(y, u, l) && IN_TOL(z, u, l)) {             \
      int j_;                                                                \
      hexvue_float_t *_vals_[4], N_[4];                                      \
      if (side > 0) {                                                        \
        _vals_[0] = poly_projections[I].cp->points[0].values_on_plus_side;   \
        _vals_[1] = poly_projections[I].cp->points[1].values_on_plus_side;   \
        _vals_[2] = poly_projections[I].cp->points[2].values_on_plus_side;   \
        _vals_[3] = poly_projections[I].cp->points[3].values_on_plus_side;   \
      } else {                                                               \
        _vals_[0] =poly_projections[I].cp->points[0].values_on_minus_side;   \
        _vals_[1] =poly_projections[I].cp->points[1].values_on_minus_side;   \
        _vals_[2] =poly_projections[I].cp->points[2].values_on_minus_side;   \
        _vals_[3] =poly_projections[I].cp->points[3].values_on_minus_side;   \
      }                                                                      \
      hexvue_q4_N(N_, x, y);                                                 \
      FOR(j_, nvals) {                                                       \
        dest_vals[j_] = (  N_[0] * _vals_[0][j_]                             \
                         + N_[1] * _vals_[1][j_]                             \
                         + N_[2] * _vals_[2][j_]                             \
                         + N_[3] * _vals_[3][j_]);                           \
      }                                                                      \
      return;                                                                \
    }                                                                        \
  }
  
  tol  = 0;
  while (1) {
    FOR(i, num_poly_projections) {
      CHECK_IN_AND_IPOL(i); 
    }
    tol += 0.02;
    if (tol > 1) REPORT_ERROR("too many attempts");
  }
}



static int
which_side_of_poly(int poly, hexvue_loc_t *hex_center)
{
  hexvue_plane_t plane;
  hexvue_crack_poly_t *cp = poly_projections[poly].cp;
  hexvue_loc_t a;
#define CPLOC(N) &cp->points[N].loc
  
  if (cp->npts == 3) {
    plane_consts_3(&plane, CPLOC(0), CPLOC(1), CPLOC(2));
  } else {
    plane_consts_4(&plane, CPLOC(0), CPLOC(1), CPLOC(2), CPLOC(3));
  }
  v3_sub(hex_center, &cp->points[0].loc, &a);
  return (v3_dot(&plane.n, &a) > 0
          ? +1
          : -1);
}


static int
write_header (hexvue_model_t *v, ckit_iobuf_t IOB)
{
  int i;
  char buf[132];

  RUN_IOP (ckit_io_put_line_no_nl (IOB, "! HEXVUE dumped from "));
  RUN_IOP (ckit_io_put_line (IOB, v->in_file));
  RUN_IOP (ckit_io_put_line (IOB, v->comment));
  sprintf (buf, "%d", v->data.num_of_values);
  RUN_IOP (ckit_io_put_line (IOB, buf));
  FOR(i, v->data.num_of_values) {
    sprintf (buf, "\"%s\"", v->data.values[i].description);
    RUN_IOP (ckit_io_put_line (IOB, buf));
  }
  return 1;
}


static int
write_hex(hexvue_model_t *v, hexvue_hex_volume_t *h, ckit_iobuf_t IOB);


static int
write_data(hexvue_model_t *v, ckit_iobuf_t IOB, int what)
{
  hexvue_hex_volume_t *h;

  switch (what) {
  case WRITE_HEX_ONLY:
    loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
      if (!write_hex (v, h, IOB))
        return 0;
    }
    break;
  case WRITE_CRACK_ONLY:
    break;
  }
  
  RUN_IOP (ckit_io_put_line (IOB, "END ! HEXVUE file."));
  return 1;
}


hexvue_bool_t
hexvue_write_model (hexvue_model_t *v, char *file, int what)
{
  ckit_iobuf_t IOB;
  FILE *fp;

  fp = ckit_io_open_stream (file, "w");
  if (fp == NULL) {
    REPORT_ERROR ((char *)ckit_io_error_explanation ());
    return 0;
  }
  IOB = ckit_io_new_buffer (fp);
  if (IOB == NULL) {
    REPORT_ERROR ((char *)ckit_io_error_explanation ());
    return 0;
  }

  if (!write_header (v, IOB))
    goto error_exit;
  
  if (!write_data (v, IOB, what))
    goto error_exit;

  fp = ckit_io_delete_buffer (IOB);
  fclose (fp);
  
  return TRUE;
  
 error_exit:
  if (IOB != NULL) {
    fp = ckit_io_delete_buffer (IOB);
  }
  if (fp != NULL)
    fclose (fp);
    
  return FALSE;
}





static int
write_hex (hexvue_model_t *v, hexvue_hex_volume_t *h, ckit_iobuf_t IOB)
{
  int j, i, n = v->data.num_of_values;
  int reqlen = n;
  float geom[3];
  static int buflen        = 0;
  static float *minv       = NULL;
  static float *maxv       = NULL;
  static unsigned char *lo = NULL;
  static unsigned char *hi = NULL;

  if (buflen < reqlen) {
    minv   = hexvue_realloc_array (minv, reqlen, float);
    maxv   = hexvue_realloc_array (maxv, reqlen, float);
    lo     = hexvue_realloc_array (lo, reqlen, unsigned char);
    hi     = hexvue_realloc_array (hi, reqlen, unsigned char);
    buflen = reqlen;
  }

  FOR (i, n) {
    minv[i] = FLT_MAX;
    maxv[i] = -FLT_MAX;
    FOR (j, 8) {
      minv[i] = min (minv[i], h->points[j].values[i]);
      maxv[i] = max (maxv[i], h->points[j].values[i]);
    }
  }
  
  RUN_IOP (ckit_io_put_line (IOB, "BHV"));
  RUN_IOP (ckit_io_put_data (IOB, minv, n * sizeof (float)));
  RUN_IOP (ckit_io_put_data (IOB, maxv, n * sizeof (float)));
  FOR (j, 8) {
    geom[0] = h->points[j].loc.x;
    geom[1] = h->points[j].loc.y;
    geom[2] = h->points[j].loc.z;
    RUN_IOP (ckit_io_put_data (IOB, geom, 3 * sizeof (float)));
    FOR (i, n) {
      ckit_io_convert_double_to_bytes (h->points[j].values[i],
                                       minv[i], maxv[i],
                                       &lo[i], &hi[i]);
    }      
    RUN_IOP (ckit_io_put_data (IOB, lo, n * sizeof (unsigned char)));
    RUN_IOP (ckit_io_put_data (IOB, hi, n * sizeof (unsigned char)));
  }

  return 1;
}


void
hexvue_set_allow_presplit(hexvue_model_t *v, hexvue_bool_t flag)
{
  v->allow_presplit = flag;
}


static int
pt_inside_hexvol(hexvue_hex_volume_t *h, /* IN */
                 hexvue_loc_t *pt, /* IN */
                 hexvue_loc_t *param_pt /* OUT */
                 );
static int
interpolate_values(hexvue_model_t *v,
                   hexvue_hex_volume_t *h,
                   hexvue_loc_t *param_pt,
                   int *nvalues
                   );


static double *values = NULL;

double const *
hexvue_values_at_point(hexvue_model_t *v, /* IN */
                       double x, double y, double z, /* IN */
                       int *nvalues /* OUT */
                       )
{
  hexvue_hex_volume_t *h;
  hexvue_loc_t pt, param_pt;

  pt.x = x;
  pt.y = y;
  pt.z = z;
  loop_through_list(v->hex_volumes, h, hexvue_hex_volume_t *) {
    if (pt_inside_hexvol(h, &pt, &param_pt)) {
      interpolate_values(v, h, &param_pt, nvalues);
      return values;
    }
  }
  return NULL;
}


static int
interpolate_values(hexvue_model_t *v,
                   hexvue_hex_volume_t *h,
                   hexvue_loc_t *param_pt,
                   int *nvalues
                   )
{
  int i;
  int n = v->data.num_of_values;
  double Nmat[8];

  if (values == NULL) {
    values = hexvue_new_array(n, double);
  } else {
    values = hexvue_realloc_array(values, n, double);
  }
  if (values == NULL)
    return FALSE;
  
  hex_compute_N_matrix(Nmat, param_pt->x, param_pt->y, param_pt->z);
  FOR(i, n) {
    values[i] = (  Nmat[0] * h->points[0].values[i]
                 + Nmat[1] * h->points[1].values[i]
                 + Nmat[2] * h->points[2].values[i]
                 + Nmat[3] * h->points[3].values[i]
                 + Nmat[4] * h->points[4].values[i]
                 + Nmat[5] * h->points[5].values[i]
                 + Nmat[6] * h->points[6].values[i]
                 + Nmat[7] * h->points[7].values[i]
                   );
  }
  *nvalues = n;
  return TRUE;
}



static int
pt_inside_hexvol(hexvue_hex_volume_t *h, /* IN */
                 hexvue_loc_t *pt, /* IN */
                 hexvue_loc_t *param_pt /* OUT */
                 )
{
#define HVLOC(h, i) h->points[i].loc
#define IN_BI_UNIT(val) ((val >= -1-eps) && (val <= 1+eps))
  hexvue_box_t hbbox;
  double tol, eps = 0.0001;
  G3D_hex_t ah;
#define TOL() tol = 0.001 * (  (hbbox.upper.x - hbbox.lower.x)   \
                             + (hbbox.upper.y - hbbox.lower.y )  \
                             + (hbbox.upper.z - hbbox.lower.z))
#define INSIDE_BBOX(pt, bbox)                                                 \
         (   ((pt).x <= bbox.upper.x + tol) && ((pt).x >= bbox.lower.x - tol) \
          && ((pt).y <= bbox.upper.y + tol) && ((pt).y >= bbox.lower.y - tol) \
          && ((pt).z <= bbox.upper.z + tol) && ((pt).z >= bbox.lower.z - tol) \
          )
#define ASSP1(num, dir) ah.loc[num].dir = HVLOC(h, num).dir
#define ASSP(num) { ASSP1(num, x); ASSP1(num, y); ASSP1(num, z); }
#define INIT_AH \
  {ASSP(0); ASSP(1); ASSP(2); ASSP(3); ASSP(4); ASSP(5); ASSP(6); ASSP(7);}
#undef CBND
#define CBND(hbbox, hex, which, func, N)                    \
  {                                                         \
    hbbox.which.x = func(hbbox.which.x, HVLOC(h, N).x);  \
    hbbox.which.y = func(hbbox.which.y, HVLOC(h, N).y);  \
    hbbox.which.z = func(hbbox.which.z, HVLOC(h, N).z);  \
  }
#undef DEFINE_HBBOX
#define DEFINE_HBBOX(hex, hbbox)                                     \
  {                                                                  \
    hbbox.lower.x = hbbox.upper.x = HVLOC(h, 0).x;                   \
    hbbox.lower.y = hbbox.upper.y = HVLOC(h, 0).y;                   \
    hbbox.lower.z = hbbox.upper.z = HVLOC(h, 0).z;                   \
    CBND(hbbox, hex, lower, min, 1);                                 \
    CBND(hbbox, hex, lower, min, 2);                                 \
    CBND(hbbox, hex, lower, min, 3);                                 \
    CBND(hbbox, hex, lower, min, 4);                                 \
    CBND(hbbox, hex, lower, min, 5);                                 \
    CBND(hbbox, hex, lower, min, 6);                                 \
    CBND(hbbox, hex, lower, min, 7);                                 \
    CBND(hbbox, hex, upper, max, 1);                                 \
    CBND(hbbox, hex, upper, max, 2);                                 \
    CBND(hbbox, hex, upper, max, 3);                                 \
    CBND(hbbox, hex, upper, max, 4);                                 \
    CBND(hbbox, hex, upper, max, 5);                                 \
    CBND(hbbox, hex, upper, max, 6);                                 \
    CBND(hbbox, hex, upper, max, 7);                                 \
  }

  DEFINE_HBBOX(h, hbbox);
  TOL();
  if (!INSIDE_BBOX(*pt, hbbox))
    return FALSE;

  INIT_AH;
  if (G3D_hex_inv_map(&ah, pt, eps/10, param_pt)) {
    if (   IN_BI_UNIT(param_pt->x)
        && IN_BI_UNIT(param_pt->y)
        && IN_BI_UNIT(param_pt->z)) {
      return TRUE;
    } 
  } else {
    REPORT_WARNING("Failed to evaluate parametric coordinates");
  }
  return FALSE;
}
4049 lines
No id keywords (cm7)
