/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* This file holds the redefinition of graphic methods for the 10-node
   tetrahedron */

#include "Econfig.h"
#include "Ecfg.h"
#include <stdlib.h>

#include "Eview.h"
#include "Egraphic.h"
#include "Egeomops.h"
#include "Ecolors.h"
#include "Eactset.h"

#define TET10_PRIVATE_HEADER
#include "Etet10.h"

TET10_SUBTRI_VERTS;

#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

/* ========================================================================= */
/* PRIVATE DATA */
/* ========================================================================= */

#define STRING_BUFF_SIZE 512

static char string_rep_buffer[STRING_BUFF_SIZE];

/* ========================================================================= */
/* PUBLIC */
/* ========================================================================= */

GraphicObj *
CreateTet10 (WCRec *points)
{
  ETet10WCRec *thedp;
  GraphicObj *g_p;
 
  thedp = (ETet10WCRec *)make_node(sizeof(ETet10WCRec));
  g_p = EGCreateGraphics(&tet10_graphic_methods, (caddrt)thedp,
			 sizeof(ETet10WCRec));
  ModifyGeometry (g_p, points);
  thedp->attributes.spec.fill_style = FILL_SOLID;
  thedp->attributes.spec.color      = 0x0;
  thedp->attributes.spec.edge_flag  = NO;
  thedp->attributes.spec.edge_color = 0x0;
  thedp->attributes.spec.shrink     = 1.0;
  thedp->attributes.spec.have_any   = YES;
  return g_p;
}



static void 
Draw (EView *v_p, GraphicObj *p)
{
  ETet10WCRec *thedp;

  thedp = (ETet10WCRec *)(p->spec.data);

  EVTet10Primitive (v_p, p,
                    thedp->points,
                    thedp->attributes.spec.color,
                    thedp->attributes.spec.fill_style,
                    thedp->attributes.spec.edge_flag,
                    thedp->attributes.spec.edge_color,
                    thedp->attributes.spec.shrink,
                    DRAW_MODE_DRAW);
}



static EGraphicType 
Type(GraphicObj *g_p)
{
  return EG_TET10;
}



static char *
AsString (GraphicObj *g_p)
{
  ETet10WCRec *l;

  l = (ETet10WCRec *)g_p->spec.data;
  sprintf(string_rep_buffer,
	  "TET10 [%f,%f,%f] [%f,%f,%f] [%f,%f,%f] [%f,%f,%f]"
          " [%f,%f,%f] [%f,%f,%f] [%f,%f,%f] [%f,%f,%f] [%f,%f,%f] [%f,%f,%f]",
	  l->points[0].x, l->points[0].y, l->points[0].z,
          l->points[1].x, l->points[1].y, l->points[1].z,
          l->points[2].x, l->points[2].y, l->points[2].z,
          l->points[3].x, l->points[3].y, l->points[3].z,
          l->points[4].x, l->points[4].y, l->points[4].z,
          l->points[5].x, l->points[5].y, l->points[5].z,
          l->points[6].x, l->points[6].y, l->points[6].z,
          l->points[7].x, l->points[7].y, l->points[7].z,
          l->points[8].x, l->points[8].y, l->points[8].z,
          l->points[9].x, l->points[9].y, l->points[9].z);
  
  return string_rep_buffer; /* Don't free this memory!!! */
}



static void 
RetrieveFrom (FILE *file_p, GraphicSpecObj *gsp)
{
  ETet10WCRec *p;
  int filled, edge_flag;
  char color[64];
  EPixel pixel;
  BOOLEAN success;
  double x, y, z;
  float shrink;
  int i;
  
  if ((p = (ETet10WCRec *)make_node(sizeof(ETet10WCRec)))
      != NULL) {
    for (i = 0; i < 10; i++) {
      fscanf(file_p, "%lf %lf %lf", &x, &y, &z);
      p->points[i].x = x;
      p->points[i].y = y;
      p->points[i].z = z;
    }
    fscanf(file_p,
	   "%d %d %d %f",
	   &(p->attributes.spec.have_any),
	   &filled, &edge_flag, &shrink);
    p->attributes.spec.fill_style = filled;
    p->attributes.spec.edge_flag  = edge_flag;
    p->attributes.spec.shrink     = shrink;
    fscanf(file_p, "%s", color);
    pixel = ColorGetPixelFromString(color, &success);
    if (success)
      p->attributes.spec.color = pixel;
    fscanf(file_p, "%s", color);
    pixel = ColorGetPixelFromString(color, &success);
    if (success)
      p->attributes.spec.edge_color = pixel;
  }
  gsp->data        = (caddrt)p;
  gsp->data_length = sizeof(ETet10WCRec);
}
 


static void 
StoreOn (FILE *file_p, GraphicObj *g_p)
{
  ETet10WCRec *p;
  char *color, *edge_color;
  int i;

  p = (ETet10WCRec *)g_p->spec.data;
  if (p->attributes.spec.have_any == YES) {
    color = ColorGetStringFromPixel(p->attributes.spec.color);
    if (color == NULL)
      color = "white";
    edge_color = ColorGetStringFromPixel(p->attributes.spec.edge_color);
    if (edge_color == NULL)
      edge_color = "white";
  } else {
    edge_color = color = "white";
  }
  for (i = 0; i < 10; i++) {
    fprintf (file_p, "%f %f %f ",
             p->points[i].x, p->points[i].y, p->points[i].z);
  }
  fprintf (file_p, "\n%d %d %d %f\n%s\n%s\n",
           (p->attributes.spec.have_any),
           p->attributes.spec.fill_style,
           p->attributes.spec.edge_flag,
           p->attributes.spec.shrink,
           color, edge_color);
}




static void 
BoundingBoxVC (EView *v_p, GraphicObj *g_p,
               VCRec *ll_return, VCRec *ur_return)
{
  ETet10WCRec *p;
  VCRec vcp;
  int i;
  
  p = (ETet10WCRec *)(g_p->spec.data);
  
  EVWCtoVC (v_p, &(p->points[0]), &vcp);
  ll_return->u = vcp.u;
  ur_return->u = vcp.u;
  ll_return->v = vcp.v;
  ur_return->v = vcp.v;
  ll_return->n = vcp.n;
  ur_return->n = vcp.n;
  for (i = 1; i < 10; i++) {
    EVWCtoVC (v_p, &(p->points[i]), &vcp);
    ll_return->u = min(ll_return->u, vcp.u);
    ll_return->v = min(ll_return->v, vcp.v);
    ll_return->n = min(ll_return->n, vcp.n);
    ur_return->u = max(ur_return->u, vcp.u);
    ur_return->v = max(ur_return->v, vcp.v);
    ur_return->n = max(ur_return->n, vcp.n);
  }
}



static void 
BoundingBoxWC (GraphicObj *g_p,
               WCRec *ll_return, WCRec *ur_return)
{
  ETet10WCRec *p;
  int i;
  
  p = (ETet10WCRec *)(g_p->spec.data);
  
  ur_return->x = ll_return->x = p->points[0].x;
  ur_return->y = ll_return->y = p->points[0].y;
  ur_return->z = ll_return->z = p->points[0].z;
  for (i = 1; i < 10; i++) {
    ll_return->x = min(ll_return->x, p->points[i].x);
    ll_return->y = min(ll_return->y, p->points[i].y);
    ll_return->z = min(ll_return->z, p->points[i].z);
    ur_return->x = max(ur_return->x, p->points[i].x);
    ur_return->y = max(ur_return->y, p->points[i].y);
    ur_return->z = max(ur_return->z, p->points[i].z);
  }
}




static void 
XorDraw (EView *v_p, GraphicObj *g_p)
{
  ETet10WCRec *thedp;

  thedp = (ETet10WCRec *)(g_p->spec.data);

  EVTet10Primitive (v_p, g_p,
                    thedp->points,
                    thedp->attributes.spec.color,
                    thedp->attributes.spec.fill_style,
                    NO, 0,
                    thedp->attributes.spec.shrink,
                    DRAW_MODE_XORDRAW);
}

 


static GraphicObj *
ModifyGeometry (GraphicObj *g_p, WCRec *points)
{
  ETet10WCRec *thedp;
  int i;

  thedp = (ETet10WCRec *)(g_p->spec.data);
  for (i = 0; i < 10; i++) {
    thedp->points[i] = points[i];
  }

  return g_p;
}



static BOOLEAN 
GetGeometry (GraphicObj *g_p, WCRec *points, int *npoints)
{
  ETet10WCRec *thedp;
  int i;
  
  thedp = (ETet10WCRec *)(g_p->spec.data);
  for (i = 0; i < 10; i++) {
    points[i] = thedp->points[i];
  }
  *npoints    = 10;
  
  return YES;
}



static void 
Hilite (EView *v_p, GraphicObj *g_p)
{
  ETet10WCRec *thedp;
  
  thedp = (ETet10WCRec *)(g_p->spec.data);

  EVTet10Primitive (v_p, g_p,
                    thedp->points,
                    thedp->attributes.spec.color,
                    FILL_HOLLOW, NO, 0,
                    1.0,
                    DRAW_MODE_HILITE);
}



static void 
Unhilite (EView *v_p, GraphicObj *g_p)
{
  ETet10WCRec *thedp;
  
  thedp = (ETet10WCRec *)(g_p->spec.data);

  EVTet10Primitive (v_p, g_p,
                    thedp->points,
                    thedp->attributes.spec.color,
                    FILL_HOLLOW, NO, 0,
                    1.0,
                    DRAW_MODE_UNHILITE);
}



static void 
Erase (EView *v_p, GraphicObj *g_p)
{
  ETet10WCRec *thedp;
  
  thedp = (ETet10WCRec *)(g_p->spec.data);

  EVTet10Primitive (v_p, g_p,
                    thedp->points,
                    thedp->attributes.spec.color,
                    thedp->attributes.spec.fill_style,
                    NO, 0,
                    thedp->attributes.spec.shrink,
                    DRAW_MODE_ERASE);
}



static GraphicObj *
Translate (GraphicObj *g_p, WCRec *by)
{
  ETet10WCRec *p;
  int i;

  p = (ETet10WCRec *)(g_p->spec.data);
  for (i = 0; i < 10; i++) {
    p->points[i].x += by->x;
    p->points[i].y += by->y;
    p->points[i].z += by->z;
  }
  return g_p;
}



static GraphicObj *
Rotate (GraphicObj *g_p, WCRec *center, WCRec *axial)
{
  ETet10WCRec *p;
  WCRec x_c;
  int i;
  
  p = (ETet10WCRec *)g_p->spec.data;
  for (i = 0; i < 10; i++) {
    x_c.x = p->points[i].x - center->x;
    x_c.y = p->points[i].y - center->y;
    x_c.z = p->points[i].z - center->z;
    RotVectAboutVect(axial, &x_c);
    p->points[i].x = center->x + x_c.x;
    p->points[i].y = center->y + x_c.y;
    p->points[i].z = center->z + x_c.z;
  }
  return g_p;
}



static GraphicObj  *
MirrorPlane(GraphicObj *g_p, WCRec *point_on_plane,
                 WCRec *unit_plane_normal)
{
  ETet10WCRec *dp;
  int i;
  
  dp = (ETet10WCRec *)g_p->spec.data;
  for (i = 0; i < 10; i++) {
    MirrorPointInPlane (point_on_plane, unit_plane_normal, &dp->points[i]);
  }
  return g_p;
}



static GraphicObj  *
MirrorCenter (GraphicObj *g_p, WCRec *center)
{
  ETet10WCRec *dp;
  int i;

  dp = (ETet10WCRec *)g_p->spec.data;
  for (i = 0; i < 10; i++) {
    dp->points[i].x = 2.*center->x - dp->points[i].x;
    dp->points[i].y = 2.*center->y - dp->points[i].y;
    dp->points[i].z = 2.*center->z - dp->points[i].z;
  }
  return g_p;
}


 
static GraphicObj *
Scale (GraphicObj *g_p, WCRec *center, 
		   FPNum sx, FPNum sy, FPNum sz)
{
  ETet10WCRec *dp;
  int i;

  dp = (ETet10WCRec *)g_p->spec.data;
  for (i = 0; i < 10; i++) {
    dp->points[i].x = sx * (dp->points[i].x - center->x) + center->x;
    dp->points[i].y = sy * (dp->points[i].y - center->y) + center->y;
    dp->points[i].z = sz * (dp->points[i].z - center->z) + center->z;
  }
  return g_p;
}



static BOOLEAN 
HitByBox (EView *v_p, EBoxVC2DRec *box,
          ESelectCriteria sel_crit, GraphicObj *g_p)
{
  VCRec ll, ur;

  switch (sel_crit) {
  case INSIDE:
    BoundingBoxVC (v_p, g_p, &ll, &ur);
    if (ll.u >= box->left   && ur.u <= box->right && 
	ll.v >= box->bottom && ur.v <= box->top)
      return YES;
    else
      return NO;
  case OVERLAP:
    {
      VCRec p1, p2, p3;
      ETet10WCRec *p;
      int k, i;
      
      p = (ETet10WCRec *)g_p->spec.data;
      for (k = 0; k < 16; k++) {
        i = tet10_subtri_verts[k][0];
        EVWCtoVC(v_p, &(p->points[i]), &p1);
        i = tet10_subtri_verts[k][1];
        EVWCtoVC(v_p, &(p->points[i]), &p2);
        i = tet10_subtri_verts[k][2];
        EVWCtoVC(v_p, &(p->points[i]), &p3);
        if (BoxOverlapsTriangle (box,
                                 (VC2DRec *)&p1,
                                 (VC2DRec *)&p2,
                                 (VC2DRec *)&p3))
          return YES;
      }
      return NO;
    }
  case INTERSECT:
  default:
    {
      VCRec p1, p2, p3;
      ETet10WCRec *p;
      int k, i;
      
      p = (ETet10WCRec *)g_p->spec.data;
      for (k = 0; k < 16; k++) {
        i = tet10_subtri_verts[k][0];
        EVWCtoVC(v_p, &(p->points[i]), &p1);
        i = tet10_subtri_verts[k][1];
        EVWCtoVC(v_p, &(p->points[i]), &p2);
        i = tet10_subtri_verts[k][2];
        EVWCtoVC(v_p, &(p->points[i]), &p3);
        if (BoxIntersectsTriangle (box,
                                   (VC2DRec *)&p1,
                                   (VC2DRec *)&p2,
                                   (VC2DRec *)&p3))
          return YES;
      }
      return NO;
    }
  }
}



static LIST 
KeyPoints (GraphicObj *p)
{
  LIST temp_list;
  KeyPointRec *key_p;
  ETet10WCRec *l_p;
  int i;

  l_p       = (ETet10WCRec *)(p->spec.data);

  temp_list = make_list();
  for (i = 0; i < 10; i++) {
    key_p = (KeyPointRec*)
      add_to_tail(temp_list, make_node(sizeof(KeyPointRec)));
    key_p->coords.x = l_p->points[i].x;
    key_p->coords.y = l_p->points[i].y;
    key_p->coords.z = l_p->points[i].z;
    key_p->dist_from_hit = 0.0;
  }
  return temp_list;
}



static GraphicObj *
Reshape (GraphicObj *p, EHandleNum hn, WCRec *pos)
{
  ETet10WCRec *_p;

  _p = (ETet10WCRec *)(p->spec.data);
  
  if (hn >= 1 && hn <= 10) {
    _p->points[hn-1] = *pos;
  }
  return p;
}



static EHandleNum 
ReshapeHandle (EView *v_p, GraphicObj *p, WCRec *nearp)
{
  ETet10WCRec *l_p;
  FPNum d, mind;
  VCRec nvc, pvc;
  int imin;
  int i;

  l_p       = (ETet10WCRec *)(p->spec.data);
  EVWCtoVC(v_p, nearp, &nvc); 
  EVWCtoVC(v_p, &(l_p->points[0]), &pvc);
  d = DistOfPnts (pvc.u, pvc.v, nvc.u, nvc.v);
  imin = 0; mind = d;
  for (i = 1; i <= 10; i++) {
    EVWCtoVC(v_p, &(l_p->points[i]), &pvc);
    d = DistOfPnts (pvc.u, pvc.v, nvc.u, nvc.v);
    if (d < mind) {
      imin = i;
      mind = d;
    }
  }
  return imin+1;
}



static void 
ShowHandles (EView *v_p, GraphicObj *p)
{
  ETet10WCRec *l_p;
  int i;

  l_p = (ETet10WCRec *)p->spec.data;
  for (i = 0; i < 10; i++) {
    EGDrawHandle(v_p, &(l_p->points[i]));
  }
}
  


static void 
EraseHandles (EView *v_p, GraphicObj *p)
{
  ETet10WCRec *l_p;
  int i;

  l_p = (ETet10WCRec *)p->spec.data;
  for (i = 0; i < 10; i++) {
    EGEraseHandle(v_p, &(l_p->points[i]));
  }
}



static void 
XYofHandle (GraphicObj *p, EHandleNum hn, WCRec *pos)
{
  ETet10WCRec *l_p;

  l_p = (ETet10WCRec *)p->spec.data;
  switch((int)hn) {
  case 1:
  case 2:
  case 3:
  case 4:
  case 5:
  case 6:
  case 7:
  case 8:
  case 9:
  case 10:
    pos->x = l_p->points[hn-1].x;
    pos->y = l_p->points[hn-1].y;
    pos->z = l_p->points[hn-1].z;
    break;
  default:
    pos->x = l_p->points[0].x;
    pos->y = l_p->points[0].y;
    pos->z = l_p->points[0].z;
    break;
  }
}



static void 
ChangeAttributes (GraphicObj *p)
{
  ETet10WCRec *dp;
  unsigned long mask;

  dp = (ETet10WCRec *)p->spec.data;

  mask = EASValsGetChangeMask();
  
  if ((mask & FILL_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.fill_style = EASValsGetFillStyle();
    dp->attributes.spec.have_any = YES;
  }
  if ((mask & COLOR_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.color = EASValsGetColor();
    dp->attributes.spec.have_any = YES;
  }
  if ((mask & EDGE_COLOR_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.edge_color = EASValsGetEdgeColor();
    dp->attributes.spec.have_any = YES;
  }
  if ((mask & EDGE_FLAG_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.edge_flag = EASValsGetEdgeFlag();
    dp->attributes.spec.have_any = YES;
  }
  if ((mask & SHRINK_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.shrink = EASValsGetShrink();
    dp->attributes.spec.have_any = YES;
  }
}



static void 
CopyAttributesToActiveSet (GraphicObj *p)
{
  ETet10WCRec *dp;

  dp = (ETet10WCRec *)p->spec.data;
  
  if (dp->attributes.spec.have_any) {
    EASValsSetFillStyle(dp->attributes.spec.fill_style);
    EASValsSetColor(dp->attributes.spec.color);
    EASValsSetEdgeColor(dp->attributes.spec.edge_color);
    EASValsSetEdgeFlag(dp->attributes.spec.edge_flag);
    EASValsSetShrink(dp->attributes.spec.shrink);
  }
}

static BOOLEAN
AttribValAsInAS (GraphicObj *p, unsigned long mask)  
{
  ETet10WCRec *dp;

  dp = (ETet10WCRec *)p->spec.data;
  
  if ((mask & FILL_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.fill_style != EASValsGetFillStyle())
      return NO;
  if ((mask & COLOR_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.color != EASValsGetColor())
      return NO;
  if ((mask & EDGE_COLOR_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.edge_color != EASValsGetEdgeColor())
      return NO;
  if ((mask & EDGE_FLAG_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.edge_flag != EASValsGetEdgeFlag())
      return NO;
  if ((mask & SHRINK_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.shrink != EASValsGetShrink())
      return NO;

  return YES;
}
