/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "Econfig.h"
#include "Ecfg.h"
#include <stdlib.h>

#include "Eview.h"
#include "Egraphic.h"
#include "Egeomops.h"
#include "Ecolors.h"
#include "Eactset.h"
#include "newP.h"
#define VECMAC_VECTOR_TYPE 1
#include "vecmacP.h"

typedef struct interact_rec_t {
  int   have_point_in_common;
  WCRec pts[3]; /* the point pts[1] is the common point */
  double cosa;
}              interact_rec_t;

typedef struct line_t {
  WCRec pts[2];
}              line_t;

static double
cosa (WCRec pts[2])
{
  WCRec d1, d2;
  VECMAC_3_VEC(d1, =, pts[0], -, pts[1]);
  VECMAC_3_VEC(d2, =, pts[2], -, pts[1]);
  if ((VECMAC_NORM(d1) >0) && (VECMAC_NORM(d2) >0)) {
    return VECMAC_DOT(d1, d2) / VECMAC_NORM(d1) / VECMAC_NORM(d2);
  } else
    return -1;
}

static interact_rec_t
get_ir (WCRec pts1[2], WCRec pts2[2], double merge_tol)
{
  interact_rec_t ir={0,{{0,0,0},{0,0,0},{0,0,0}},-1};
  WCRec d;
  VECMAC_3_VEC(d, =, pts1[0], -, pts2[0]);
  if (VECMAC_NORM(d) < merge_tol) {
    ir.have_point_in_common = 1;
    ir.pts[0] = pts1[1]; ir.pts[1] = pts1[0]; ir.pts[2] = pts2[1];
    ir.cosa = cosa (ir.pts);
    return ir;
  }
  VECMAC_3_VEC(d, =, pts1[1], -, pts2[0]);
  if (VECMAC_NORM(d) < merge_tol) {
    ir.have_point_in_common = 1;
    ir.pts[0] = pts1[0]; ir.pts[1] = pts1[1]; ir.pts[2] = pts2[1];
    ir.cosa = cosa (ir.pts);
    return ir;
  }
  VECMAC_3_VEC(d, =, pts1[1], -, pts2[1]);
  if (VECMAC_NORM(d) < merge_tol) {
    ir.have_point_in_common = 1;
    ir.pts[0] = pts1[0]; ir.pts[1] = pts1[1]; ir.pts[2] = pts2[0];
    ir.cosa = cosa (ir.pts);
    return ir;
  }
  VECMAC_3_VEC(d, =, pts1[0], -, pts2[1]);
  if (VECMAC_NORM(d) < merge_tol) {
    ir.have_point_in_common = 1;
    ir.pts[0] = pts1[1]; ir.pts[1] = pts1[0]; ir.pts[2] = pts2[0];
    ir.cosa = cosa (ir.pts);
    return ir;
  }
  return ir;
}

static int
have_line (line_t *ls[], int nl, WCRec p0, WCRec p1, double merge_tol)
{
  int i;
  for (i=0; i<nl; i++) {
    if (ls[i]) {
      WCRec d;
      VECMAC_3_VEC(d, =, ls[i]->pts[0], -, p0);
      if (VECMAC_NORM(d) < merge_tol) {
        VECMAC_3_VEC(d, =, ls[i]->pts[1], -, p1);
        if (VECMAC_NORM(d) < merge_tol) 
          return 1;
      }
      VECMAC_3_VEC(d, =, ls[i]->pts[0], -, p1);
      if (VECMAC_NORM(d) < merge_tol) {
        VECMAC_3_VEC(d, =, ls[i]->pts[1], -, p0);
        if (VECMAC_NORM(d) < merge_tol) 
          return 1;
      }
    }
  }
  return 0;
}

static LIST
update_line_list (LIST ll, double merge_tol, WCRec pts[3])
{
  int nl = list_length(ll);
  interact_rec_t ir, maxir;
  line_t **ls = NEWP_OBJ_ARRAY(nl, line_t *);
  int i;
  line_t *l, *maxl1, *maxl2;
  /* initialize */
  l=0; i=0;
  while (l = (line_t *)get_list_next(ll, l)) {
    ls[i] = l;
    i++;
  }
  /* find common points, calculate smallest angle (largest cosine) */
  maxl1 = 0; maxl2 = 0; maxir.cosa = -2;
  for (i=0; i<nl; i++) {
    int j;
    for (j=i+1; j<nl; j++) { /* notice: we do not take line w/ itself */
      ir = get_ir (ls[i]->pts, ls[j]->pts, merge_tol);
      if (ir.have_point_in_common && (ir.cosa > maxir.cosa)) {
        maxl1 = ls[i]; maxl2 = ls[j]; maxir = ir;
      }
    }
  }
  if (maxl1 && maxl1) { /* assign output */
    pts[0] = maxir.pts[0];
    pts[1] = maxir.pts[1];
    pts[2] = maxir.pts[2];
    for (i=0; i<nl; i++) { /* delete used lines */
      if (ls[i] == maxl1) ls[i] = 0;
      if (ls[i] == maxl2) ls[i] = 0;
    }
  } else { /* no points in common? could happen... */
    for (i=0; i<nl; i++) { /* delete all line references */
      ls[i] = 0;
    }
  }    
  /* rebuild ll */
  flush_list(ll, NOT_NODES);
  for (i=0; i<nl; i++) {
    if (ls[i]) add_to_tail(ll, ls[i]);
  }
  /* add the third */
  if (!have_line (ls, nl, maxir.pts[0], maxir.pts[2], merge_tol)) {
    line_t *al = NEWP_OBJ(line_t);
    al->pts[0]=maxir.pts[0];
    al->pts[1]=maxir.pts[2];
    add_to_tail(ll, al);
  }
  /* free */
  NEWP_FREE_OBJ(ls);
  return ll;
}

LIST
tripatch_trilist_from_line_loop (LIST inpll, double merge_tol)
{
  LIST ll = make_list();
  LIST tl = make_list ();
  GraphicObj *l = 0;
  while (l = (GraphicObj *)get_list_next(inpll, l)) {
    line_t *lr = NEWP_OBJ(line_t);
    WCRec pts[2];
    int npts;
    EGGetGraphicsGeometry(l, pts, &npts);
    lr->pts[0]=pts[0];
    lr->pts[1]=pts[1];
    add_to_tail(ll, lr);
  }
  while (list_length(ll) > 1) {
    WCRec pts[3];
    ll = update_line_list (ll, merge_tol, pts);
    if (! list_empty(ll))
      add_to_tail (tl, CreateTriangle3D(pts));
  }
  free_list(ll, AND_NODES);
  return tl;
}
