/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "Econfig.h"  
#include "Ecfg.h"  
#include <ctype.h>
#include "Etypein.h"
#include "ioP.h"
#include "tokensP.h" /* of Ckit library */

#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

/* ========================================================================= */
/* PRIVATE */
/* ========================================================================= */

static top_command *top_command_table = NULL;

static int commands_total = 0;	/* number of fields in the commands[] array */

static tokens_parser parser = NULL;


/* ========================================================================= */
/* PUBLIC */
/* ========================================================================= */

/* ========================================================================= */

void 
TypeInInputData(char *in_file) /* TypeInSetCommandTable must be */
				      /* called as the first thing before */
				      /* entering TypeInInputData */
{
  TypeInInitCommandLine();
  TypeInOpenInFile(in_file);
  while(TypeInReadNextLine() != EOF) {
    top_command_table[TypeInCurrentCommand()].handle_it();
  }
}

/* ========================================================================= */

void 
TypeInInterpretCurrentCommand()
{
  if (top_command_table[TypeInCurrentCommand()].handle_it != NULL) {
    /* TypeInPrintLineContents (); */
    top_command_table[TypeInCurrentCommand()].handle_it();
  }
}
 
/* ========================================================================= */
/* The table MUST start by { NULL, 0, TypeInErrorInInput }. */

void 
TypeInSetCommandTable(top_command *table, int table_size)
{
  top_command_table = table;
  commands_total    = table_size;
  TypeInInitCommandLine();
}

/* ========================================================================= */
/* Works similarly to TypeInSetCommandTable when there was no command */
/* table active; otherwise appends the new command table at the end. */
/* The table must not start by { NULL, 0, TypeInErrorInInput }. */

top_command *
TypeInAugmentCommandTable (top_command *table, int table_size)
{
  top_command *old_table, *new_table;
  int old_commands_total, new_commands_total, i, j;

  if (table == NULL)
    return top_command_table;   /* request to get the current command table */
  
  old_table          = top_command_table;
  old_commands_total = commands_total;
  if (old_table == NULL) {
    old_commands_total = 1; /* take account of the nil command */
  }
  new_commands_total = old_commands_total + table_size;

  new_table = (top_command *)make_node(sizeof(top_command)*new_commands_total);
  if (new_table == NULL) {
    fprintf(stderr, "Failed allocate memory\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  if (old_table == NULL) {
    new_table[0].keyword        = NULL;
    new_table[0].keyword_length = 0;
    new_table[0].handle_it      = TypeInErrorInInput;
    new_table[0].syntax         = NULL;
  } else {
    for (i = 0; i < old_commands_total; i++) {
      if (old_table[i].keyword != NULL)
        new_table[i].keyword        = strdup(old_table[i].keyword);
      else
        new_table[i].keyword        = NULL;
      new_table[i].keyword_length = old_table[i].keyword_length;
      new_table[i].handle_it      = old_table[i].handle_it;
      if (old_table[i].syntax != NULL)
        new_table[i].syntax         = strdup(old_table[i].syntax);
      else
        new_table[i].syntax         = NULL;
    }
  }
  for (j = 0; j < table_size; j++) {
    i = j + old_commands_total;
    new_table[i].keyword        = strdup(table[j].keyword);
    new_table[i].keyword_length = table[j].keyword_length;
    new_table[i].handle_it      = table[j].handle_it;
    if (table[j].syntax != NULL)
      new_table[i].syntax         = strdup(table[j].syntax);
    else
      new_table[i].syntax         = NULL;
  }
  top_command_table = new_table;
  commands_total    = new_commands_total;
  TypeInInitCommandLine();
  return old_table;
}

/* ========================================================================= */

int 
TypeInCurrentCommand ()
{
  int     i;
  char *tok;
  int chosen = 0;
  int matched_for_chosen = 0;
  int matched = 0;

  tok = TypeInGetTokenAsString(1);

  /* Note that we start at i = 1 (the error_handler is at the zero index) */
  /* Find maximum match, ie, compare with all command keywords in table
     and pick that with maximum match.
   */
  for(i = 1; i < commands_total; i++) {
    matched = TypeInCompareTokenWithKeyword (tok,
                                             top_command_table[i].keyword);
    if (matched > 0) {
      if (matched > matched_for_chosen) {
        chosen = i;
        matched_for_chosen = matched;
      }
    }
  }

  return chosen;
}

/* ========================================================================= */

void 
TypeInInitCommandLine ()
{
  if (top_command_table == NULL) {
    fprintf(stderr, "No command table!?\n");
    exit (ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  parser = tokens_new_parser ();
  if (!parser) {
    fprintf(stderr, "No parser!?\n");
    exit (ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
}

/* ========================================================================= */

int 
TypeInReadNextLine () /* returns YES for valid input obtained or EOF */
{
  return tokens_next_line (parser);
}

/* ========================================================================= */

void 
TypeInOpenInFile (char *in_file) 
{
  if (!tokens_open_file (parser, in_file)) {
    fprintf(stderr, "\nFailed to open input file \"%s\".", in_file);
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
}

/* ========================================================================= */

int 
TypeInParseLine (char *to_be_parsed)
{
  return tokens_parse_line (parser, to_be_parsed);
}

/* ========================================================================= */

int 
TypeInGetTokensTotal ()
{
  return tokens_total_of_tokens (parser);
}

/* ========================================================================= */

token_type 
TypeInGetTokenType (int token_number)
{
  return tokens_type_of_token (parser, token_number);
}

/* ========================================================================= */

char *
TypeInGetTokenAsString (int token_number)
{
  return (char *) tokens_token_as_string (parser, token_number);
}

/* ========================================================================= */

double 
TypeInGetTokenAsDouble (int token_number)
{
  return tokens_token_as_double (parser, token_number);
}     

/* ========================================================================= */

int 
TypeInGetTokenAsInt (int token_number)
{
  return tokens_token_as_int (parser, token_number);
}     

/* ========================================================================= */

int 
TypeInBadToken (int ith, char *should_be_type)
{
  if (ith <= TypeInGetTokensTotal()) {
    fprintf(stderr, "\nBad token type (should be a %s) in line #%d:\n "
            "   %dth token = \"%s\".\n",
            should_be_type,
            tokens_curr_line_num (parser),
            ith, TypeInGetTokenAsString (ith));
  } else {
    fprintf(stderr, "\nBad token type (should be a %s) in line #%d:\n "
            "   NO %dth token (total of tokens = %d).\n",
            should_be_type,
            tokens_curr_line_num (parser), ith, TypeInGetTokensTotal());
  }
  return 0;
}
	
void 
TypeInPrintLineContents ()
{
  int tot, i;
  token_type type;
  
  printf("\nLine # %d --> number of tokens: %d",
         tokens_curr_line_num (parser), tot = TypeInGetTokensTotal());
  for (i=1; i <= tot; i++){
    type = TypeInGetTokenType(i);
    switch((int)type)	{
    case (int)NUMBER:
      printf("\n  %d type = NUMBER, val = %f",
	     i,TypeInGetTokenAsDouble(i));
      break;
    case (int)STRING:
      printf("\n  %d type = STRING, val = %s",
	     i,TypeInGetTokenAsString(i));
      break;
    default:
      printf("\n  %d type=UNKNOWN",i);
    }
  }
  printf("\n==========================================\n");
}

/* ========================================================================= */

/*
   Keyword      Token      Match?
    MV           MVNUMS     no
    MVNUM        MVNUMS     no (note that the S is missing in keyword)
    MV           MV         yes
    MVNUMS       MV         yes, but provided token `MVNUMS' 
                            is also present it will be chosen 
                            as a better match
    MVNUMS       MVN        yes (remark above applies)
 */
int 
TypeInIsKeywordInLine (int start_at, char *keyword)
{
  int i, tot;
  char  *tok;
  int chosen = 0;
  int matched_for_chosen = 0;
  int matched = 0;
  int kwlen = strlen (keyword);
  int toklen;
 
  for(i = (start_at > 0)?start_at:1,
      tot = TypeInGetTokensTotal(); i <= tot; i++)    {
    tok = TypeInGetTokenAsString(i);
    toklen = strlen (tok);
    matched = TypeInCompareTokenWithKeyword (tok, keyword);
    if (matched > 0 &&
        matched == kwlen &&
        kwlen >= toklen) {
      if (matched > matched_for_chosen) {
        chosen = i;
        matched_for_chosen = matched;
      }
    }
  }
  return chosen;
}

/* ========================================================================= */

void 
TypeInErrorInInput ()
{
  fprintf(stderr, "\nUnknown top keyword \"%s\" in line#%d. Ignoring.\n",
	  TypeInGetTokenAsString(1),
          tokens_curr_line_num (parser));
}

/* ========================================================================= */

/*
   Denoting ltok and lkey the number of characters in the strings tok
   and keyword respectively, the function compares the min (ltok, lkey)
   characters.  If these match, the number min (ltok, lkey) is
   returned; otherwise the function returns 0.
 */
int
TypeInCompareTokenWithKeyword (char *tok, char *keyword)
{
  int j, len;

  len = min (strlen (tok), strlen (keyword));
  for(j = 0; j < len; j++) {
    if (toupper (*(tok+j)) != toupper (*(keyword+j)))	{
      return 0;
    }
  }
  return len;
}

/* ========================================================================= */

void 
TypeInBadSyntax (char *to_print)
{
  fprintf(stderr, "\nBad syntax in line#%d. %s\n",
          tokens_curr_line_num (parser), to_print);
}


top_command *
TypeInGetCommandData (char *top_keyword)
{
  int i;
  int matched = 0;

  for (i = 1; i < commands_total; i++) {
    matched = TypeInCompareTokenWithKeyword (top_keyword,
                                             top_command_table[i].keyword);
    if (matched == strlen (top_keyword)) {
      return &top_command_table[i];
    }
  }

  return NULL; /* No command table or no such keyword */
}


top_command *
TypeInReplaceCommand (top_command *cmd)
{
  int i;
  int matched = 0;
  
  if (top_command_table != NULL) {
    for (i = 1; i < commands_total; i++) {
      matched
        = TypeInCompareTokenWithKeyword (cmd->keyword,
                                         top_command_table[i].keyword);
      if (matched == strlen (cmd->keyword)) {
        /* The command was already in the table */
        top_command_table[i].handle_it = cmd->handle_it;
        if (top_command_table[i].syntax != NULL)
          free(top_command_table[i].syntax);
        if (cmd->syntax != NULL)
          top_command_table[i].syntax = strdup(cmd->syntax);
        return NULL;
      }
    }
  } 
  /* The command is new */
  return TypeInAugmentCommandTable(cmd, 1);
}



int 
TypeInCommandTableSize(void)
{
  return commands_total;
}
